
// A86T.c : A86 test layer

#include "types.h"
#include "a86.h"
#include "goptions.h"    // for address_size (decides whether 32 or 64-bit)

void generate_test_assembler_code ()
{
  {
    const int OFS[3] = {0, 0x12, 0xFFD2};
    int d, siz;
    for (siz=1; siz<=address_size; siz<<=1)
    {
      for (d=0; d<3; d++)
      {
        int ofs = OFS[d];

        EA m1 = {NONE, NONE, 1, ofs};  // global
        EA m2 = {RBP, NONE,  1, ofs};  // local
        EA m3 = {RBP, RDI,   2, ofs};  // local + index*2

        EA m4 = {RAX, NONE,  1, ofs};
        EA m5 = {R12, NONE,  1, ofs};
        EA m6 = {R13, NONE,  1, ofs};
        EA m7 = {R15, NONE,  1, ofs};

        EA m8  = {NONE, RAX,  4, ofs};
        EA m9  = {NONE, R12,  4, ofs};
        EA m10 = {NONE, R13,  4, ofs};
        EA m11 = {NONE, R15,  4, ofs};
        EA m12 = {RBP,  RAX,  4, ofs};

        EA m13 = {RSI, R12,  1, ofs};
        EA m14 = {R12, R13,  1, ofs};
        EA m15 = {R13, RDI,  1, ofs};
        EA m16 = {R13, R12,  1, ofs};

        EA m17 = {RSP, R8,   1, ofs};
        EA m18 = {RIP, NONE, 1, ofs};

        c_add_reg_mem (RAX, &m1, siz);
        c_add_reg_mem (RBX, &m2, siz);
        c_add_reg_mem (RCX, &m3, siz);
        c_add_reg_mem (RSP, &m4, siz);

        if (address_size == 8)
        {
          c_add_reg_mem (RAX, &m5, siz);
          c_add_reg_mem (RBX, &m6, siz);
          c_add_reg_mem (RDI, &m7, siz);
        }

        c_add_reg_mem (RSI, &m8, siz);

        if (address_size == 8)
        {
          c_add_reg_mem (R8, &m9, siz);
          c_add_reg_mem (R9, &m10, siz);
          c_add_reg_mem (R10, &m11, siz);
        }

        c_add_reg_mem (R11, &m12, siz);

        if (address_size == 8)
        {
          c_add_reg_mem (R12, &m13, siz);
          c_add_reg_mem (R13, &m14, siz);
          c_add_reg_mem (R14, &m15, siz);
          c_add_reg_mem (R15, &m16, siz);

          c_add_reg_mem (RAX, &m17, siz);
          c_add_reg_mem (RBX, &m18, siz);
        }

        c_add_reg_mem (RCX, &m1, siz);
        c_add_reg_mem (RDX, &m1, siz);

        if (address_size == 8)
        {
          c_add_reg_mem (RSP, &m1, siz);
          c_add_reg_mem (RSP, &m2, siz);
          c_add_reg_mem (RSI, &m3, siz);
        }

        c_add_mem_reg (&m1, RDI, siz);

        if (address_size == 8)
        {
          c_add_mem_reg (&m1, R12, siz);
          c_add_mem_reg (&m2, R13, siz);
          c_add_mem_reg (&m3, R14, siz);

        }

        c_add_mem_imm (&m1, 0x67, siz);
      }
    }
  }


  {
    int siz;
    for (siz=1; siz<=address_size; siz<<=1)
    {
      c_add_reg_reg (RAX, RBX, siz);
      c_mov_reg_reg (RAX, RBX, siz);

      c_add_reg_reg (RAX, R12, siz);
      c_mov_reg_reg (RAX, R12, siz);

      c_add_reg_reg (RSI, RBX, siz);
      c_mov_reg_reg (RSI, RBX, siz);

      c_add_reg_reg (R12, RBX, siz);
      c_mov_reg_reg (R12, RBX, siz);

      c_add_reg_reg (RSI, R13, siz);
      c_mov_reg_reg (RSI, R13, siz);
    }
  }

  {
    int siz;
    for (siz=1; siz<=address_size; siz<<=1)
    {
      c_add_reg_imm (RAX, 1, siz);

      if (address_size == 8)
      {
        c_add_reg_imm (RSI, 0x200, siz);
        c_add_reg_imm (R12, 0x02FFFF, siz);
        c_add_reg_imm (R13, 0x02FFFF, siz);
        c_add_reg_imm (R15, 15, siz);
      }
    }
  }

  c_sub_reg_reg (RAX, RAX, 1);
  c_sub_reg_reg (RAX, RSI, 1);
  c_sub_reg_reg (RAX, RBX, 2);

  {
    EA m = {RBP, RDI, 2, 0x023};  // local + index
    int siz;
    for (siz=1; siz<=address_size; siz<<=1)
    {
      c_sub_reg_mem (RAX, &m, siz);
      c_sub_reg_mem (RSI, &m, siz);
      c_sub_reg_mem (R12, &m, siz);
    }
  }


  c_call_reg (RAX, address_size, 0);
  c_call_reg (RBX, address_size, 0);
  c_call_reg (RSP, address_size, 0);
  c_call_reg (R8, address_size, 0);
  c_call_reg (R12, address_size, 0);
  c_call_reg (R13, address_size, 0);
  c_call_reg (R15, address_size, 0);


  {
    const int OFS[3] = {0, 0x12, 0xFFD2};
    int d;
    for (d=0; d<3; d++)
    {
      int ofs = OFS[d];

      EA m1 = {NONE, NONE, 1, ofs};  // global
      EA m2 = {RBP, NONE,  1, ofs};  // local
      EA m3 = {RBP, RDI,   2, ofs};  // local + index*2

      EA m4 = {RAX, NONE,  1, ofs};
      EA m5 = {R12, NONE,  1, ofs};
      EA m6 = {R13, RAX,  1, ofs};
      EA m7 = {R15, NONE,  1, ofs};

      c_call_indirect (&m1, address_size, 0);
      c_call_indirect (&m2, address_size, 0);
      c_call_indirect (&m3, address_size, 0);

      c_call_indirect (&m4, address_size, 0);
      c_call_indirect (&m5, address_size, 0);
      c_call_indirect (&m6, address_size, 0);
      c_call_indirect (&m7, address_size, 0);
    }
  }

  {
    const int OFS[3] = {0, 0x12, 0xFFD2};
    int d;
    for (d=0; d<3; d++)
    {
      int ofs = OFS[d];

      EA m1 = {NONE, NONE, 1, ofs};  // global
      EA m2 = {RBP, NONE,  1, ofs};  // local
      EA m3 = {RBP, RDI,   2, ofs};  // local + index*2

      EA m4 = {RAX, NONE,  1, ofs};
      EA m5 = {R12, NONE,  1, ofs};
      EA m6 = {R13, RAX,  1, ofs};
      EA m7 = {R15, NONE,  1, ofs};

      c_jump_indirect (&m1, address_size);
      c_jump_indirect (&m2, address_size);
      c_jump_indirect (&m3, address_size);

      c_jump_indirect (&m4, address_size);
      c_jump_indirect (&m5, address_size);
      c_jump_indirect (&m6, address_size);
      c_jump_indirect (&m7, address_size);
    }
  }


  c_setcond_reg (1, true,  RAX, 1);
  c_setcond_reg (1, false, RBX, 1);

  if (address_size == 8)
  {
    c_setcond_reg (1, true, RBP, 1);
    c_setcond_reg (1, true, R12, 1);
  }


  {
    EA m1 = {NONE, NONE, 1, 0x37};  // global
    EA m2 = {RBP, NONE,  1, 0x43};  // local
    EA m3 = {RBP, RDI,   2, 0x98};  // local + index*2

    c_setcond_mem (1, true, &m1, 1);
    c_setcond_mem (1, true, &m2, 1);
    c_setcond_mem (1, true, &m3, 1);


    c_ret (0);
    c_ret (8);

    c_int (5);

    c_lea_reg_mem (RCX, &m1, 4);
    c_lea_reg_mem (RCX, &m1, address_size);

    c_lea_reg_mem (RCX, &m2, 4);
    c_lea_reg_mem (RCX, &m2, address_size);

    c_lea_reg_mem (RCX, &m3, 4);
    c_lea_reg_mem (RCX, &m3, address_size);


    c_push_reg (RAX, address_size);
    c_push_reg (RBX, address_size);
    c_push_reg (RBP, address_size);
    c_push_reg (R9, address_size);

    c_push_mem (&m1, address_size);
    c_push_mem (&m2, address_size);
    c_push_mem (&m3, address_size);

    c_push_imm (12);
    c_push_imm (0xFFD2);

    c_pop_reg (RBX, address_size);
    c_pop_reg (RBP, address_size);
    c_pop_reg (R12, address_size);

    c_pop_mem (&m1, address_size);
    c_pop_mem (&m2, address_size);
    c_pop_mem (&m3, address_size);
  }



  {
    int siz;
    for (siz=1; siz<=address_size; siz<<=1)
    {
      c_neg_reg (RAX, siz);
      c_neg_reg (RBX, siz);
      c_neg_reg (RSP, siz);
      c_neg_reg (R8, siz);
      c_neg_reg (R12, siz);
      c_neg_reg (R13, siz);
      c_neg_reg (R15, siz);
    }
  }



  {
    const int OFS[3] = {0, 0x12, 0xFFD2};
    int d, siz;
    for (siz=1; siz<=address_size; siz<<=1)
    {
      for (d=0; d<3; d++)
      {
        int ofs = OFS[d];

        EA m1 = {NONE, NONE, 1, ofs};  // global
        EA m2 = {RBP, NONE,  1, ofs};  // local
        EA m3 = {RBP, RDI,   2, ofs};  // local + index*2

        EA m4 = {RAX, NONE,  1, ofs};
        EA m5 = {R12, NONE,  1, ofs};
        EA m6 = {R13, RAX,  1, ofs};
        EA m7 = {R15, NONE,  1, ofs};

        c_neg_mem (&m1, siz);
        c_neg_mem (&m2, siz);
        c_neg_mem (&m3, siz);

        c_neg_mem (&m4, siz);
        c_neg_mem (&m5, siz);
        c_neg_mem (&m6, siz);
        c_neg_mem (&m7, siz);
      }
    }
  }


  {
    int siz;
    for (siz=1; siz<=address_size; siz<<=1)
    {
      c_not_reg (RAX, siz);
      c_not_reg (RBX, siz);
      c_not_reg (RSP, siz);
      c_not_reg (R8, siz);
      c_not_reg (R12, siz);
      c_not_reg (R13, siz);
      c_not_reg (R15, siz);
    }
  }


  {
    const int OFS[3] = {0, 0x12, 0xFFD2};
    int d, siz;
    for (siz=1; siz<=address_size; siz<<=1)
    {
      for (d=0; d<3; d++)
      {
        int ofs = OFS[d];

        EA m1 = {NONE, NONE, 1, ofs};  // global
        EA m2 = {RBP, NONE,  1, ofs};  // local
        EA m3 = {RBP, RDI,   2, ofs};  // local + index*2

        EA m4 = {RAX, NONE,  1, ofs};
        EA m5 = {R12, NONE,  1, ofs};
        EA m6 = {R13, RAX,  1, ofs};
        EA m7 = {R15, NONE,  1, ofs};

        c_not_mem (&m1, siz);
        c_not_mem (&m2, siz);
        c_not_mem (&m3, siz);

        c_not_mem (&m4, siz);
        c_not_mem (&m5, siz);
        c_not_mem (&m6, siz);
        c_not_mem (&m7, siz);
      }
    }
  }

  {
    int siz;
    for (siz=1; siz<=address_size; siz<<=1)
    {
      c_inc_reg (RAX, siz);
      c_inc_reg (RBX, siz);
      c_inc_reg (RSP, siz);
      c_inc_reg (R8, siz);
      c_inc_reg (R12, siz);
      c_inc_reg (R13, siz);
      c_inc_reg (R15, siz);
    }
  }

  {
    int siz;
    for (siz=1; siz<=address_size; siz<<=1)
    {
      c_dec_reg (RAX, siz);
      c_dec_reg (RBX, siz);
      c_dec_reg (RSP, siz);
      c_dec_reg (R8, siz);
      c_dec_reg (R12, siz);
      c_dec_reg (R13, siz);
      c_dec_reg (R15, siz);
    }
  }


  {
    const int OFS[3] = {0, 0x12, 0xFFD2};
    int d, siz;
    for (siz=1; siz<=address_size; siz<<=1)
    {
      for (d=0; d<3; d++)
      {
        int ofs = OFS[d];

        EA m1 = {NONE, NONE, 1, ofs};  // global
        EA m2 = {RBP, NONE,  1, ofs};  // local
        EA m3 = {RBP, RDI,   2, ofs};  // local + index*2

        EA m4 = {RAX, NONE,  1, ofs};
        EA m5 = {R12, NONE,  1, ofs};
        EA m6 = {R13, RAX,  1, ofs};
        EA m7 = {R15, NONE,  1, ofs};

        c_inc_mem (&m1, siz);
        c_inc_mem (&m2, siz);
        c_inc_mem (&m3, siz);

        c_inc_mem (&m4, siz);
        c_inc_mem (&m5, siz);
        c_inc_mem (&m6, siz);
        c_inc_mem (&m7, siz);
      }
    }
  }


  {
    const int OFS[3] = {0, 0x12, 0xFFD2};
    int d, siz;
    for (siz=1; siz<=address_size; siz<<=1)
    {
      for (d=0; d<3; d++)
      {
        int ofs = OFS[d];

        EA m1 = {NONE, NONE, 1, ofs};  // global
        EA m2 = {RBP, NONE,  1, ofs};  // local
        EA m3 = {RBP, RDI,   2, ofs};  // local + index*2

        EA m4 = {RAX, NONE,  1, ofs};
        EA m5 = {R12, NONE,  1, ofs};
        EA m6 = {R13, RAX,  1, ofs};
        EA m7 = {R15, NONE,  1, ofs};

        c_dec_mem (&m1, siz);
        c_dec_mem (&m2, siz);
        c_dec_mem (&m3, siz);

        c_dec_mem (&m4, siz);
        c_dec_mem (&m5, siz);
        c_dec_mem (&m6, siz);
        c_dec_mem (&m7, siz);
      }
    }
  }


  {
    int siz;
    for (siz=1; siz<=address_size; siz<<=1)
    {
      c_xchg_reg_reg (RAX, RBX, siz);
      c_xchg_reg_reg (RAX, RBX, siz);
      c_xchg_reg_reg (RAX, RSP, siz);
      c_xchg_reg_reg (RAX, R12, siz);
      c_xchg_reg_reg (RAX, R13, siz);
      c_xchg_reg_reg (RAX, R15, siz);

      c_xchg_reg_reg (RAX, R8, siz);
      c_xchg_reg_reg (RBX, R8, siz);
      c_xchg_reg_reg (RSP, RAX, siz);
      c_xchg_reg_reg (R12, R8, siz);
      c_xchg_reg_reg (R13, R8, siz);
      c_xchg_reg_reg (R15, R8, siz);
    }
  }


  {
    const int OFS[3] = {0, 0x12, 0xFFD2};
    int d, siz;
    for (siz=1; siz<=address_size; siz<<=1)
    {
      for (d=0; d<3; d++)
      {
        int ofs = OFS[d];

        EA m1 = {NONE, NONE, 1, ofs};  // global
        EA m2 = {RBP, NONE,  1, ofs};  // local
        EA m3 = {RBP, RDI,   2, ofs};  // local + index*2

        EA m4 = {RAX, NONE,  1, ofs};
        EA m5 = {R12, NONE,  1, ofs};
        EA m6 = {R13, RAX,   1, ofs};
        EA m7 = {R15, NONE,  1, ofs};

        c_xchg_reg_mem (RAX, &m1, siz);
        c_xchg_reg_mem (RBX, &m2, siz);
        c_xchg_reg_mem (RCX, &m3, siz);
        c_xchg_reg_mem (RSP, &m4, siz);
        c_xchg_reg_mem (R8,  &m5, siz);
        c_xchg_reg_mem (R12, &m6, siz);
        c_xchg_reg_mem (R13, &m7, siz);
      }
    }
  }


  {
    int siz;
    for (siz=1; siz<=address_size; siz<<=1)
    {
      c_test_reg_reg (RAX, RAX, siz);
      c_test_reg_reg (RAX, RBX, siz);
      c_test_reg_reg (RAX, RSP, siz);
      c_test_reg_reg (RAX, R12, siz);
      c_test_reg_reg (RAX, R13, siz);
      c_test_reg_reg (RAX, R15, siz);

      c_test_reg_reg (RAX, R8, siz);
      c_test_reg_reg (RBX, R8, siz);
      c_test_reg_reg (RSP, R8, siz);
      c_test_reg_reg (R12, R8, siz);
      c_test_reg_reg (R13, R8, siz);
      c_test_reg_reg (R15, R8, siz);
    }
  }


  {
    const int OFS[3] = {0, 0x12, 0xFFD2};
    int d, siz;
    for (siz=1; siz<=address_size; siz<<=1)
    {
      for (d=0; d<3; d++)
      {
        int ofs = OFS[d];

        EA m1 = {NONE, NONE, 1, ofs};  // global
        EA m2 = {RBP, NONE,  1, ofs};  // local
        EA m3 = {RBP, RDI,   2, ofs};  // local + index*2

        EA m4 = {RAX, NONE,  1, ofs};
        EA m5 = {R12, NONE,  1, ofs};
        EA m6 = {R13, RAX,   1, ofs};
        EA m7 = {R15, NONE,  1, ofs};

        c_test_mem_reg (&m1, RAX, siz);
        c_test_mem_reg (&m2, RBX, siz);
        c_test_mem_reg (&m3, RCX, siz);
        c_test_mem_reg (&m4, RSP, siz);
        c_test_mem_reg (&m5, R8,  siz);
        c_test_mem_reg (&m6, R12, siz);
        c_test_mem_reg (&m7, R13, siz);
      }
    }
  }


  {
    int siz;
    for (siz=1; siz<=address_size; siz<<=1)
    {
      c_test_reg_imm (RAX, 1, siz);
      c_test_reg_imm (RBX, 2, siz);
      c_test_reg_imm (RSP, 3, siz);
      c_test_reg_imm (R12, 4, siz);
      c_test_reg_imm (R13, 5, siz);
      c_test_reg_imm (R15, 6, siz);
    }
  }


  {
    const int OFS[3] = {0, 0x12, 0xFFD2};
    int d, siz;
    for (siz=1; siz<=address_size; siz<<=1)
    {
      for (d=0; d<3; d++)
      {
        int ofs = OFS[d];

        EA m1 = {NONE, NONE, 1, ofs};  // global
        EA m2 = {RBP, NONE,  1, ofs};  // local
        EA m3 = {RBP, RDI,   2, ofs};  // local + index*2

        EA m4 = {RAX, NONE,  1, ofs};
        EA m5 = {R12, NONE,  1, ofs};
        EA m6 = {R13, RAX,   1, ofs};
        EA m7 = {R15, NONE,  1, ofs};

        c_test_mem_imm (&m1, 1, siz);
        c_test_mem_imm (&m2, 2, siz);
        c_test_mem_imm (&m3, 3, siz);
        c_test_mem_imm (&m4, 4, siz);
        c_test_mem_imm (&m5, 5,  siz);
        c_test_mem_imm (&m6, 6, siz);
        c_test_mem_imm (&m7, 7, siz);
      }
    }
  }


  {
    int siz;
    for (siz=1; siz<=address_size; siz<<=1)
    {
      c_mov_reg_reg (RAX, RAX, siz);
      c_mov_reg_reg (RAX, RBX, siz);
      c_mov_reg_reg (RAX, RSP, siz);
      c_mov_reg_reg (RAX, R12, siz);
      c_mov_reg_reg (RAX, R13, siz);
      c_mov_reg_reg (RAX, R15, siz);

      c_mov_reg_reg (RAX, R8, siz);
      c_mov_reg_reg (RBX, R8, siz);
      c_mov_reg_reg (RSP, R8, siz);
      c_mov_reg_reg (R12, R8, siz);
      c_mov_reg_reg (R13, R8, siz);
      c_mov_reg_reg (R15, R8, siz);
    }
  }



  {
    const int OFS[3] = {0, 0x12, 0xFFD2};
    int d, siz;
    for (siz=1; siz<=address_size; siz<<=1)
    {
      for (d=0; d<3; d++)
      {
        int ofs = OFS[d];

        EA m1 = {NONE, NONE, 1, ofs};  // global
        EA m2 = {RBP, NONE,  1, ofs};  // local
        EA m3 = {RBP, RDI,   2, ofs};  // local + index*2

        EA m4 = {RAX, NONE,  1, ofs};
        EA m5 = {R12, NONE,  1, ofs};
        EA m6 = {R13, RAX,   1, ofs};
        EA m7 = {R15, NONE,  1, ofs};

        c_mov_reg_mem (RAX, &m1, siz);
        c_mov_reg_mem (RAX, &m2, siz);
        c_mov_reg_mem (RAX, &m3, siz);
        c_mov_reg_mem (RAX, &m4, siz);
        c_mov_reg_mem (RAX, &m5, siz);
        c_mov_reg_mem (RAX, &m6, siz);
        c_mov_reg_mem (RAX, &m7, siz);
      }
    }
  }


  {
    const int OFS[3] = {0, 0x12, 0xFFD2};
    int d, siz;
    for (siz=1; siz<=address_size; siz<<=1)
    {
      for (d=0; d<3; d++)
      {
        int ofs = OFS[d];

        EA m1 = {NONE, NONE, 1, ofs};  // global
        EA m2 = {RBP, NONE,  1, ofs};  // local
        EA m3 = {RBP, RDI,   2, ofs};  // local + index*2

        EA m4 = {RAX, NONE,  1, ofs};
        EA m5 = {R12, NONE,  1, ofs};
        EA m6 = {R13, RAX,   1, ofs};
        EA m7 = {R15, NONE,  1, ofs};

        c_mov_reg_mem (RAX, &m1, siz);
        c_mov_reg_mem (RBX, &m2, siz);
        c_mov_reg_mem (RCX, &m3, siz);
        c_mov_reg_mem (RDX, &m4, siz);
        c_mov_reg_mem (RSP, &m5, siz);
        c_mov_reg_mem (R12, &m6, siz);
        c_mov_reg_mem (R14, &m7, siz);
      }
    }
  }


  {
    const int OFS[3] = {0, 0x12, 0xFFD2};
    int d, siz;
    for (siz=1; siz<=address_size; siz<<=1)
    {
      for (d=0; d<3; d++)
      {
        int ofs = OFS[d];

        EA m1 = {NONE, NONE, 1, ofs};  // global
        EA m2 = {RBP, NONE,  1, ofs};  // local
        EA m3 = {RBP, RDI,   2, ofs};  // local + index*2

        EA m4 = {RAX, NONE,  1, ofs};
        EA m5 = {R12, NONE,  1, ofs};
        EA m6 = {R13, RAX,   1, ofs};
        EA m7 = {R15, NONE,  1, ofs};

        c_mov_mem_reg (&m1, RAX, siz);
        c_mov_mem_reg (&m2, RAX, siz);
        c_mov_mem_reg (&m3, RAX, siz);
        c_mov_mem_reg (&m4, RAX, siz);
        c_mov_mem_reg (&m5, RAX, siz);
        c_mov_mem_reg (&m6, RAX, siz);
        c_mov_mem_reg (&m7, RAX, siz);
      }
    }
  }


  {
    int siz;
    for (siz=1; siz<=address_size; siz<<=1)
    {
      c_mov_reg_imm (RAX, 0, siz);
      c_mov_reg_imm (RBX, 1, siz);
      c_mov_reg_imm (RSP, 2, siz);
      c_mov_reg_imm (R12, 3, siz);
      c_mov_reg_imm (R13, 4, siz);
      c_mov_reg_imm (R15, 5, siz);
    }
  }

  c_mov_reg_imm (RAX, -7, address_size);
  c_mov_reg_imm (R12, -7, address_size);
  c_mov_reg_imm (RAX, 3000000000, address_size);
  c_mov_reg_imm (R12, 3000000000, address_size);
  c_mov_reg_imm (RAX, 30000000000, address_size);
  c_mov_reg_imm (R12, 30000000000, address_size);


  {
    const int OFS[3] = {0, 0x12, 0xFFD2};
    int d, siz;
    for (siz=1; siz<=address_size; siz<<=1)
    {
      for (d=0; d<3; d++)
      {
        int ofs = OFS[d];

        EA m1 = {NONE, NONE, 1, ofs};  // global
        EA m2 = {RBP, NONE,  1, ofs};  // local
        EA m3 = {RBP, RDI,   2, ofs};  // local + index*2

        EA m4 = {RAX, NONE,  1, ofs};
        EA m5 = {R12, NONE,  1, ofs};
        EA m6 = {R13, RAX,   1, ofs};
        EA m7 = {R15, NONE,  1, ofs};

        c_mov_mem_imm (&m1, 1, siz);
        c_mov_mem_imm (&m2, 2, siz);
        c_mov_mem_imm (&m3, 3, siz);
        c_mov_mem_imm (&m4, 4, siz);
        c_mov_mem_imm (&m5, 5,  siz);
        c_mov_mem_imm (&m6, 6, siz);
        c_mov_mem_imm (&m7, 7, siz);
      }
    }
  }


  {
    int siz, siz0;
    for (siz=2; siz<=address_size; siz<<=1)
    {
      for (siz0=1; siz0<siz; siz0<<=1)
      {
        c_movsx_reg_reg (RAX, siz, RAX, siz0);
      }
    }
  }

  {
    int siz, siz0;
    for (siz=2; siz<=address_size; siz<<=1)
    {
      for (siz0=1; siz0<siz; siz0<<=1)
      {
        c_movsx_reg_reg (RAX, siz, RBX, siz0);
        c_movsx_reg_reg (RAX, siz, RBP, siz0);
        c_movsx_reg_reg (RAX, siz, R12, siz0);
        c_movsx_reg_reg (RAX, siz, R13, siz0);
        c_movsx_reg_reg (RAX, siz, R15, siz0);

        c_movsx_reg_reg (RBX, siz, RAX, siz0);
        c_movsx_reg_reg (RBP, siz, RAX, siz0);
        c_movsx_reg_reg (R12, siz, RAX, siz0);
        c_movsx_reg_reg (R13, siz, RAX, siz0);
        c_movsx_reg_reg (R15, siz, RAX, siz0);
      }
    }
  }


  {
    const int ofs = 3;

    EA m1 = {NONE, NONE, 1, ofs};  // global
    EA m2 = {RBP, NONE,  1, ofs};  // local
    EA m3 = {RBP, RDI,   2, ofs};  // local + index*2
    EA m4 = {RAX, NONE,  1, ofs};
    EA m5 = {R12, NONE,  1, ofs};

    int siz, siz0;
    for (siz=2; siz<=address_size; siz<<=1)
    {
      for (siz0=1; siz0<siz; siz0<<=1)
      {
        c_movsx_reg_mem (RBX, siz, &m1, siz0);
        c_movsx_reg_mem (RBP, siz, &m2, siz0);
        c_movsx_reg_mem (R12, siz, &m3, siz0);
        c_movsx_reg_mem (R13, siz, &m4, siz0);
        c_movsx_reg_mem (R15, siz, &m5, siz0);
      }
    }
  }




  {
    int siz, siz0;
    for (siz=2; siz<=address_size; siz<<=1)
    {
      for (siz0=1; siz0<siz; siz0<<=1)
      {
        c_movzx_reg_reg (RAX, siz, RAX, siz0);
      }
    }
  }


  {
    int siz, siz0;
    for (siz=2; siz<=address_size; siz<<=1)
    {
      for (siz0=1; siz0<siz; siz0<<=1)
      {
        c_movzx_reg_reg (RAX, siz, RBX, siz0);
        c_movzx_reg_reg (RAX, siz, RBP, siz0);
        c_movzx_reg_reg (RAX, siz, R12, siz0);
        c_movzx_reg_reg (RAX, siz, R13, siz0);
        c_movzx_reg_reg (RAX, siz, R15, siz0);

        c_movzx_reg_reg (RBX, siz, RAX, siz0);
        c_movzx_reg_reg (RBP, siz, RAX, siz0);
        c_movzx_reg_reg (R12, siz, RAX, siz0);
        c_movzx_reg_reg (R13, siz, RAX, siz0);
        c_movzx_reg_reg (R15, siz, RAX, siz0);
      }
    }
  }


  {
    const int ofs = 3;

    EA m1 = {NONE, NONE, 1, ofs};  // global
    EA m2 = {RBP, NONE,  1, ofs};  // local
    EA m3 = {RBP, RDI,   2, ofs};  // local + index*2
    EA m4 = {RAX, NONE,  1, ofs};
    EA m5 = {R12, NONE,  1, ofs};

    int siz, siz0;
    for (siz=2; siz<=address_size; siz<<=1)
    {
      for (siz0=1; siz0<siz; siz0<<=1)
      {
        c_movzx_reg_mem (RBX, siz, &m1, siz0);
        c_movzx_reg_mem (RBP, siz, &m2, siz0);
        c_movzx_reg_mem (R12, siz, &m3, siz0);
        c_movzx_reg_mem (R13, siz, &m4, siz0);
        c_movzx_reg_mem (R15, siz, &m5, siz0);
      }
    }
  }


  {
    int siz;
    for (siz=2; siz<=address_size; siz<<=1)
    {
      c_imul_reg_reg (RAX, RAX, siz);
      c_imul_reg_reg (RBX, RSI, siz);
      c_imul_reg_reg (RSP, R8, siz);
      c_imul_reg_reg (R12, RAX, siz);
      c_imul_reg_reg (R13, RBX, siz);
      c_imul_reg_reg (R15, RCX, siz);
    }
  }



  {
    const int OFS[3] = {0, 0x12, 0xFFD2};
    int d, siz;
    for (siz=2; siz<=address_size; siz<<=1)
    {
      for (d=0; d<3; d++)
      {
        int ofs = OFS[d];

        EA m1 = {NONE, NONE, 1, ofs};  // global
        EA m2 = {RBP, NONE,  1, ofs};  // local
        EA m3 = {RBP, RDI,   2, ofs};  // local + index*2

        EA m4 = {RAX, NONE,  1, ofs};
        EA m5 = {R12, NONE,  1, ofs};
        EA m6 = {R13, RAX,   1, ofs};
        EA m7 = {R15, NONE,  1, ofs};

        c_imul_reg_mem (RAX, &m1, siz);
        c_imul_reg_mem (RBX, &m2, siz);
        c_imul_reg_mem (RCX, &m3, siz);
        c_imul_reg_mem (RDX, &m4, siz);
        c_imul_reg_mem (R8, &m5, siz);
        c_imul_reg_mem (R9, &m6, siz);
        c_imul_reg_mem (R13, &m7, siz);
      }
    }
  }

  {
    int siz;
    for (siz=2; siz<=address_size; siz<<=1)
    {
      c_imul_reg_imm (RAX, 0, siz);
      c_imul_reg_imm (RBX, -1, siz);
      c_imul_reg_imm (RSP, 2,  siz);
      c_imul_reg_imm (R12, 4, siz);
      c_imul_reg_imm (R13, 8, siz);
      c_imul_reg_imm (R15, 1<<30, siz);

      c_imul_reg_imm (R11, 3, siz);
      c_imul_reg_imm (R12, 5, siz);
      c_imul_reg_imm (R13, 9, siz);
      c_imul_reg_imm (R14, 10, siz);
    }
  }


  {
    int siz;
    for (siz=2; siz<=address_size; siz<<=1)
    {
      c_imul_reg_reg_imm (RAX, RSI, 3, siz);
      c_imul_reg_reg_imm (RBX, RSI, 3, siz);
      c_imul_reg_reg_imm (RSP, RSI, 3, siz);
      c_imul_reg_reg_imm (R12, RAX, 3, siz);
      c_imul_reg_reg_imm (R13, RBX, 3, siz);
      c_imul_reg_reg_imm (R15, RCX, 3, siz);

      c_imul_reg_reg_imm (RAX, RAX, 277, siz);
      c_imul_reg_reg_imm (RBX, R8, 277, siz);
      c_imul_reg_reg_imm (RSP, R8, 277, siz);
      c_imul_reg_reg_imm (R12, R8, 277, siz);
      c_imul_reg_reg_imm (R13, R8, 277, siz);
      c_imul_reg_reg_imm (R15, R8, 277, siz);
    }
  }



  {
    const int ofs = 3;

    EA m1 = {NONE, NONE, 1, ofs};  // global
    EA m2 = {RBP, NONE,  1, ofs};  // local
    EA m3 = {RBP, RDI,   2, ofs};  // local + index*2
    EA m4 = {RAX, NONE,  1, ofs};
    EA m5 = {R12, NONE,  1, ofs};

    int siz;
    for (siz=2; siz<=address_size; siz<<=1)
    {
      c_imul_reg_mem_imm (RAX, &m1, 3, siz);
      c_imul_reg_mem_imm (RBX, &m2, 3, siz);
      c_imul_reg_mem_imm (RSP, &m3, 3, siz);
      c_imul_reg_mem_imm (R12, &m4, 3, siz);
      c_imul_reg_mem_imm (R13, &m5, 3, siz);
      c_imul_reg_mem_imm (R15, &m1, 3, siz);

      c_imul_reg_mem_imm (RAX, &m1, 277, siz);
      c_imul_reg_mem_imm (RBX, &m2, 277, siz);
      c_imul_reg_mem_imm (RSP, &m3, 277, siz);
      c_imul_reg_mem_imm (R12, &m4, 277, siz);
      c_imul_reg_mem_imm (R13, &m5, 277, siz);
      c_imul_reg_mem_imm (R15, &m1, 277, siz);
    }
  }


  {
    int siz;
    for (siz=1; siz<=address_size; siz<<=1)
    {
      c_umul_rax_reg (RAX, siz);
      c_umul_rax_reg (RBX, siz);
      c_umul_rax_reg (RSP, siz);
      c_umul_rax_reg (R12, siz);
      c_umul_rax_reg (R13, siz);
      c_umul_rax_reg (R15, siz);
    }
  }

  {
    const int OFS[3] = {0, 0x12, 0xFFD2};
    int d, siz;
    for (siz=1; siz<=address_size; siz<<=1)
    {
      for (d=0; d<3; d++)
      {
        int ofs = OFS[d];

        EA m1 = {NONE, NONE, 1, ofs};  // global
        EA m2 = {RBP, NONE,  1, ofs};  // local
        EA m3 = {RBP, RDI,   2, ofs};  // local + index*2

        EA m4 = {RAX, NONE,  1, ofs};
        EA m5 = {R12, NONE,  1, ofs};
        EA m6 = {R13, RAX,   1, ofs};
        EA m7 = {R15, NONE,  1, ofs};

        c_umul_rax_mem (&m1, siz);
        c_umul_rax_mem (&m2, siz);
        c_umul_rax_mem (&m3, siz);
        c_umul_rax_mem (&m4, siz);
        c_umul_rax_mem (&m5, siz);
        c_umul_rax_mem (&m6, siz);
        c_umul_rax_mem (&m7, siz);
      }
    }
  }

  {
    int siz;
    for (siz=4; siz<=address_size; siz<<=1)
    {
      c_idiv_rax_reg (RAX, siz);
      c_idiv_rax_reg (RBX, siz);
      c_idiv_rax_reg (RSP, siz);
      c_idiv_rax_reg (R12, siz);
      c_idiv_rax_reg (R13, siz);
      c_idiv_rax_reg (R15, siz);
    }
  }

  {
    const int OFS[3] = {0, 0x12, 0xFFD2};
    int d, siz;
    for (siz=4; siz<=address_size; siz<<=1)
    {
      for (d=0; d<3; d++)
      {
        int ofs = OFS[d];

        EA m1 = {NONE, NONE, 1, ofs};  // global
        EA m2 = {RBP, NONE,  1, ofs};  // local
        EA m3 = {RBP, RDI,   2, ofs};  // local + index*2

        EA m4 = {RAX, NONE,  1, ofs};
        EA m5 = {R12, NONE,  1, ofs};
        EA m6 = {R13, RAX,   1, ofs};
        EA m7 = {R15, NONE,  1, ofs};

        c_idiv_rax_mem (&m1, siz);
        c_idiv_rax_mem (&m2, siz);
        c_idiv_rax_mem (&m3, siz);
        c_idiv_rax_mem (&m4, siz);
        c_idiv_rax_mem (&m5, siz);
        c_idiv_rax_mem (&m6, siz);
        c_idiv_rax_mem (&m7, siz);
      }
    }
  }

  {
    int siz;
    for (siz=4; siz<=address_size; siz<<=1)
    {
      c_imod_rax_reg (RAX, siz);
      c_imod_rax_reg (RBX, siz);
      c_imod_rax_reg (RSP, siz);
      c_imod_rax_reg (R12, siz);
      c_imod_rax_reg (R13, siz);
      c_imod_rax_reg (R15, siz);
    }
  }

  {
    const int OFS[3] = {0, 0x12, 0xFFD2};
    int d, siz;
    for (siz=4; siz<=address_size; siz<<=1)
    {
      for (d=0; d<3; d++)
      {
        int ofs = OFS[d];

        EA m1 = {NONE, NONE, 1, ofs};  // global
        EA m2 = {RBP, NONE,  1, ofs};  // local
        EA m3 = {RBP, RDI,   2, ofs};  // local + index*2

        EA m4 = {RAX, NONE,  1, ofs};
        EA m5 = {R12, NONE,  1, ofs};
        EA m6 = {R13, RAX,   1, ofs};
        EA m7 = {R15, NONE,  1, ofs};

        c_imod_rax_mem (&m1, siz);
        c_imod_rax_mem (&m2, siz);
        c_imod_rax_mem (&m3, siz);
        c_imod_rax_mem (&m4, siz);
        c_imod_rax_mem (&m5, siz);
        c_imod_rax_mem (&m6, siz);
        c_imod_rax_mem (&m7, siz);
      }
    }
  }


  {
    int siz;
    for (siz=4; siz<=address_size; siz<<=1)
    {
      c_udiv_rax_reg (RAX, siz);
      c_udiv_rax_reg (RBX, siz);
      c_udiv_rax_reg (RSP, siz);
      c_udiv_rax_reg (R12, siz);
      c_udiv_rax_reg (R13, siz);
      c_udiv_rax_reg (R15, siz);
    }
  }

  {
    const int OFS[3] = {0, 0x12, 0xFFD2};
    int d, siz;
    for (siz=4; siz<=address_size; siz<<=1)
    {
      for (d=0; d<3; d++)
      {
        int ofs = OFS[d];

        EA m1 = {NONE, NONE, 1, ofs};  // global
        EA m2 = {RBP, NONE,  1, ofs};  // local
        EA m3 = {RBP, RDI,   2, ofs};  // local + index*2

        EA m4 = {RAX, NONE,  1, ofs};
        EA m5 = {R12, NONE,  1, ofs};
        EA m6 = {R13, RAX,   1, ofs};
        EA m7 = {R15, NONE,  1, ofs};

        c_udiv_rax_mem (&m1, siz);
        c_udiv_rax_mem (&m2, siz);
        c_udiv_rax_mem (&m3, siz);
        c_udiv_rax_mem (&m4, siz);
        c_udiv_rax_mem (&m5, siz);
        c_udiv_rax_mem (&m6, siz);
        c_udiv_rax_mem (&m7, siz);
      }
    }
  }


  {
    int siz;
    for (siz=4; siz<=address_size; siz<<=1)
    {
      c_umod_rax_reg (RAX, siz);
      c_umod_rax_reg (RBX, siz);
      c_umod_rax_reg (RSP, siz);
      c_umod_rax_reg (R12, siz);
      c_umod_rax_reg (R13, siz);
      c_umod_rax_reg (R15, siz);
    }
  }

  {
    const int OFS[3] = {0, 0x12, 0xFFD2};
    int d, siz;
    for (siz=4; siz<=address_size; siz<<=1)
    {
      for (d=0; d<3; d++)
      {
        int ofs = OFS[d];

        EA m1 = {NONE, NONE, 1, ofs};  // global
        EA m2 = {RBP, NONE,  1, ofs};  // local
        EA m3 = {RBP, RDI,   2, ofs};  // local + index*2

        EA m4 = {RAX, NONE,  1, ofs};
        EA m5 = {R12, NONE,  1, ofs};
        EA m6 = {R13, RAX,   1, ofs};
        EA m7 = {R15, NONE,  1, ofs};

        c_umod_rax_mem (&m1, siz);
        c_umod_rax_mem (&m2, siz);
        c_umod_rax_mem (&m3, siz);
        c_umod_rax_mem (&m4, siz);
        c_umod_rax_mem (&m5, siz);
        c_umod_rax_mem (&m6, siz);
        c_umod_rax_mem (&m7, siz);
      }
    }
  }


  {
    int siz;
    for (siz=1; siz<=address_size; siz<<=1)
    {
      c_shl_reg_imm (RAX, 1, siz);
      c_shl_reg_imm (RCX, 1, siz);
      c_shl_reg_imm (RBX, 2, siz);
      c_shl_reg_imm (RSP, 3, siz);
      c_shl_reg_imm (R12, 4, siz);
      c_shl_reg_imm (R13, 5, siz);
      c_shl_reg_imm (R15, 6, siz);
    }
  }

  {
    int siz;
    for (siz=1; siz<=address_size; siz<<=1)
    {
      c_shl_reg_CL (RAX, siz);
      c_shl_reg_CL (RCX, siz);
      c_shl_reg_CL (RBX, siz);
      c_shl_reg_CL (RSP, siz);
      c_shl_reg_CL (R12, siz);
      c_shl_reg_CL (R13, siz);
      c_shl_reg_CL (R15, siz);
    }
  }


  {
    int siz;
    for (siz=1; siz<=address_size; siz<<=1)
    {
      int ofs = 3;

      EA m1 = {NONE, NONE, 1, ofs};  // global
      EA m2 = {RBP, NONE,  1, ofs};  // local
      EA m3 = {RBP, RDI,   2, ofs};  // local + index*2

      EA m4 = {RAX, NONE,  1, ofs};
      EA m5 = {R12, NONE,  1, ofs};
      EA m6 = {R13, RAX,   1, ofs};
      EA m7 = {R15, NONE,  1, ofs};

      c_shl_mem_imm (&m1, 1, siz);
      c_shl_mem_imm (&m2, 1, siz);
      c_shl_mem_imm (&m3, 1, siz);
      c_shl_mem_imm (&m4, 1, siz);
      c_shl_mem_imm (&m5, 1, siz);
      c_shl_mem_imm (&m6, 1, siz);
      c_shl_mem_imm (&m7, 1, siz);
    }
  }

  {
    int siz;
    for (siz=1; siz<=address_size; siz<<=1)
    {
      int ofs = 3;

      EA m1 = {NONE, NONE, 1, ofs};  // global
      EA m2 = {RBP, NONE,  1, ofs};  // local
      EA m3 = {RBP, RDI,   2, ofs};  // local + index*2

      EA m4 = {RAX, NONE,  1, ofs};
      EA m5 = {R12, NONE,  1, ofs};
      EA m6 = {R13, RAX,   1, ofs};
      EA m7 = {R15, NONE,  1, ofs};

      c_shl_mem_imm (&m1, 2, siz);
      c_shl_mem_imm (&m2, 3, siz);
      c_shl_mem_imm (&m3, 4, siz);
      c_shl_mem_imm (&m4, 5, siz);
      c_shl_mem_imm (&m5, 6, siz);
      c_shl_mem_imm (&m6, 7, siz);
      c_shl_mem_imm (&m7, 63, siz);
    }
  }

  {
    int siz;
    for (siz=1; siz<=address_size; siz<<=1)
    {
      int ofs = 3;

      EA m1 = {NONE, NONE, 1, ofs};  // global
      EA m2 = {RBP, NONE,  1, ofs};  // local
      EA m3 = {RBP, RDI,   2, ofs};  // local + index*2

      EA m4 = {RAX, NONE,  1, ofs};
      EA m5 = {R12, NONE,  1, ofs};
      EA m6 = {R13, RAX,   1, ofs};
      EA m7 = {R15, NONE,  1, ofs};

      c_shl_mem_CL (&m1, siz);
      c_shl_mem_CL (&m2, siz);
      c_shl_mem_CL (&m3, siz);
      c_shl_mem_CL (&m4, siz);
      c_shl_mem_CL (&m5, siz);
      c_shl_mem_CL (&m6, siz);
      c_shl_mem_CL (&m7, siz);
    }
  }


  {
    int siz;
    for (siz=1; siz<=address_size; siz<<=1)
    {
      c_sar_reg_imm (RAX, 1, siz);
      c_sar_reg_imm (RCX, 1, siz);
      c_sar_reg_imm (RBX, 2, siz);
      c_sar_reg_imm (RSP, 3, siz);
      c_sar_reg_imm (R12, 4, siz);
      c_sar_reg_imm (R13, 5, siz);
      c_sar_reg_imm (R15, 6, siz);
    }
  }

  {
    int siz;
    for (siz=1; siz<=address_size; siz<<=1)
    {
      c_sar_reg_CL (RAX, siz);
      c_sar_reg_CL (RCX, siz);
      c_sar_reg_CL (RBX, siz);
      c_sar_reg_CL (RSP, siz);
      c_sar_reg_CL (R12, siz);
      c_sar_reg_CL (R13, siz);
      c_sar_reg_CL (R15, siz);
    }
  }


  {
    int siz;
    for (siz=1; siz<=address_size; siz<<=1)
    {
      int ofs = 3;

      EA m1 = {NONE, NONE, 1, ofs};  // global
      EA m2 = {RBP, NONE,  1, ofs};  // local
      EA m3 = {RBP, RDI,   2, ofs};  // local + index*2

      EA m4 = {RAX, NONE,  1, ofs};
      EA m5 = {R12, NONE,  1, ofs};
      EA m6 = {R13, RAX,   1, ofs};
      EA m7 = {R15, NONE,  1, ofs};

      c_sar_mem_imm (&m1, 1, siz);
      c_sar_mem_imm (&m2, 1, siz);
      c_sar_mem_imm (&m3, 1, siz);
      c_sar_mem_imm (&m4, 1, siz);
      c_sar_mem_imm (&m5, 1, siz);
      c_sar_mem_imm (&m6, 1, siz);
      c_sar_mem_imm (&m7, 1, siz);
    }
  }

  {
    int siz;
    for (siz=1; siz<=address_size; siz<<=1)
    {
      int ofs = 3;

      EA m1 = {NONE, NONE, 1, ofs};  // global
      EA m2 = {RBP, NONE,  1, ofs};  // local
      EA m3 = {RBP, RDI,   2, ofs};  // local + index*2

      EA m4 = {RAX, NONE,  1, ofs};
      EA m5 = {R12, NONE,  1, ofs};
      EA m6 = {R13, RAX,   1, ofs};
      EA m7 = {R15, NONE,  1, ofs};

      c_sar_mem_imm (&m1, 2, siz);
      c_sar_mem_imm (&m2, 3, siz);
      c_sar_mem_imm (&m3, 4, siz);
      c_sar_mem_imm (&m4, 5, siz);
      c_sar_mem_imm (&m5, 6, siz);
      c_sar_mem_imm (&m6, 7, siz);
      c_sar_mem_imm (&m7, 63, siz);
    }
  }

  {
    int siz;
    for (siz=1; siz<=address_size; siz<<=1)
    {
      int ofs = 3;

      EA m1 = {NONE, NONE, 1, ofs};  // global
      EA m2 = {RBP, NONE,  1, ofs};  // local
      EA m3 = {RBP, RDI,   2, ofs};  // local + index*2

      EA m4 = {RAX, NONE,  1, ofs};
      EA m5 = {R12, NONE,  1, ofs};
      EA m6 = {R13, RAX,   1, ofs};
      EA m7 = {R15, NONE,  1, ofs};

      c_sar_mem_CL (&m1, siz);
      c_sar_mem_CL (&m2, siz);
      c_sar_mem_CL (&m3, siz);
      c_sar_mem_CL (&m4, siz);
      c_sar_mem_CL (&m5, siz);
      c_sar_mem_CL (&m6, siz);
      c_sar_mem_CL (&m7, siz);
    }
  }

  {
    int siz;
    for (siz=1; siz<=address_size; siz<<=1)
    {
      c_shr_reg_imm (RAX, 1, siz);
      c_shr_reg_imm (RCX, 1, siz);
      c_shr_reg_imm (RBX, 2, siz);
      c_shr_reg_imm (RSP, 3, siz);
      c_shr_reg_imm (R12, 4, siz);
      c_shr_reg_imm (R13, 5, siz);
      c_shr_reg_imm (R15, 6, siz);
    }
  }

  {
    int siz;
    for (siz=1; siz<=address_size; siz<<=1)
    {
      c_shr_reg_CL (RAX, siz);
      c_shr_reg_CL (RCX, siz);
      c_shr_reg_CL (RBX, siz);
      c_shr_reg_CL (RSP, siz);
      c_shr_reg_CL (R12, siz);
      c_shr_reg_CL (R13, siz);
      c_shr_reg_CL (R15, siz);
    }
  }


  {
    int siz;
    for (siz=1; siz<=address_size; siz<<=1)
    {
      int ofs = 3;

      EA m1 = {NONE, NONE, 1, ofs};  // global
      EA m2 = {RBP, NONE,  1, ofs};  // local
      EA m3 = {RBP, RDI,   2, ofs};  // local + index*2

      EA m4 = {RAX, NONE,  1, ofs};
      EA m5 = {R12, NONE,  1, ofs};
      EA m6 = {R13, RAX,   1, ofs};
      EA m7 = {R15, NONE,  1, ofs};

      c_shr_mem_imm (&m1, 1, siz);
      c_shr_mem_imm (&m2, 1, siz);
      c_shr_mem_imm (&m3, 1, siz);
      c_shr_mem_imm (&m4, 1, siz);
      c_shr_mem_imm (&m5, 1, siz);
      c_shr_mem_imm (&m6, 1, siz);
      c_shr_mem_imm (&m7, 1, siz);
    }
  }

  {
    int siz;
    for (siz=1; siz<=address_size; siz<<=1)
    {
      int ofs = 3;

      EA m1 = {NONE, NONE, 1, ofs};  // global
      EA m2 = {RBP, NONE,  1, ofs};  // local
      EA m3 = {RBP, RDI,   2, ofs};  // local + index*2

      EA m4 = {RAX, NONE,  1, ofs};
      EA m5 = {R12, NONE,  1, ofs};
      EA m6 = {R13, RAX,   1, ofs};
      EA m7 = {R15, NONE,  1, ofs};

      c_shr_mem_imm (&m1, 2, siz);
      c_shr_mem_imm (&m2, 3, siz);
      c_shr_mem_imm (&m3, 4, siz);
      c_shr_mem_imm (&m4, 5, siz);
      c_shr_mem_imm (&m5, 6, siz);
      c_shr_mem_imm (&m6, 7, siz);
      c_shr_mem_imm (&m7, 63, siz);
    }
  }

  {
    int siz;
    for (siz=1; siz<=address_size; siz<<=1)
    {
      int ofs = 3;

      EA m1 = {NONE, NONE, 1, ofs};  // global
      EA m2 = {RBP, NONE,  1, ofs};  // local
      EA m3 = {RBP, RDI,   2, ofs};  // local + index*2

      EA m4 = {RAX, NONE,  1, ofs};
      EA m5 = {R12, NONE,  1, ofs};
      EA m6 = {R13, RAX,   1, ofs};
      EA m7 = {R15, NONE,  1, ofs};

      c_shr_mem_CL (&m1, siz);
      c_shr_mem_CL (&m2, siz);
      c_shr_mem_CL (&m3, siz);
      c_shr_mem_CL (&m4, siz);
      c_shr_mem_CL (&m5, siz);
      c_shr_mem_CL (&m6, siz);
      c_shr_mem_CL (&m7, siz);
    }
  }

  {
    int siz;
    for (siz=2; siz<=address_size; siz<<=1)
    {
      c_shld_reg_reg_CL (RAX, RAX, siz);
      c_shld_reg_reg_CL (RAX, RBX, siz);
      c_shld_reg_reg_CL (RAX, RSP, siz);
      c_shld_reg_reg_CL (RAX, R12, siz);
      c_shld_reg_reg_CL (RAX, R13, siz);
      c_shld_reg_reg_CL (RAX, R15, siz);

      c_shld_reg_reg_CL (RAX, R8, siz);
      c_shld_reg_reg_CL (RBX, R8, siz);
      c_shld_reg_reg_CL (RSP, R8, siz);
      c_shld_reg_reg_CL (R12, R8, siz);
      c_shld_reg_reg_CL (R13, R8, siz);
      c_shld_reg_reg_CL (R15, R8, siz);
    }
  }


  {
    int siz;
    for (siz=2; siz<=address_size; siz<<=1)
    {
      c_shld_reg_reg_imm (RAX, RAX, 3, siz);
      c_shld_reg_reg_imm (RAX, RBX, 3, siz);
      c_shld_reg_reg_imm (RAX, RSP, 3, siz);
      c_shld_reg_reg_imm (RAX, R12, 3, siz);
      c_shld_reg_reg_imm (RAX, R13, 3, siz);
      c_shld_reg_reg_imm (RAX, R15, 3, siz);

      c_shld_reg_reg_imm (RAX, R8, 3, siz);
      c_shld_reg_reg_imm (RBX, R8, 3, siz);
      c_shld_reg_reg_imm (RSP, R8, 3, siz);
      c_shld_reg_reg_imm (R12, R8, 3, siz);
      c_shld_reg_reg_imm (R13, R8, 3, siz);
      c_shld_reg_reg_imm (R15, R8, 3, siz);
    }
  }


  {
    int siz;
    for (siz=2; siz<=address_size; siz<<=1)
    {
      int ofs = 3;

      EA m1 = {NONE, NONE, 1, ofs};  // global
      EA m2 = {RBP, NONE,  1, ofs};  // local
      EA m3 = {RBP, RDI,   2, ofs};  // local + index*2

      EA m4 = {RAX, NONE,  1, ofs};
      EA m5 = {R12, NONE,  1, ofs};
      EA m6 = {R13, RAX,   1, ofs};
      EA m7 = {R15, NONE,  1, ofs};

      c_shld_mem_reg_CL (&m1, RAX, siz);
      c_shld_mem_reg_CL (&m2, RBX, siz);
      c_shld_mem_reg_CL (&m3, RCX, siz);
      c_shld_mem_reg_CL (&m4, RDX, siz);
      c_shld_mem_reg_CL (&m5, RSP, siz);
      c_shld_mem_reg_CL (&m6, R12, siz);
      c_shld_mem_reg_CL (&m7, R14, siz);
    }
  }

  {
    int siz;
    for (siz=2; siz<=address_size; siz<<=1)
    {
      int ofs = 3;

      EA m1 = {NONE, NONE, 1, ofs};  // global
      EA m2 = {RBP, NONE,  1, ofs};  // local
      EA m3 = {RBP, RDI,   2, ofs};  // local + index*2

      EA m4 = {RAX, NONE,  1, ofs};
      EA m5 = {R12, NONE,  1, ofs};
      EA m6 = {R13, RAX,   1, ofs};
      EA m7 = {R15, NONE,  1, ofs};

      c_shld_mem_reg_imm (&m1, RAX, 9, siz);
      c_shld_mem_reg_imm (&m2, RBX, 9, siz);
      c_shld_mem_reg_imm (&m3, RCX, 9, siz);
      c_shld_mem_reg_imm (&m4, RDX, 9, siz);
      c_shld_mem_reg_imm (&m5, RSP, 9, siz);
      c_shld_mem_reg_imm (&m6, R12, 9, siz);
      c_shld_mem_reg_imm (&m7, R14, 9, siz);
    }
  }



  {
    int siz;
    for (siz=2; siz<=address_size; siz<<=1)
    {
      c_shrd_reg_reg_CL (RAX, RAX, siz);
      c_shrd_reg_reg_CL (RAX, RBX, siz);
      c_shrd_reg_reg_CL (RAX, RSP, siz);
      c_shrd_reg_reg_CL (RAX, R12, siz);
      c_shrd_reg_reg_CL (RAX, R13, siz);
      c_shrd_reg_reg_CL (RAX, R15, siz);

      c_shrd_reg_reg_CL (RAX, R8, siz);
      c_shrd_reg_reg_CL (RBX, R8, siz);
      c_shrd_reg_reg_CL (RSP, R8, siz);
      c_shrd_reg_reg_CL (R12, R8, siz);
      c_shrd_reg_reg_CL (R13, R8, siz);
      c_shrd_reg_reg_CL (R15, R8, siz);
    }
  }


  {
    int siz;
    for (siz=2; siz<=address_size; siz<<=1)
    {
      c_shrd_reg_reg_imm (RAX, RAX, 3, siz);
      c_shrd_reg_reg_imm (RAX, RBX, 3, siz);
      c_shrd_reg_reg_imm (RAX, RSP, 3, siz);
      c_shrd_reg_reg_imm (RAX, R12, 3, siz);
      c_shrd_reg_reg_imm (RAX, R13, 3, siz);
      c_shrd_reg_reg_imm (RAX, R15, 3, siz);

      c_shrd_reg_reg_imm (RAX, R8, 3, siz);
      c_shrd_reg_reg_imm (RBX, R8, 3, siz);
      c_shrd_reg_reg_imm (RSP, R8, 3, siz);
      c_shrd_reg_reg_imm (R12, R8, 3, siz);
      c_shrd_reg_reg_imm (R13, R8, 3, siz);
      c_shrd_reg_reg_imm (R15, R8, 3, siz);
    }
  }


  {
    int siz;
    for (siz=2; siz<=address_size; siz<<=1)
    {
      int ofs = 3;

      EA m1 = {NONE, NONE, 1, ofs};  // global
      EA m2 = {RBP, NONE,  1, ofs};  // local
      EA m3 = {RBP, RDI,   2, ofs};  // local + index*2

      EA m4 = {RAX, NONE,  1, ofs};
      EA m5 = {R12, NONE,  1, ofs};
      EA m6 = {R13, RAX,   1, ofs};
      EA m7 = {R15, NONE,  1, ofs};

      c_shrd_mem_reg_CL (&m1, RAX, siz);
      c_shrd_mem_reg_CL (&m2, RBX, siz);
      c_shrd_mem_reg_CL (&m3, RCX, siz);
      c_shrd_mem_reg_CL (&m4, RDX, siz);
      c_shrd_mem_reg_CL (&m5, RSP, siz);
      c_shrd_mem_reg_CL (&m6, R12, siz);
      c_shrd_mem_reg_CL (&m7, R14, siz);
    }
  }

  {
    int siz;
    for (siz=2; siz<=address_size; siz<<=1)
    {
      int ofs = 3;

      EA m1 = {NONE, NONE, 1, ofs};  // global
      EA m2 = {RBP, NONE,  1, ofs};  // local
      EA m3 = {RBP, RDI,   2, ofs};  // local + index*2

      EA m4 = {RAX, NONE,  1, ofs};
      EA m5 = {R12, NONE,  1, ofs};
      EA m6 = {R13, RAX,   1, ofs};
      EA m7 = {R15, NONE,  1, ofs};

      c_shrd_mem_reg_imm (&m1, RAX, 9, siz);
      c_shrd_mem_reg_imm (&m2, RBX, 9, siz);
      c_shrd_mem_reg_imm (&m3, RCX, 9, siz);
      c_shrd_mem_reg_imm (&m4, RDX, 9, siz);
      c_shrd_mem_reg_imm (&m5, RSP, 9, siz);
      c_shrd_mem_reg_imm (&m6, R12, 9, siz);
      c_shrd_mem_reg_imm (&m7, R14, 9, siz);
    }
  }

  {
    int siz;
    for (siz=1; siz<=address_size; siz<<=1)
    {
      c_rcl_reg (RAX, siz);
      c_rcl_reg (RCX, siz);
      c_rcl_reg (RBX, siz);
      c_rcl_reg (RSP, siz);
      c_rcl_reg (R12, siz);
      c_rcl_reg (R13, siz);
      c_rcl_reg (R15, siz);
    }
  }


  {
    int siz;
    for (siz=1; siz<=address_size; siz<<=1)
    {
      int ofs = 3;

      EA m1 = {NONE, NONE, 1, ofs};  // global
      EA m2 = {RBP, NONE,  1, ofs};  // local
      EA m3 = {RBP, RDI,   2, ofs};  // local + index*2

      EA m4 = {RAX, NONE,  1, ofs};
      EA m5 = {R12, NONE,  1, ofs};
      EA m6 = {R13, RAX,   1, ofs};
      EA m7 = {R15, NONE,  1, ofs};

      c_rcl_mem (&m1, siz);
      c_rcl_mem (&m2, siz);
      c_rcl_mem (&m3, siz);
      c_rcl_mem (&m4, siz);
      c_rcl_mem (&m5, siz);
      c_rcl_mem (&m6, siz);
      c_rcl_mem (&m7, siz);
    }
  }

  {
    EA ea = {RBP, NONE, 1, 3};  // local
    c_lock_prefix ();
    c_add_mem_reg (&ea, RAX, 4);
  }

  c_pause ();

  c_cmc ();

  if (address_size == 4)
  {
    c_lahf ();
    c_sahf ();
    c_mov_al_ah ();
    c_and_ah_al ();
  }

  {
    EA ea = {RBP, NONE, 1, 3};  // local

    c_fld (&ea, 4);
    c_fld (&ea, 8);

    c_fld_zero ();
    c_fld_one ();

    c_fld_imem (&ea, 2);
    c_fld_imem (&ea, 4);
    c_fld_imem (&ea, 8);

    c_fstp (&ea, 4);
    c_fstp (&ea, 8);

    c_fistp (&ea, 2);
    c_fistp (&ea, 4);
    c_fistp (&ea, 8);

    c_fisttp (&ea, 2);
    c_fisttp (&ea, 4);
    c_fisttp (&ea, 8);

    c_fldcw (&ea);
    c_fstcw (&ea);

    c_push_fltp (4);
    c_push_fltp (8);

    c_fadd_fmem (&ea, 4);
    c_fadd_fmem (&ea, 8);

    c_fadd_imem (&ea, 2);
    c_fadd_imem (&ea, 4);

    c_faddp ();

    c_fsub_fmem (&ea, 4);
    c_fsub_fmem (&ea, 8);

    c_fsub_imem (&ea, 2);
    c_fsub_imem (&ea, 4);

    c_fsubp ();

    c_fsubr_fmem (&ea, 4);
    c_fsubr_fmem (&ea, 8);

    c_fsubr_imem (&ea, 2);
    c_fsubr_imem (&ea, 4);

    c_fsubrp ();

    c_fmul_fmem (&ea, 4);
    c_fmul_fmem (&ea, 8);

    c_fmul_imem (&ea, 2);
    c_fmul_imem (&ea, 4);

    c_fmulp ();

    c_fdiv_fmem (&ea, 4);
    c_fdiv_fmem (&ea, 8);

    c_fdiv_imem (&ea, 2);
    c_fdiv_imem (&ea, 4);

    c_fdivp ();

    c_fdivr_fmem (&ea, 4);
    c_fdivr_fmem (&ea, 8);

    c_fdivr_imem (&ea, 2);
    c_fdivr_imem (&ea, 4);

    c_fdivrp ();

    c_fneg ();

    if (address_size == 4)
    {
      c_fcomp_fmem (&ea, 4);
      c_fcomp_fmem (&ea, 8);
    }

    c_fcomp ();

    c_fpop ();

    ea.base = NONE;  ea.index = NONE;   ea.scale = 1;  ea.reloc.kind = RELOC_NONE; ea.reloc.nr = 0;
    c_mov_reg_mem (RAX, &ea, 4);
    c_mov_mem_reg (&ea, RAX, 4);

    c_fisttp (&ea, 2);
    c_fisttp (&ea, 4);
    c_fisttp (&ea, 8);
  }
}



void generate_test_assembler_code_xmm ()
{

  if (false)
  {
    c_mov_xm_xm  (X2,  X5,  4);
    c_mov_xm_xm  (X2,  X15, 4);
    c_mov_xm_xm  (X12, X3,  4);

    c_mov_xm_xm  (X2, X5,   8);
    c_mov_xm_xm  (X2, X15,  8);
    c_mov_xm_xm  (X12, X3,  8);
  }

  if (false)
  {
    {
      EA ea = {R11, NONE,  1, 15};

      c_mov_xm_mem (X5, &ea, 4);
      c_mov_xm_mem (X12, &ea, 4);

      c_mov_xm_mem (X5, &ea, 8);
      c_mov_xm_mem (X12, &ea, 8);
    }

    {
      EA ea = {RCX, NONE,  1, 15};

      c_mov_xm_mem (X5, &ea, 4);
      c_mov_xm_mem (X12, &ea, 4);

      c_mov_xm_mem (X5, &ea, 8);
      c_mov_xm_mem (X12, &ea, 8);
    }
  }

  if (false)
  {
    {
      EA ea = {R11, NONE,  1, 15};

      c_mov_mem_xm (&ea, X5, 4);
      c_mov_mem_xm (&ea, X12, 4);

      c_mov_mem_xm (&ea, X5, 8);
      c_mov_mem_xm (&ea, X12, 8);
    }

    {
      EA ea = {RCX, NONE,  1, 15};

      c_mov_mem_xm (&ea, X5, 4);
      c_mov_mem_xm (&ea, X12, 4);

      c_mov_mem_xm (&ea, X5, 8);
      c_mov_mem_xm (&ea, X12, 8);
    }
  }

  if (false)
  {
    c_mov_xm_reg (X2, RCX, 4);
    c_mov_xm_reg (X2, RCX, 8);

    c_mov_xm_reg (X12, RCX, 4);
    c_mov_xm_reg (X12, RCX, 8);

    c_mov_xm_reg (X2, R9, 4);
    c_mov_xm_reg (X2, R9, 8);

    c_mov_xm_reg (X12, R10, 4);
    c_mov_xm_reg (X12, R10, 8);
  }

  if (false)
  {
    c_mov_reg_xm (RCX, X2, 4);
    c_mov_reg_xm (RCX, X2, 8);

    c_mov_reg_xm (RCX, X12, 4);
    c_mov_reg_xm (RCX, X12, 8);

    c_mov_reg_xm (R9, X2, 4);
    c_mov_reg_xm (R9, X2, 8);

    c_mov_reg_xm (R10, X12, 4);
    c_mov_reg_xm (R10, X12, 8);
  }

  if (false)
  {
    c_cnv_xm_xm (X3, X2, 4, 8);
    c_cnv_xm_xm (X3, X2, 8, 4);

    c_cnv_xm_xm (X3, X12, 4, 8);
    c_cnv_xm_xm (X3, X12, 8, 4);

    c_cnv_xm_xm (X13, X2, 4, 8);
    c_cnv_xm_xm (X13, X2, 8, 4);

    c_cnv_xm_xm (X13, X12, 4, 8);
    c_cnv_xm_xm (X13, X12, 8, 4);
  }

  if (false)
  {
    c_load_xm_size_ireg_size (X1, RCX, 8, 4);   // int4/8 to float/double
    c_load_xm_size_ireg_size (X1, RCX, 4, 8);   // int4/8 to float/double

    c_load_xm_size_ireg_size (X12, RCX, 8, 4);   // int4/8 to float/double
    c_load_xm_size_ireg_size (X12, RCX, 4, 8);   // int4/8 to float/double

    c_load_xm_size_ireg_size (X1, R13, 8, 4);   // int4/8 to float/double
    c_load_xm_size_ireg_size (X1, R13, 4, 8);   // int4/8 to float/double

    c_load_xm_size_ireg_size (X12, R13, 8, 4);   // int4/8 to float/double
    c_load_xm_size_ireg_size (X12, R13, 4, 8);   // int4/8 to float/double

/*
0000000000401004 | F2:0F2AC9                         | cvtsi2sd xmm1,ecx                         |
0000000000401008 | F348:0F2AC9                       | cvtsi2ss xmm1,rcx                         |
000000000040100D | F244:0F2AE1                       | cvtsi2sd xmm12,ecx                        |
0000000000401012 | F34C:0F2AE1                       | cvtsi2ss xmm12,rcx                        |
0000000000401017 | F241:0F2ACD                       | cvtsi2sd xmm1,r13d                        |
000000000040101C | F349:0F2ACD                       | cvtsi2ss xmm1,r13                         |
0000000000401021 | F245:0F2AE5                       | cvtsi2sd xmm12,r13d                       |
0000000000401026 | F34D:0F2AE5                       | cvtsi2ss xmm12,r13                        |
*/
  }


  if (false)
  {
    c_load_xm_size_ireg_size (X1, RCX, 4, 4);   // int4/8 to float/double
    c_load_xm_size_ireg_size (X12, RCX, 4, 4);   // int4/8 to float/double
    c_load_xm_size_ireg_size (X1, R13, 4, 4);   // int4/8 to float/double
    c_load_xm_size_ireg_size (X12, R13, 4, 4);   // int4/8 to float/double
  }
  
/*  
000000000040102B | F3:0F2AC9                         | cvtsi2ss xmm1,ecx                         |
0000000000401033 | F344:0F2AE1                       | cvtsi2ss xmm12,ecx                        |
000000000040103D | F341:0F2ACD                       | cvtsi2ss xmm1,r13d                        |
0000000000401047 | F345:0F2AE5                       | cvtsi2ss xmm12,r13d                       |
*/    


  if (false)
  {
    c_load_xm_size_ireg_size (X1, RCX, 8, 8);   // int4/8 to float/double
    c_load_xm_size_ireg_size (X12, RCX, 8, 8);   // int4/8 to float/double
    c_load_xm_size_ireg_size (X1, R13, 8, 8);   // int4/8 to float/double
    c_load_xm_size_ireg_size (X12, R13, 8, 8);   // int4/8 to float/double
  }

/*
0000000000401051 | F248:0F2AC9                       | cvtsi2sd xmm1,rcx                         |
000000000040105B | F24C:0F2AE1                       | cvtsi2sd xmm12,rcx                        |
0000000000401065 | F249:0F2ACD                       | cvtsi2sd xmm1,r13                         |
000000000040106F | F24D:0F2AE5                       | cvtsi2sd xmm12,r13                        |
*/


  if (false)
  {
    int dst, src;
    for (dst=4; dst<=8; dst+=4)
    {
      for (src=4; src<=8; src+=4)
      {
        {
          EA ea = {RCX, NONE,  1, 15};

          c_load_xm_size_imem_size (X5, &ea, dst, src);
          c_load_xm_size_imem_size (X12, &ea, dst, src);
        }

        {
          EA ea = {R11, NONE,  1, 15};

          c_load_xm_size_imem_size (X5, &ea, dst, src);
          c_load_xm_size_imem_size (X12, &ea, dst, src);
        }
      }
    }
  }

/*
0000000000401004 | F3:0F2A69 0F                      | cvtsi2ss xmm5,dword ptr ds:[rcx+F]        |
0000000000401009 | F344:0F2A61 0F                    | cvtsi2ss xmm12,dword ptr ds:[rcx+F]       |
000000000040100F | F341:0F2A6B 0F                    | cvtsi2ss xmm5,dword ptr ds:[r11+F]        |
0000000000401015 | F345:0F2A63 0F                    | cvtsi2ss xmm12,dword ptr ds:[r11+F]       |
000000000040101B | F348:0F2A69 0F                    | cvtsi2ss xmm5,qword ptr ds:[rcx+F]        |
0000000000401021 | F34C:0F2A61 0F                    | cvtsi2ss xmm12,qword ptr ds:[rcx+F]       |
0000000000401027 | F349:0F2A6B 0F                    | cvtsi2ss xmm5,qword ptr ds:[r11+F]        |
000000000040102D | F34D:0F2A63 0F                    | cvtsi2ss xmm12,qword ptr ds:[r11+F]       |
0000000000401033 | F2:0F2A69 0F                      | cvtsi2sd xmm5,dword ptr ds:[rcx+F]        |
0000000000401038 | F244:0F2A61 0F                    | cvtsi2sd xmm12,dword ptr ds:[rcx+F]       |
000000000040103E | F241:0F2A6B 0F                    | cvtsi2sd xmm5,dword ptr ds:[r11+F]        |
0000000000401044 | F245:0F2A63 0F                    | cvtsi2sd xmm12,dword ptr ds:[r11+F]       |
000000000040104A | F248:0F2A69 0F                    | cvtsi2sd xmm5,qword ptr ds:[rcx+F]        |
0000000000401050 | F24C:0F2A61 0F                    | cvtsi2sd xmm12,qword ptr ds:[rcx+F]       |
0000000000401056 | F249:0F2A6B 0F                    | cvtsi2sd xmm5,qword ptr ds:[r11+F]        |
000000000040105C | F24D:0F2A63 0F                    | cvtsi2sd xmm12,qword ptr ds:[r11+F]       |
*/


  if (false)
  {
    c_store_ireg_size_xm_size (RCX, X1,  4, 4);   // int4/8 to float/double
    c_store_ireg_size_xm_size (RCX, X12, 4, 4);   // int4/8 to float/double
    c_store_ireg_size_xm_size (R13, X1,  4, 4);   // int4/8 to float/double
    c_store_ireg_size_xm_size (R13, X12, 4, 4);   // int4/8 to float/double

    c_store_ireg_size_xm_size (RCX, X1,  8, 4);   // int4/8 to float/double
    c_store_ireg_size_xm_size (RCX, X12, 8, 4);   // int4/8 to float/double
    c_store_ireg_size_xm_size (R13, X1,  8, 4);   // int4/8 to float/double
    c_store_ireg_size_xm_size (R13, X12, 8, 4);   // int4/8 to float/double

    c_store_ireg_size_xm_size (RCX, X1,  4, 8);   // int4/8 to float/double
    c_store_ireg_size_xm_size (RCX, X12, 4, 8);   // int4/8 to float/double
    c_store_ireg_size_xm_size (R13, X1,  4, 8);   // int4/8 to float/double
    c_store_ireg_size_xm_size (R13, X12, 4, 8);   // int4/8 to float/double

    c_store_ireg_size_xm_size (RCX, X1,  8, 8);   // int4/8 to float/double
    c_store_ireg_size_xm_size (RCX, X12, 8, 8);   // int4/8 to float/double
    c_store_ireg_size_xm_size (R13, X1,  8, 8);   // int4/8 to float/double
    c_store_ireg_size_xm_size (R13, X12, 8, 8);   // int4/8 to float/double
  }

/*
0000000000401004 | F3:0F2CC9                         | cvttss2si ecx,xmm1                        |
0000000000401008 | F341:0F2CCC                       | cvttss2si ecx,xmm12                       |
000000000040100D | F344:0F2CE9                       | cvttss2si r13d,xmm1                       |
0000000000401012 | F345:0F2CEC                       | cvttss2si r13d,xmm12                      |
0000000000401017 | F348:0F2CC9                       | cvttss2si rcx,xmm1                        |
000000000040101C | F349:0F2CCC                       | cvttss2si rcx,xmm12                       |
0000000000401021 | F34C:0F2CE9                       | cvttss2si r13,xmm1                        |
0000000000401026 | F34D:0F2CEC                       | cvttss2si r13,xmm12                       |
000000000040102B | F2:0F2CC9                         | cvttsd2si ecx,xmm1                        |
000000000040102F | F241:0F2CCC                       | cvttsd2si ecx,xmm12                       |
0000000000401034 | F244:0F2CE9                       | cvttsd2si r13d,xmm1                       |
0000000000401039 | F245:0F2CEC                       | cvttsd2si r13d,xmm12                      |
000000000040103E | F248:0F2CC9                       | cvttsd2si rcx,xmm1                        |
0000000000401043 | F249:0F2CCC                       | cvttsd2si rcx,xmm12                       |
0000000000401048 | F24C:0F2CE9                       | cvttsd2si r13,xmm1                        |
000000000040104D | F24D:0F2CEC                       | cvttsd2si r13,xmm12                       |
*/

  if (false)
  {
    int dst, src;
    for (dst=4; dst<=8; dst+=4)
    {
      for (src=4; src<=8; src+=4)
      {
        {
          EA ea = {RCX, NONE,  1, 15};
          c_store_ireg_size_memf_size (RCX, &ea, dst, src);
          c_store_ireg_size_memf_size (R12, &ea, dst, src);
        }

        {
          EA ea = {R11, NONE,  1, 15};
          c_store_ireg_size_memf_size (RCX, &ea, dst, src);
          c_store_ireg_size_memf_size (R12, &ea, dst, src);
        }
      }
    }
  }

/*
0000000000401052 | F3:0F2C49 0F                      | cvttss2si ecx,dword ptr ds:[rcx+F]        |
0000000000401057 | F344:0F2C61 0F                    | cvttss2si r12d,dword ptr ds:[rcx+F]       |
000000000040105D | F341:0F2C4B 0F                    | cvttss2si ecx,dword ptr ds:[r11+F]        |
0000000000401063 | F345:0F2C63 0F                    | cvttss2si r12d,dword ptr ds:[r11+F]       |

0000000000401069 | F2:0F2C49 0F                      | cvttsd2si ecx,qword ptr ds:[rcx+F]        |
000000000040106E | F244:0F2C61 0F                    | cvttsd2si r12d,qword ptr ds:[rcx+F]       |
0000000000401074 | F241:0F2C4B 0F                    | cvttsd2si ecx,qword ptr ds:[r11+F]        |
000000000040107A | F245:0F2C63 0F                    | cvttsd2si r12d,qword ptr ds:[r11+F]       |

0000000000401080 | F348:0F2C49 0F                    | cvttss2si rcx,dword ptr ds:[rcx+F]        |
0000000000401086 | F34C:0F2C61 0F                    | cvttss2si r12,dword ptr ds:[rcx+F]        |
000000000040108C | F349:0F2C4B 0F                    | cvttss2si rcx,dword ptr ds:[r11+F]        |
0000000000401092 | F34D:0F2C63 0F                    | cvttss2si r12,dword ptr ds:[r11+F]        |

0000000000401098 | F248:0F2C49 0F                    | cvttsd2si rcx,qword ptr ds:[rcx+F]        |
000000000040109E | F24C:0F2C61 0F                    | cvttsd2si r12,qword ptr ds:[rcx+F]        |
00000000004010A4 | F249:0F2C4B 0F                    | cvttsd2si rcx,qword ptr ds:[r11+F]        |
00000000004010AA | F24D:0F2C63 0F                    | cvttsd2si r12,qword ptr ds:[r11+F]        |
*/



  if (false)
  {
    c_add_xm_xm (X1,  X1,  4);
    c_sub_xm_xm (X5,  X12, 4);
    c_mul_xm_xm (R12, X5,  4);
    c_div_xm_xm (R13, X12, 4);
    c_sqr_xm_xm (R13, X15, 4);
    
    c_add_xm_xm (X1,  X1,  8);
    c_sub_xm_xm (X5,  X12, 8);
    c_mul_xm_xm (R12, X5,  8);
    c_div_xm_xm (R13, X12, 8);
    c_sqr_xm_xm (R13, X15, 8);
  }
  
/*  
0000000000401004 | F3:0F58C9                         | addss xmm1,xmm1                           |
0000000000401008 | F341:0F5CEC                       | subss xmm5,xmm12                          |
000000000040100D | F344:0F59E5                       | mulss xmm12,xmm5                          |
0000000000401012 | F345:0F5EEC                       | divss xmm13,xmm12                         |
0000000000401017 | F345:0F51EF                       | sqrtss xmm13,xmm15                        |
000000000040101C | F2:0F58C9                         | addsd xmm1,xmm1                           |
0000000000401020 | F241:0F5CEC                       | subsd xmm5,xmm12                          |
0000000000401025 | F244:0F59E5                       | mulsd xmm12,xmm5                          |
000000000040102A | F245:0F5EEC                       | divsd xmm13,xmm12                         |
000000000040102F | F245:0F51EF                       | sqrtsd xmm13,xmm15                        |
*/
  
  if (false)
  {
    int size;
    for (size=4; size<=8; size+=4)
    {
      EA ea = {RCX, NONE,  1, 15};
      
      c_add_xm_mem (X2, &ea, size);
      c_sub_xm_mem (X5, &ea, size);
      c_mul_xm_mem (X12, &ea, size);
      c_div_xm_mem (X13, &ea, size);
      c_sqr_xm_mem (X14, &ea, size);
    }
  }

/*
0000000000401034 | F3:0F5851 0F                      | addss xmm2,dword ptr ds:[rcx+F]           |
0000000000401039 | F3:0F5C51 0F                      | subss xmm2,dword ptr ds:[rcx+F]           |
000000000040103E | F3:0F5951 0F                      | mulss xmm2,dword ptr ds:[rcx+F]           |
0000000000401043 | F3:0F5E51 0F                      | divss xmm2,dword ptr ds:[rcx+F]           |
0000000000401048 | F3:0F5151 0F                      | sqrtss xmm2,dword ptr ds:[rcx+F]          |
000000000040104D | F2:0F5851 0F                      | addsd xmm2,qword ptr ds:[rcx+F]           |
0000000000401052 | F2:0F5C51 0F                      | subsd xmm2,qword ptr ds:[rcx+F]           |
0000000000401057 | F2:0F5951 0F                      | mulsd xmm2,qword ptr ds:[rcx+F]           |
000000000040105C | F2:0F5E51 0F                      | divsd xmm2,qword ptr ds:[rcx+F]           |
0000000000401061 | F2:0F5151 0F                      | sqrtsd xmm2,qword ptr ds:[rcx+F]          |
*/



  if (false)
  {
    int size;
    for (size=4; size<=8; size+=4)
    {
      c_cmp_xm_xm (X1,  X1,  size);
      c_cmp_xm_xm (X5,  X12, size);
      c_cmp_xm_xm (R12, X5,  size);
      c_cmp_xm_xm (R13, X12, size);
    }
  }

/*
0000000000401004 | 0F2EC9                            | ucomiss xmm1,xmm1                         |
0000000000401007 | 41:0F2EEC                         | ucomiss xmm5,xmm12                        |
000000000040100B | 44:0F2EE5                         | ucomiss xmm12,xmm5                        |
000000000040100F | 45:0F2EEC                         | ucomiss xmm13,xmm12                       |
0000000000401013 | 66:0F2EC9                         | ucomisd xmm1,xmm1                         |
0000000000401017 | 6641:0F2EEC                       | ucomisd xmm5,xmm12                        |
000000000040101C | 6644:0F2EE5                       | ucomisd xmm12,xmm5                        |
0000000000401021 | 6645:0F2EEC                       | ucomisd xmm13,xmm12                       |
*/


  if (false)
  {
    int size;
    for (size=4; size<=8; size+=4)
    {
      EA ea = {RCX, NONE,  1, 15};
      c_cmp_xm_mem (X5, &ea, size);
      c_cmp_xm_mem (X12, &ea, size);
    }
  }


/*
0000000000401026 | 0F2E69 0F                         | ucomiss xmm5,dword ptr ds:[rcx+F]         |
000000000040102A | 44:0F2E61 0F                      | ucomiss xmm12,dword ptr ds:[rcx+F]        |
000000000040102F | 66:0F2E69 0F                      | ucomisd xmm5,qword ptr ds:[rcx+F]         |
0000000000401034 | 6644:0F2E61 0F                    | ucomisd xmm12,qword ptr ds:[rcx+F]        |
*/


}
