
// ASLAB.C

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "tracing.h"
#include "bintree.h"

#include "types.h"
#include "error.h"

#include "aslab.h"

/***********************************************************************************/

// tree that stores labels for jumping under the function's code and raise int5 exceptions

static BINARY_TREE ll_tree;

/***********************************************************************************/

static int ll_compare (void                  *user,
                       const LINE_LABEL_DATA *data1,
                       const unsigned int    size1,
                       const LINE_LABEL_DATA *data2,
                       const unsigned int    size2)
{
  if (data1->line < data2->line)
    return -1;
  if (data1->line > data2->line)
    return +1;
  if (data1->occurence < data2->occurence)
    return -1;
  if (data1->occurence > data2->occurence)
    return +1;
  return 0;
}

/***********************************************************************************/

// returns label_nr (either new of from existing node for that line)

int4 store_ll (int line, int4 label_nr)
{
  LINE_LABEL_DATA u;
  int             rc, dummy;

  u.line      = line;
  u.occurence = 0;
  u.label_nr  = label_nr;

  rc = insert_btree (&ll_tree, &u, sizeof(u));
  if (rc == 0)
    return label_nr;

  if (rc != BT_DUPLICATE_KEY)
    fatal_compiler_error0 ("store_ll(1)");

  u.line = line;

  rc = retrieve_btree (&ll_tree, &u, sizeof(u), sizeof(u), &dummy, BT_EQUAL);
  if (rc < 0)
    fatal_compiler_error0 ("store_ll(2)");

  return u.label_nr;
}

/***********************************************************************************/

// stores always the label, with an occurence >= 1 (for P_ASSERT)

void store_ll_forced (int line, int4 label_nr)
{
  LINE_LABEL_DATA u;
  int             rc, i;

  for (i=1; ; i++)
  {
    u.line      = line;
    u.occurence = i;
    u.label_nr  = label_nr;

    rc = insert_btree (&ll_tree, &u, sizeof(u));
    if (rc == 0)
      return;

    if (rc != BT_DUPLICATE_KEY)
      fatal_compiler_error0 ("store_ll_forced(1)");
  }
}

/***********************************************************************************/

void create_ll_tree ()
{
  create_btree (&ll_tree, NULL, ll_compare);
}

/***********************************************************************************/

void traverse_ll_tree (LL_FUNC func)
{
  traverse_btree (&ll_tree, func, +1);
}

/***********************************************************************************/

void close_ll_tree ()
{
  close_btree (&ll_tree);
}

/***********************************************************************************/
