
// bug.c : get crash location

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <io.h>
#include <malloc.h>
#include <fcntl.h>
#include "types.h"

int main (int argc, char *argv[])
{
  unsigned long ip, size, current_unit, unit, line, *l, len;
  unsigned char *p;
  bool          found;
  int           fd;

  if (argc != 3 || sscanf (argv[2], "%X", &ip) != 1)
  {
    printf ("usage: bug <file.dbg> <IP>\n");
    exit (-1);
  }

//  ip += 0x00400000;

  fd = open (argv[1], O_RDONLY|O_BINARY);
  if (fd < 0)
  {
    printf ("error: cannot open file %s\n", argv[1]);
    exit (-1);
  }

  size = lseek (fd, 0L, SEEK_END);
  lseek (fd, 0L, SEEK_SET);

  p = malloc(size);
  if (p == null)
  {
    printf ("error: out of memory\n");
    exit (-1);
  }

  if (read (fd, p, size) != (int)size)
  {
    printf ("error: cannot read %s\n", argv[1]);
    exit (-1);
  }

  close (fd);

  l = (unsigned long *)p;
  if (size < 12 || l[0] != 0xF2761287 || l[1] != 0xD5016423)
  {
    printf ("error: illegal file %s\n", argv[1]);
    exit (-1);
  }

  l += 2;

  found = false;
  unit = 0;
  current_unit = 0;

  while (l[1] != 0x80000000 && (char *)&l[2] <= &p[size])
  {
    if (l[0] == 0)
      current_unit = l[1];
    else if (ip >= l[0] && ip < l[2] && l[1] != 0)
    {
      found = true;
      unit = current_unit;
      line = l[1];
    }

    l += 2;
  }

  printf ("IP Address  : %06LX\n", ip);

  if (!found)
  {
    printf ("IP not found\n");
    exit (-1);
  }

  l += 2;
  while (*l)
  {
    if (*l == unit)
    {
      l++;

      len = *l;
      l++;
      if (len)
        printf ("library : %.*s\n", len, l);
      l = (unsigned long *)(((char *)l) + len);

      len = *l;
      l++;
      printf ("source file : %.*s\n", len, l);
      l = (unsigned long *)(((char *)l) + len);
    }
    else
    {
      l++;

      len = *l;
      l++;
      l = (unsigned long *)(((char *)l) + len);

      len = *l;
      l++;
      l = (unsigned long *)(((char *)l) + len);
    }
  }

  printf ("source line : %d\n", line);

  return 0;
}

