
// bug2.c : add filenames and lines to crash report

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <io.h>
#include <ctype.h>
#include <malloc.h>
#include <fcntl.h>
#include "types.h"

void translate_ip (char *dbg_filename, 
                   unsigned long ip, 
                   char *info)  // out
{
  unsigned long size, current_unit, unit, line, *l, len;
  unsigned char *p;
  bool          found;
  int           fd;

  strcpy (info, "");

  fd = open (dbg_filename, O_RDONLY|O_BINARY);
  if (fd < 0)
  {
    printf ("error: cannot open file %s\n", dbg_filename);
    exit (-1);
  }

  size = lseek (fd, 0L, SEEK_END);
  lseek (fd, 0L, SEEK_SET);

  p = malloc(size);
  if (p == null)
  {
    printf ("error: out of memory\n");
    exit (-1);
  }

  if (read (fd, p, size) != (int)size)
  {
    printf ("error: cannot read %s\n", dbg_filename);
    exit (-1);
  }

  close (fd);

  l = (unsigned long *)p;
  if (size < 12 || l[0] != 0xF2761287 || l[1] != 0xD5016423)
  {
    printf ("error: illegal file %s\n", dbg_filename);
    exit (-1);
  }

  l += 2;

  found = false;
  unit = 0;
  current_unit = 0;

  while (l[1] != 0x80000000 && (char *)&l[2] <= &p[size])
  {
    if (l[0] == 0)
      current_unit = l[1];
    else if (ip >= l[0] && ip < l[2] && l[1] != 0)
    {
      found = true;
      unit = current_unit;
      line = l[1];
    }

    l += 2;
  }

  if (!found)
  {
    return;
  }

  l += 2;
  while (*l)
  {
    if (*l == unit)
    {
      l++;

      len = *l;
      l++;
//      if (len)   printf ("library : %.*s\n", len, l);
      l = (unsigned long *)(((char *)l) + len);

      len = *l;
      l++;

      sprintf (info, "  file: %.*s  line: %d", len, l, line);
      
      l = (unsigned long *)(((char *)l) + len);
    }
    else
    {
      l++;

      len = *l;
      l++;
      l = (unsigned long *)(((char *)l) + len);

      len = *l;
      l++;
      l = (unsigned long *)(((char *)l) + len);
    }
  }

  free (p);
}


int main (int argc, char *argv[])
{
  FILE *fp1, *fp2;
  char line[4096], info[512];
  
  if (argc != 4)
  {
    printf ("usage: bug2 <file.dbg> <crash-report.txt>  <modified-crash-report>\n");
    exit (-1);
  }

  fp1 = fopen (argv[2], "r");
  if (fp1 == NULL)
    return -1;
    
  fp2 = fopen (argv[3], "w+");
  if (fp2 == NULL)
    return -1;
    
  while (fgets (line, sizeof(line), fp1) != NULL)
  {
    unsigned long ip;
    int           len = strlen(line);
    int           count;
    
    info[0] = '\0';

    // cut trailing \n
    if (len > 0 && line[len-1] == '\n')
      line[--len] = 0;

    count = 0;
    while (len > 0 && isxdigit(line[len-1]))
    {
      len--;
      count++;
    }
    
    if (count >= 6 && count <= 8)
    {
      if (sscanf (&line[len], "%x", &ip) == 1)
        translate_ip (argv[1], ip, info);
    }

    fprintf (fp2, "%s%s\n", line, info);
  }

  fclose (fp1);  
  fclose (fp2);  
  return 0;
}

