
// codstatem.c : code generator for statements

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <malloc.h>

#include "binsort.h"

#include "pcodes.h"
#include "token.h"
#include "entities.h"
#include "error.h"
#include "goptions.h"
#include "codout.h"
#include "codgen.h"
#include "codexp.h"
#include "type.h"
#include "codstatem.h"

/*******************************************************************************************/

static bool is_empty_statement_sequence (PENTITY e_list)
{
  PENTITY e = e_list;
  
  while (e != null)
  {
    if (e->kind == AN_UNUSED_STATEMENT)
      ;
    else if (e->kind == A_BLOCK_STATEMENT)
    {
      if (!is_empty_statement_sequence (e->info->the_block_statement.inner->entities.first))
        return false;
    }
    else
    {
      return false;
    }
    
    e = e->next;
  }
  
  return true;
}

/*******************************************************************************************/

// returns true if code after statement is unreachable

bool generate_code_for_if_statement (PENTITY    e,
                                     wstring    func_id,
                                     FRAME_INFO *frame,
                                     int        outer_label_break,
                                     int        outer_label_continue,
                                     bool       *any_jumps_to_label_break,
                                     bool       *any_jumps_to_label_continue)
{
  PEXPRESSION exp;
  bool        unreachable, unreachable2;

  exp = e->info->the_if_statement.condition;


  // optimize constant case

  if (exp->kind == A_CONST_ENUMERATION_VALUE)
  {
    if (exp->info.const_enumeration_value_info.value == 1)  // true
    {
      unreachable = generate_code_for_region
                               (func_id,
                                e,
                                e->info->the_if_statement.true_branch->entities.first,
                                frame,
                                outer_label_break,
                                outer_label_continue,
                                any_jumps_to_label_break,
                                any_jumps_to_label_continue);
    }
    else
    {
      unreachable = generate_code_for_region
                               (func_id,
                                e,
                                e->info->the_if_statement.false_branch->entities.first,
                                frame,
                                outer_label_break,
                                outer_label_continue,
                                any_jumps_to_label_break,
                                any_jumps_to_label_continue);
    }

    return unreachable;
  }


  generate_p_location (&e->info->the_if_statement.loc);


  // optimize   "if (e) break;"  and   "if (e) continue;"

  {
    PENTITY first_st;

    first_st = e->info->the_if_statement.true_branch->entities.first;

    if (first_st != null)
    {
      if (first_st->kind == A_BREAK_STATEMENT)
      {
        if (release_objects_in_outer_scope (K_FOR | K_WHILE | K_SWITCH, e, false) == false)
        {
          generate_code_for_branching (e->info->the_if_statement.condition,
                                       /* branch_iftrue = */ true,
                                       outer_label_break,
                                       frame);
          *any_jumps_to_label_break = true;

          unreachable = generate_code_for_region
                               (func_id,
                                e,
                                e->info->the_if_statement.false_branch->entities.first,
                                frame,
                                outer_label_break,
                                outer_label_continue,
                                any_jumps_to_label_break,
                                any_jumps_to_label_continue);
          return unreachable;
        }
      }
      else if (first_st->kind == A_CONTINUE_STATEMENT)
      {
        if (release_objects_in_outer_scope (K_FOR | K_WHILE, e, false) == false)
        {
          generate_code_for_branching (e->info->the_if_statement.condition,
                                       /* branch_iftrue = */ true,
                                       outer_label_continue,
                                       frame);
          *any_jumps_to_label_continue = true;

          unreachable = generate_code_for_region
                               (func_id,
                                e,
                                e->info->the_if_statement.false_branch->entities.first,
                                frame,
                                outer_label_break,
                                outer_label_continue,
                                any_jumps_to_label_break,
                                any_jumps_to_label_continue);
          return unreachable;
        }
      }
    }
  }


  // optimize   "if (e) st; else break;"  and   "if (e) st; else continue;"

  {
    PENTITY first_st;

    first_st = e->info->the_if_statement.false_branch->entities.first;

    if (first_st != null)
    {
      if (first_st->kind == A_BREAK_STATEMENT)
      {
        if (release_objects_in_outer_scope (K_FOR | K_WHILE | K_SWITCH, e, false) == false)
        {
          generate_code_for_branching (e->info->the_if_statement.condition,
                                       /* branch_iftrue = */ false,
                                       outer_label_break,
                                       frame);
          *any_jumps_to_label_break = true;

          unreachable = generate_code_for_region
                               (func_id,
                                e,
                                e->info->the_if_statement.true_branch->entities.first,
                                frame,
                                outer_label_break,
                                outer_label_continue,
                                any_jumps_to_label_break,
                                any_jumps_to_label_continue);
          return unreachable;
        }
      }
      else if (first_st->kind == A_CONTINUE_STATEMENT)
      {
        if (release_objects_in_outer_scope (K_FOR | K_WHILE, e, false) == false)
        {
          generate_code_for_branching (e->info->the_if_statement.condition,
                                       /* branch_iftrue = */ false,
                                       outer_label_continue,
                                       frame);
          *any_jumps_to_label_continue = true;

          unreachable = generate_code_for_region
                               (func_id,
                                e,
                                e->info->the_if_statement.true_branch->entities.first,
                                frame,
                                outer_label_break,
                                outer_label_continue,
                                any_jumps_to_label_break,
                                any_jumps_to_label_continue);
          return unreachable;
        }
      }
    }
  }


  // optimize   "if (e) ; else ;"

  if (is_empty_statement_sequence (e->info->the_if_statement.true_branch->entities.first) &&   // true branch empty
      is_empty_statement_sequence (e->info->the_if_statement.false_branch->entities.first))    // false branch empty
  {
    generate_code_for_bool_expression (e->info->the_if_statement.condition, frame);
    put_code (P_DROP_BOOL);
    return false;
  }


  // optimize   "if (e) ; else st;"

  if (is_empty_statement_sequence (e->info->the_if_statement.true_branch->entities.first))   // true branch empty
  {
    int label_nr1;

    label_nr1 = get_new_near_label_nr ();

    generate_code_for_branching (e->info->the_if_statement.condition,
                                 /* branch_iftrue = */ true,
                                 label_nr1,
                                 frame);

    generate_code_for_region (func_id,
                              e,
                              e->info->the_if_statement.false_branch->entities.first,
                              frame,
                              outer_label_break,
                              outer_label_continue,
                              any_jumps_to_label_break,
                              any_jumps_to_label_continue);
    put_code (P_NEAR_LABEL);
    put_int4 (label_nr1);

    return false;
  }


  // general case

  {
    int label_nr1, label_nr2;

    label_nr1 = get_new_near_label_nr ();

    generate_code_for_branching (e->info->the_if_statement.condition,
                                 /* branch_iftrue = */ false,
                                 label_nr1,
                                 frame);

    unreachable = generate_code_for_region
                             (func_id,
                              e,
                              e->info->the_if_statement.true_branch->entities.first,
                              frame,
                              outer_label_break,
                              outer_label_continue,
                              any_jumps_to_label_break,
                              any_jumps_to_label_continue);

    if (is_empty_statement_sequence (e->info->the_if_statement.false_branch->entities.first))
    {
      put_code (P_NEAR_LABEL);
      put_int4 (label_nr1);

      return false;    // reachable through jmp to label_nr1
    }
    else  // there is an "else"
    {
      if (!unreachable)
      {
        label_nr2 = get_new_near_label_nr ();

        put_code (P_GOTO);
        put_int4 (label_nr2);
      }

      put_code (P_NEAR_LABEL);
      put_int4 (label_nr1);

      unreachable2 = generate_code_for_region
                               (func_id,
                                e,
                                e->info->the_if_statement.false_branch->entities.first,
                                frame,
                                outer_label_break,
                                outer_label_continue,
                                any_jumps_to_label_break,
                                any_jumps_to_label_continue);
      if (!unreachable)
      {
        put_code (P_NEAR_LABEL);
        put_int4 (label_nr2);
      }

      return unreachable & unreachable2;  // unreachable if both are unreachable
    }
  }
}

/*******************************************************************************************/

// returns true if code after statement is unreachable

bool generate_code_for_while_statement (PENTITY    e,
                                        wstring    func_id,
                                        FRAME_INFO *frame)
{
  PEXPRESSION exp;
  int         label_break, label_continue;
  bool        unreachable;
  bool        any_jumps_to_label_break    = false;
  bool        any_jumps_to_label_continue = false;

  exp = e->info->the_while_statement.condition;

  label_continue = get_new_near_label_nr ();
  label_break    = get_new_near_label_nr ();

  if (exp->kind == A_CONST_ENUMERATION_VALUE)
  {
    if (exp->info.const_enumeration_value_info.value == 1)  // true
    {
      put_code (P_NEAR_LABEL);
      put_int4 (label_continue);

      unreachable = generate_code_for_region
                               (func_id,
                                e,
                                e->info->the_while_statement.inner->entities.first,
                                frame,
                                label_break,
                                label_continue,
                                &any_jumps_to_label_break,
                                &any_jumps_to_label_continue);
      if (!unreachable)
      {
        put_code (P_GOTO);
        put_int4 (label_continue);
      }

      put_code (P_NEAR_LABEL);
      put_int4 (label_break);

      return (!any_jumps_to_label_break);
    }
    else   // constant false
    {
      return false;   // reachable
    }
  }
  else
  {
    int label_up;

    label_up = get_new_near_label_nr ();

    if (!is_empty_statement_sequence (e->info->the_while_statement.inner->entities.first))   // there are inner statements
    {
      put_code (P_GOTO);
      put_int4 (label_continue);
    }

    put_code (P_NEAR_LABEL);
    put_int4 (label_up);

    generate_code_for_region (func_id,
                              e,
                              e->info->the_while_statement.inner->entities.first,
                              frame,
                              label_break,
                              label_continue,
                              &any_jumps_to_label_break,
                              &any_jumps_to_label_continue);

    put_code (P_NEAR_LABEL);
    put_int4 (label_continue);

    generate_p_location (&e->info->the_while_statement.loc);

    generate_code_for_branching (e->info->the_while_statement.condition,
                                 /* branch_iftrue = */ true,
                                 label_up,
                                 frame);

    put_code (P_NEAR_LABEL);
    put_int4 (label_break);

    return false;  // always reachable (through false runtime condition or break)
  }
}

/*******************************************************************************************/

// returns true if code after statement is unreachable

bool generate_code_for_for_statement (PENTITY    e,
                                      wstring    func_id,
                                      FRAME_INFO *frame)
{
  PEXPRESSION exp;
  int         label_break, label_continue, label_up;
  bool        unreachable;
  bool        any_jumps_to_label_break    = false;
  bool        any_jumps_to_label_continue = false;

  exp = e->info->the_for_statement.condition;

  label_continue = get_new_near_label_nr ();
  label_break    = get_new_near_label_nr ();
  label_up       = get_new_near_label_nr ();

  generate_code_for_region (func_id,
                            e,
                            e->info->the_for_statement.pre->entities.first,
                            frame,
                            label_break,
                            label_continue,
                            &any_jumps_to_label_break,
                            &any_jumps_to_label_continue);

  if (exp == null || exp->kind == A_CONST_ENUMERATION_VALUE)
  {
    if (exp == null || exp->info.const_enumeration_value_info.value == 1)  // true
    {
      put_code (P_NEAR_LABEL);
      put_int4 (label_up);

      unreachable = generate_code_for_region
                               (func_id,
                                e,
                                e->info->the_for_statement.inner->entities.first,
                                frame,
                                label_break,
                                label_continue,
                                &any_jumps_to_label_break,
                                &any_jumps_to_label_continue);

      put_code (P_NEAR_LABEL);
      put_int4 (label_continue);

      if ((!unreachable) | any_jumps_to_label_continue)
      {
        generate_code_for_region (func_id,
                                  e,
                                  e->info->the_for_statement.post->entities.first,
                                  frame,
                                  label_break,
                                  label_continue,
                                  &any_jumps_to_label_break,
                                  &any_jumps_to_label_continue);
        put_code (P_GOTO);
        put_int4 (label_up);
      }

      put_code (P_NEAR_LABEL);
      put_int4 (label_break);

      return (!any_jumps_to_label_break);
    }
    else   // condition : constant false
    {
      return false;  // always reachable
    }
  }
  else
  {
    int label_start;

    label_start = get_new_near_label_nr ();

    if ((!is_empty_statement_sequence (e->info->the_for_statement.inner->entities.first)) ||
        (!is_empty_statement_sequence (e->info->the_for_statement.post->entities.first)))
    {
      put_code (P_GOTO);
      put_int4 (label_start);
    }

    put_code (P_NEAR_LABEL);
    put_int4 (label_up);

    unreachable = generate_code_for_region
                             (func_id,
                              e,
                              e->info->the_for_statement.inner->entities.first,
                              frame,
                              label_break,
                              label_continue,
                              &any_jumps_to_label_break,
                              &any_jumps_to_label_continue);

    put_code (P_NEAR_LABEL);
    put_int4 (label_continue);

    if ((!unreachable) | any_jumps_to_label_continue)
    {
      generate_code_for_region (func_id,
                                e,
                                e->info->the_for_statement.post->entities.first,
                                frame,
                                label_break,
                                label_continue,
                                &any_jumps_to_label_break,
                                &any_jumps_to_label_continue);
    }

    put_code (P_NEAR_LABEL);
    put_int4 (label_start);

    generate_code_for_region (func_id,
                              e,
                              e->info->the_for_statement.exp_region->entities.first,
                              frame,
                              label_break,
                              label_continue,
                              &any_jumps_to_label_break,
                              &any_jumps_to_label_continue);

    generate_p_location (&e->info->the_for_statement.loc);

    generate_code_for_branching (e->info->the_for_statement.condition,
                                 /* branch_iftrue = */ true,
                                 label_up,
                                 frame);
    put_code (P_NEAR_LABEL);
    put_int4 (label_break);

    return false;  // always reachable (through false runtime condition or break)
  }
}

/*******************************************************************************************/

void generate_code_for_break_statement (PENTITY    e,
                                        wstring    func_id,
                                        FRAME_INFO *frame,
                                        int        outer_label_break)
{
  put_code (P_GOTO);
  put_int4 (outer_label_break);
}

/*******************************************************************************************/

void generate_code_for_continue_statement (PENTITY    e,
                                           wstring    func_id,
                                           FRAME_INFO *frame,
                                           int        outer_label_continue)
{
  put_code (P_GOTO);
  put_int4 (outer_label_continue);
}

/*******************************************************************************************/

static void compute_type_info (PENTITY subtype, INTEGER_INFO *info)
{
  PENTITY type;

  type = complete_type_of (subtype);

  switch (type->kind)
  {
    case AN_ENUMERATION_TYPE:
      memcpy (info, &integer_info[type->info->the_enumeration_type.base], sizeof(INTEGER_INFO));
      break;

    case AN_INTEGER_TYPE:
      memcpy (info, &integer_info[type->info->the_integer_type.type], sizeof(INTEGER_INFO));
      break;

    default:
      fatal_compiler_error0 ("compute_type_info()");
  }
}

/*****************************************************************************/

typedef struct {
  int8             value;
  FLOW_ALTERNATIVE *f;
} CASE_NODE;

static int compare (const CASE_NODE *p1, const CASE_NODE *p2)
{
  if (p1->value < p2->value)
    return -1;
  if (p1->value > p2->value)
    return +1;
  return 0;
}

/*******************************************************************************************/

typedef struct {
  int  first;   // first index into case table
  int  last;    // last index into case table
} INODE;

/*******************************************************************************************/

static void generate_switch_flow (int8         min,     // range of expression to evaluate
                                  int8         max,
                                  INTEGER_INFO *type_info,
                                  INODE        *table2,
                                  uint4        table2_count,
                                  CASE_NODE    *table,
                                  uint4        table_count,
                                  int          label_default_drop,
                                  int          label_default_no_drop,
                                  int          label_exit,
                                  PENTITY      e,
                                  wstring      func_id,
                                  FRAME_INFO   *frame,
                                  int          outer_label_continue,
                                  bool         *any_jumps_to_label_continue)
{
  uint4 middle;
  int8  first_value, last_value;
  int   label_larger_or_equal;
  uint4 i, pool_offset;
  int8  n;
  bool  unreachable, any_jumps_to_label_break;
  POOL  *pool;

  if (table2_count > 1)
  {
    // we must split the possible intervals in two groups.
    middle = table2_count/2;

    first_value = table[table2[middle].first].value;

    label_larger_or_equal = get_new_near_label_nr ();


    // if larger or equal than first_value, branch to label_larger_or_equal

    if (type_info->size <= 4)
    {
      if (type_info->is_signed)
        put_code (P_SWITCH_CMP_S4);
      else
        put_code (P_SWITCH_CMP_U4);
      put_int4 ((int4)first_value);
    }
    else
    {
      put_code (P_SWITCH_CMP_8);
      put_int8 (first_value);
    }
    put_byte (CMP_LARGER_OR_EQUAL);
    put_int4 (label_larger_or_equal);


    // smaller than middle
    generate_switch_flow (min, first_value-1, type_info, table2, middle, table, table_count,
                          label_default_drop, label_default_no_drop, label_exit,
                          e, func_id, frame, outer_label_continue, any_jumps_to_label_continue);

    put_code (type_info->size <= 4 ? P_SHADOW_4 : P_SHADOW_8);

    put_code (P_NEAR_LABEL);
    put_int4 (label_larger_or_equal);

    // larger or equal than middle
    generate_switch_flow (first_value, max, type_info, &table2[middle], table2_count-middle, table, table_count,
                          label_default_drop, label_default_no_drop, label_exit,
                          e, func_id, frame, outer_label_continue, any_jumps_to_label_continue);
  }
  else   // table_count == 1 : only 1 interval left
  {
    first_value = table[table2[0].first].value;
    last_value  = table[table2[0].last].value;

    if (first_value < last_value)   // a jump table with several values
    {
      if (type_info->size == 8)
      {
        if (min < first_value)    // exp possibly smaller
        {
          put_code (P_SWITCH_CMP_8);
          put_int8 (first_value);
          put_byte (CMP_SMALLER);
          put_int4 (label_default_drop);
        }

        if (max > last_value)    // exp possibly larger
        {
          put_code (P_SWITCH_CMP_8);
          put_int8 (last_value);
          put_byte (CMP_LARGER);
          put_int4 (label_default_drop);
        }

        //  subtract lower bound
        put_code (P_CTE_8);
        put_int8 (first_value);
        put_code (P_SUB8);

        // conversion from int8 to uint4
        put_code (P_CONV_LONG_INT);
      }
      else    // for int4 / uint4
      {
        // subtract lower bound (yields unsigned result)
        put_code (P_CTE_4);
        put_int4 ((int4)first_value);
        put_code (P_SUB4);

        if (min < first_value || max > last_value)    // exp has be outside jump table
        {
          put_code (P_SWITCH_CMP_U4);      // always unsigned !
          put_int4 ((int4)(last_value - first_value));
          put_byte (CMP_LARGER);
          put_int4 (label_default_drop);
        }
      }

      // generate code for jump table

      pool = new_pool_constant (address_size * (int4)(last_value - first_value + 1), address_size);
      pool_offset = 0;

      // generate label table entries
      i = table2[0].first;
      for (n=first_value; n<=last_value; n++)
      {
        if (i < table_count && table[i].value == n)   // case constant exists
        {
          if (table[i].f->label_nr_drop == 0)     // alternative has no label yet
          {
            table[i].f->label_nr_drop    = get_new_near_label_nr ();
            table[i].f->label_nr_no_drop = get_new_near_label_nr ();
          }
          store_integer (pool, pool_offset, table[i].f->label_nr_no_drop, address_size);
          pool_offset += address_size;
          i++;
        }
        else   // case constant does not exist
        {
          store_integer (pool, pool_offset, label_default_no_drop, address_size);
          pool_offset += address_size;
        }
      }

      put_code (P_JUMP_4);
      put_int8 (serial_nr_of_pool_cte (pool));
    }
    else   // single value
    {
      if (min < first_value || max > last_value)
      {
        if (type_info->size <= 4)
        {
          put_code (P_SWITCH_CMP_U4);
          put_int4 ((int4)first_value);
        }
        else
        {
          put_code (P_SWITCH_CMP_8);
          put_int8 (first_value);
        }
        put_byte (CMP_NOT_EQUAL);
        put_int4 (label_default_drop);
      }


      i = table2[0].first;
      if (table[i].f->label_nr_drop == 0)     // alternative has no label yet
      {
        table[i].f->label_nr_drop    = get_new_near_label_nr ();
        table[i].f->label_nr_no_drop = get_new_near_label_nr ();
      }

      if (table[i].f->code_generated)  // code was already generated earlier
      {
        if (type_info->size <= 4)
          put_code (P_DROP_4);
        else
          put_code (P_DROP_8);

        put_code (P_GOTO);
        put_int4 (table[i].f->label_nr_no_drop);
      }
      else        // generate code for this alternative
      {
        put_code (P_NEAR_LABEL);
        put_int4 (table[i].f->label_nr_drop);

        if (type_info->size <= 4)
          put_code (P_DROP_4);
        else
          put_code (P_DROP_8);

        put_code (P_NEAR_LABEL);
        put_int4 (table[i].f->label_nr_no_drop);

        // generate code

        unreachable = generate_code_for_region
                             (func_id,
                              e,
                              table[i].f->inner->entities.first,
                              frame,
                              label_exit,
                              outer_label_continue,
                              &any_jumps_to_label_break,
                              any_jumps_to_label_continue);
        if (!unreachable)
        {
          put_code (P_GOTO);
          put_int4 (label_exit);
        }

        table[i].f->code_generated = true;
      }
    }
  }
}

/*******************************************************************************************/

void generate_code_for_switch_statement (PENTITY    e,
                                         wstring    func_id,
                                         FRAME_INFO *frame,
                                         int        outer_label_continue,
                                         bool       *any_jumps_to_label_continue)
{
  PEXPRESSION      exp;
  PENTITY          type;
  uint4            count, i, j1, j2, k;
  CASE_NODE        *table;
  FLOW_ALTERNATIVE *f;
  CASE_CONSTANT    *l;
  INODE            *table2;
  bool             found;
  int              label_default_drop, label_default_no_drop, label_exit;
  bool             unreachable, any_jumps_to_label_break;
  INTEGER_INFO     type_info;
  int8             diff;

  exp = e->info->the_switch_statement.value;
  type = complete_type_of (exp->base_type_or_null);

  if (is_constant_exp (exp))
  {
    int8 value;

    if (exp->kind == A_CONST_ENUMERATION_VALUE)
      value = exp->info.const_enumeration_value_info.value;
    else
      value = exp->info.const_integer_value_info.value;

    // find the matching alternative

    f = e->info->the_switch_statement.first_alt;
    while (f != null)
    {
      l = f->cte_list;
      while (l != null)
      {
        if (l->value == value)
          break;
        l = l->next;
      }
      if (l != null)    // value found
        break;

      f = f->next;
    }

    if (l == null)   // value not found
      f = e->info->the_switch_statement.last_alt;  // set f to last default case


    // generate code for branch f

    label_exit = get_new_near_label_nr ();

    generate_code_for_region (func_id,
                              e,
                              f->inner->entities.first,
                              frame,
                              label_exit,
                              outer_label_continue,
                              &any_jumps_to_label_break,
                              any_jumps_to_label_continue);

    put_code (P_NEAR_LABEL);
    put_int4 (label_exit);

    return;
  }

  generate_p_location (&e->info->the_switch_statement.loc);

  generate_code_for_discrete_expression (exp, frame);
  if (type == type_bool)
    put_code (P_CONV_BOOL_INT);

  compute_type_info (type, &type_info);


  count = e->info->the_switch_statement.count;

  if (count == 0)    // only 1 default branch
  {
    if (type_info.size <= 4)
      put_code (P_DROP_4);
    else
      put_code (P_DROP_8);

    // generate code for default branch

    label_exit = get_new_near_label_nr ();

    generate_code_for_region (func_id,
                              e,
                              e->info->the_switch_statement.last_alt->inner->entities.first,
                              frame,
                              label_exit,
                              outer_label_continue,
                              &any_jumps_to_label_break,
                              any_jumps_to_label_continue);

    put_code (P_NEAR_LABEL);
    put_int4 (label_exit);

    return;
  }


  // load case constants into table

  table = malloc (count * sizeof(CASE_NODE));
  if (table == null)
    fatal_out_of_memory_error ("generate_code_for_switch_statement(1)\n");

  i = 0;
  f = e->info->the_switch_statement.first_alt;
  while (f != null)
  {
    l = f->cte_list;
    while (l != null)
    {
      table[i].value = l->value;
      table[i].f     = f;
      i++;
      l = l->next;
    }

    f = f->next;
  }


  // sort table by constant value

  if (binary_sort (table, count, sizeof(CASE_NODE), compare) < 0)
    fatal_out_of_memory_error ("generate_code_for_switch_statement(2)\n");


  // allocate inode table

  table2 = malloc (count * sizeof(INODE));
  if (table2 == null)
    fatal_out_of_memory_error ("generate_code_for_switch_statement(3)\n");


  // fill inode table with jump table ranges

  k = 0;    // index into table2

  i = 0;
  while (i < count)
  {
    const uint MIN    = 4;   // minimum nb cases in jump table
    const uint FACTOR = 4;   // allow 1/FACTOR full jump table

    j1 = i;

    // advance j1 until reaching MIN cases with close values

    found = false;

    for (;;)
    {
      if (count - j1 < MIN)  // less than MIN cases left
        break;

      diff = table[j1+MIN-1].value - table[j1].value;
      if (diff >= 0 && diff <= FACTOR*MIN)   // found MIN cases with close values
      {
        found = true;
        break;
      }

      j1++;
    }

    if (found)
    {
      j2 = j1 + (MIN-1);

      // expand the table upwards
      while (j1 > i)
      {
        diff = table[j2].value - table[j1-1].value;
        if (diff < 0 || diff > FACTOR*(j2-(j1-1)+1))
          break;
        j1--;
      }

      // expand the table downwards
      while (j2+1 < count)
      {
        diff = table[j2+1].value - table[j1].value;
        if (diff < 0 || diff > FACTOR*((j2+1)-j1+1))
          break;
        j2++;
      }

      // expand the table upwards again
      while (j1 > i)
      {
        diff = table[j2].value - table[j1-1].value;
        if (diff < 0 || diff > FACTOR*(j2-(j1-1)+1))
          break;
        j1--;
      }

      // we have an interval j1 .. j2 for a jump table

      // add individual values, from [i to j1[

      while (i < j1)
      {
        table2[k].first = i;
        table2[k].last  = i;
        k++;
        i++;
      }

      // add jump table [j1 to j2]

      table2[k].first = j1;
      table2[k].last  = j2;
      k++;

      i = j2+1;
    }
    else  // no jump table found
    {
      // add individual values, from [i to count[

      while (i < count)
      {
        table2[k].first = i;
        table2[k].last  = i;
        k++;
        i++;
      }

      i = count;
    }
  }


  label_default_drop    = get_new_near_label_nr ();
  label_default_no_drop = get_new_near_label_nr ();
  label_exit            = get_new_near_label_nr ();


  // be sure value is in lowest CPU register
  if (type_info.size <= 4)
    put_code (P_SYNC_4);
  else
    put_code (P_SYNC_8);


  // table2 has k elements

  generate_switch_flow (type_info.min, type_info.max, &type_info, table2, k, table, count,
                        label_default_drop, label_default_no_drop, label_exit,
                        e, func_id, frame, outer_label_continue, any_jumps_to_label_continue);


  // generate code for all branches where it hasn't been done;
  // this includes the default branch.

  f = e->info->the_switch_statement.first_alt;
  while (f != null)
  {
    if (!f->code_generated)
    {
      if (!f->is_default)
      {
        put_code (type_info.size <= 4 ? P_SHADOW_4 : P_SHADOW_8);

        put_code (P_NEAR_LABEL);
        put_int4 (f->label_nr_drop);

        put_code (type_info.size <= 4 ? P_DROP_4 : P_DROP_8);

        put_code (P_NEAR_LABEL);
        put_int4 (f->label_nr_no_drop);

        // generate code

        unreachable = generate_code_for_region
                             (func_id,
                              e,
                              f->inner->entities.first,
                              frame,
                              label_exit,
                              outer_label_continue,
                              &any_jumps_to_label_break,
                              any_jumps_to_label_continue);
        if (!unreachable)
        {
          put_code (P_GOTO);
          put_int4 (label_exit);
        }
      }
      else      // default case
      {
        put_code (type_info.size <= 4 ? P_SHADOW_4 : P_SHADOW_8);

        put_code (P_NEAR_LABEL);
        put_int4 (label_default_drop);

        put_code (type_info.size <= 4 ? P_DROP_4 : P_DROP_8);

        put_code (P_NEAR_LABEL);
        put_int4 (label_default_no_drop);

        // generate code

        generate_code_for_region
                  (func_id,
                   e,
                   f->inner->entities.first,
                   frame,
                   label_exit,
                   outer_label_continue,
                   &any_jumps_to_label_break,
                   any_jumps_to_label_continue);

        put_code (P_NEAR_LABEL);
        put_int4 (label_exit);
      }
    }

    f = f->next;
  }


  free (table2);
  free (table);
}

/*******************************************************************************************/
