
// complete.c

#include <stdio.h>
#include <string.h>

#include "token.h"
#include "lex.h"
#include "entities.h"
#include "error.h"

//----------------------------------------------------------------------------------------

// called for function bodies and block statements

void check_local_completion (PENTITY first)
{
  PENTITY e;

  e = first;
  while (e != null)
  {
    if (e->kind == AN_INCOMPLETE_TYPE)
    {
      if (e->info->the_incomplete_type.full_type == null)
      {
        char msg[260];
        sprintf (msg, "missing full type declaration for incomplete type %.128S", e->identifier_or_null);
        semantic_error (msg, &token.pos);
      }
    }

    e = e->next;
  }
}

//----------------------------------------------------------------------------------------

bool is_body_required (PENTITY first)
{
  PENTITY e;

  e = first;
  while (e != null)
  {
    if (e->kind == AN_INCOMPLETE_TYPE && e->info->the_incomplete_type.full_type == null)
      return true;

    if (e->kind == AN_OPAQUE_TYPE)
      return true;

    if (e->kind == A_PACKAGE_DECLARATION && e->info->the_package_declaration.is_body_required)
      return true;

    if (e->kind == A_FUNCTION_DECLARATION &&
        e->info->the_function_declaration.to_type
         ->info->the_function_pointer_type.extern_dll_or_null == null)
      return true;

    e = e->next;
  }

  return false;
}

//----------------------------------------------------------------------------------------

// called by unit bodies and package bodies
// to check the interface and body parts
// uses current token position

void check_global_completion (PENTITY first)
{
  PENTITY e;
  char    msg[260];

  e = first;
  while (e != null)
  {
    if (e->kind == AN_INCOMPLETE_TYPE && e->info->the_incomplete_type.full_type == null)
    {
      sprintf (msg, "missing full type declaration for incomplete type %.128S", e->identifier_or_null);
      semantic_error (msg, &token.pos);
    }

    if (e->kind == AN_OPAQUE_TYPE && e->info->the_opaque_type.full_type == null)
    {
      sprintf (msg, "missing full type declaration for opaque type %.128S", e->identifier_or_null);
      semantic_error (msg, &token.pos);
    }

    if (e->kind == A_PACKAGE_DECLARATION &&
        e->info->the_package_declaration.is_body_required &&
        e->info->the_package_declaration.to_package_body_or_null == null)
    {
      sprintf (msg, "missing body for package %.128S", e->identifier_or_null);
      semantic_error (msg, &token.pos);
    }

    if (e->kind == A_FUNCTION_DECLARATION &&
        e->info->the_function_declaration.to_type
         ->info->the_function_pointer_type.extern_dll_or_null == null &&
        e->info->the_function_declaration.to_function_body_or_null == null)
    {
      sprintf (msg, "missing body for function %.128S", e->identifier_or_null);
      semantic_error (msg, &token.pos);
    }

    e = e->next;
  }
}

//----------------------------------------------------------------------------------------
