
// dbginfo : write debug info into file

#include <string.h>
#include <io.h>
#include <fcntl.h>
#include <malloc.h>
#include "types.h"

#include "error.h"
#include "unit.h"
#include "dbginfo.h"

/*****************************************************************************/

#define SIZE  4096

/*****************************************************************************/

typedef struct block {
  byte data[SIZE];
  int  len;
  struct block *next;  
} BLOCK;

static BLOCK *first, *last, *move;

/*****************************************************************************/

static void store (byte *data, int size)
{
  BLOCK *n;

  if (last == null || last->len + size > SIZE)
  {
    n = malloc (sizeof(BLOCK));
    memset (n, 0x00, sizeof(BLOCK));
    if (first == null)
      first = n;
    else
      last->next = n;
    last = n;
  }

  memcpy (&last->data[last->len], data, size);
  last->len += size;
}

/*****************************************************************************/

void dbg_header ()
{
  uint4 data[2];

  data[0] = 0xF2761287;
  data[1] = 0xD5016423;
  store ((byte *)data, 8);
}

void dbg_new_unit (int unit)
{
  uint4 data[2];

  data[0] = 0;
  data[1] = unit;
  store ((byte *)data, 8);

  move = last;
}

void dbg_store_line (int line, int ip)
{
  uint4 data[2];
  data[0] = ip;
  data[1] = line;
  store ((byte *)data, 8);
}

void move_dbg_lines (int start_address, int rip_offset)
{
  BLOCK *p;
  int   ofs;

  p = move;
  while (p != null)
  {
    ofs = p->len;

    for (;;)
    {
      ofs -= 8;
      if (ofs < 0 || *((int *)&p->data[ofs]) < start_address)
        break;
      *((uint *)&p->data[ofs]) += rip_offset;
    }

    p = p->next;
  }
}

void dbg_store_unit_name (int unit, char *library, char *source)
{
  int len;

  store ((byte *)&unit, 4);

  len = strlen(library);
  store ((byte *)&len, 4);
  store (library, len);

  len = strlen(source);
  store ((byte *)&len, 4);
  store (source, len);
}

void dbg_save (char *debug_filename)
{
  int   fd;
  BLOCK *p;

  dbg_new_unit (0x80000000);  // signals end of list
  list_units_for_debug ();
  dbg_new_unit (0x80000000);  // signals end of list

  fd = open (debug_filename, O_CREAT|O_TRUNC|O_RDWR|O_BINARY, 0660);
  if (fd < 0)
  {
    fatal_compiler_error0 ("cannot save debug file on disk");
    return;
  }

  p = first;
  while (p != null)
  {
    if (write (fd, p->data, p->len) != (int)p->len)
    {
      close (fd);
      unlink (debug_filename);
      fatal_compiler_error0 ("cannot write debug file to disk");
      return;
    }

    p = p->next;
  }

  if (close (fd) < 0)
  {
    unlink (debug_filename);
    fatal_compiler_error0 ("cannot close debug file");
    return;
  }
}

