
// dllnames.c

#include <stdio.h>
#include <string.h>
#include <malloc.h>

#include "bintree.h"

#include "types.h"
#include "error.h"

#include "dllnames.h"

/**********************************************************************************/

typedef struct {
  uint4  nr;
  char   *dll;
  char   *func;
} NODE;

static BINARY_TREE tree;
static bool        tree_created;
static uint4       unique_nr;

/**********************************************************************************/

static int compare (void               *user,
                    const NODE         *a,
                    const unsigned int size1,
                    const NODE         *b,
                    const unsigned int size2)
{
  if (a->nr < b->nr)
    return -1;
  if (a->nr > b->nr)
    return +1;
  return 0;
}

/**********************************************************************************/

static char *new_string (char *s)
{
  char *p;
  p = malloc (strlen(s) + 1);
  if (p == null)
    fatal_out_of_memory_error ("new_string(1)");
  strcpy (p, s);
  return p;
}

/**********************************************************************************/

uint4 insert_dll_name (char *dll, char *func)
{
  NODE n;
  int  rc;

  if (!tree_created)
  {
    create_btree (&tree, NULL, compare);
    tree_created = true;
  }

  n.nr = ++unique_nr;
  n.dll = new_string (dll);
  n.func = new_string (func);

  rc = insert_btree (&tree, &n, sizeof(n));
  if (rc)
    fatal_out_of_memory_error ("insert_dll_name(1)");

  return unique_nr;
}

/**********************************************************************************/

void get_dll_name (uint4 nr, char **dll, char **func)
{
  NODE n;
  int  rc;
  unsigned int dummy;

  n.nr = nr;

  rc = retrieve_btree (&tree, &n, sizeof(n), sizeof(n), &dummy, BT_EQUAL);
  if (rc)
    fatal_out_of_memory_error ("get_dll_name(1)");

  *dll = n.dll;
  *func = n.func;
}

/**********************************************************************************/
