
// error.c

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "token.h"
#include "unit.h"

#define MAX_ERRORS 25


static char library_file[260+1], previous_library_file[260+1];
static char source_file[260+1], previous_source_file[260+1];
static long nb_warnings, nb_errors;
static int  code_generator_source_line;

int g_unit_key;    // used by entities.c when creating new entities

void set_current_error_source_file (string lib, string file, int key)
{
  strcpy (library_file, lib);
  strcpy (source_file, file);
  g_unit_key = key;   // used by entities.c when creating new entities
}

void error_set_code_generator_source_line (int line)
{
  code_generator_source_line = line;
}

void stop_compilation ()
{
  printf ("stop compilation\n");
  exit (-1);
}

void error_set_code_generator_unit_key (int key)
{
  static int previous_unit_key;
  if (key != previous_unit_key)
  {
    retrieve_library_and_source_names (key, library_file, source_file);
    previous_unit_key = key;
  }
}

static void display_source_file ()
{
  if (nb_errors == MAX_ERRORS)
  {
    printf ("number of errors >= %d\n", MAX_ERRORS);
    stop_compilation ();
  }

  if (strcmp (previous_library_file, library_file) != 0 ||
      strcmp (previous_source_file, source_file) != 0)
  {
    strcpy (previous_library_file, library_file);
    strcpy (previous_source_file, source_file);

    printf ("\n");
    if (library_file[0] == '\0')
      printf ("compiling %s:\n", source_file);
    else
      printf ("compiling %s: [%s]\n", source_file, library_file);
  }
}

void get_current_source_filename (char *filename)
{
  if (library_file[0] == '\0')
    sprintf (filename, "%s", source_file);
  else
    sprintf (filename, "%s: [%s]", source_file, library_file);
}

void warning (string s, TEXT_POSITION *pos)
{
  display_source_file ();
  printf ("line %d col %d : warning : %s\n", pos->line, pos->col, s);
  nb_warnings++;
}

void config_error (string s)
{
  display_source_file ();
  printf ("config: %s\n", s);
  nb_errors++;
  stop_compilation ();
}

void lexical_error (string s, TEXT_POSITION *pos)
{
  display_source_file ();
  printf ("line %d col %d : lexical : %s\n", pos->line, pos->col, s);
  nb_errors++;
}

void syntax_error (string s, TEXT_POSITION *pos)
{
  display_source_file ();
  printf ("line %d col %d : syntax : %s\n", pos->line, pos->col, s);
  nb_errors++;
}

void semantic_error  (string s, TEXT_POSITION *pos)
{
  display_source_file ();
  printf ("line %d col %d : semantic : %s\n", pos->line, pos->col, s);
  nb_errors++;
}

// fatal errors before code generation :

void fatal_compiler_error (string s, TEXT_POSITION *pos)
{
  display_source_file ();
  printf ("line %d col %d : FATAL COMPILER ERROR : %s\n", pos->line, pos->col, s);
  nb_errors++;
  stop_compilation ();
}


// for resource file

void resource_error (string resource_filename, string msg, int line_nr, int col)
{
  printf ("compiling %s:\n", resource_filename);
  printf ("line %d col %d : syntax : %s\n", line_nr, col, msg);
  stop_compilation ();
}


// for code generator :

void code_generator_error (string s)
{
  display_source_file ();
  printf ("\n");
  printf ("line %d : code generator : %s\n", code_generator_source_line, s);
  nb_errors++;
}


// fatal errors during code generation or in pool.c :

void fatal_compiler_error0 (string s)
{
  display_source_file ();
  printf ("line %d : FATAL COMPILER ERROR : %s\n", code_generator_source_line, s);
  nb_errors++;
  stop_compilation ();
}

void fatal_out_of_memory_error (string s)
{
  display_source_file ();
  printf ("line %d : out of memory : %s\n", code_generator_source_line, s);
  nb_errors++;
  stop_compilation ();
}

long nb_of_compilation_errors ()
{
  return nb_errors;
}

long nb_of_compilation_warnings ()
{
  return nb_warnings;
}
