
// extract resources from an EXE file

#include <windows.h>
#include <stdio.h>
#include <io.h>
#include <fcntl.h>

#define IS_INTRESOURCE(x) (((unsigned int)(x))<65536)

BOOL CALLBACK EnumResLang(
  HMODULE hModule,
  LPCTSTR lpszType,
  LPCTSTR lpszName,
  WORD    wIDLanguage,
  DWORD   lParam
)
{
  HRSRC r;
  HGLOBAL h;
  HANDLE lpResLock;
  DWORD size;
  char filename[260];
  int fd;

//  printf ("    lang: %d\n", wIDLanguage);

  r = FindResourceEx (hModule, lpszType, lpszName, wIDLanguage);
  if (r == 0)
  {
    printf ("      error: cannot find resource\n");
    return TRUE;
  }

  h = LoadResource (hModule, r);
  if (h == 0)
  {
    printf ("      error: cannot load resource\n");
    return TRUE;
  }

  lpResLock = LockResource (h);
  if (lpResLock == NULL)
    return -1;

  size = SizeofResource (hModule, r);

  sprintf (filename, "");

  if (IS_INTRESOURCE(lpszName))
    sprintf (&filename[strlen(filename)], "%d", lpszName);
  else
    sprintf (&filename[strlen(filename)], "%s", lpszName);
  strcat (filename, ".");
  
  if (IS_INTRESOURCE(lpszType))
    sprintf (&filename[strlen(filename)], "%d", lpszType);
  else
    sprintf (&filename[strlen(filename)], "%s", lpszType);
  strcat (filename, ".");

  sprintf (&filename[strlen(filename)], "%04X", wIDLanguage);

  strcat (filename, ".resource");


  if (IS_INTRESOURCE(lpszName))
    printf ("%-20d", lpszName);
  else
    printf ("%-20s", lpszName);
  printf ("  ");
  
  if (IS_INTRESOURCE(lpszType))
    printf ("%-20d", lpszType);
  else
    printf ("%-20s", lpszType);
  printf ("  ");

  printf ("%s\n", filename);


  fd = _open (filename, O_CREAT|O_TRUNC|O_RDWR|O_BINARY, 0660);
  if (fd < 0)
  {
    printf ("        error: creating %s failed\n", filename);
    return TRUE;
  }

  if (_write (fd, lpResLock, size) != (int)size)
  {
    _close (fd);
    _unlink (filename);
    printf ("        error: writing to %s failed\n", filename);
    return TRUE;
  }

  if (_close (fd) < 0)
  {
    printf ("        error: closing %s failed\n", filename);
    return TRUE;
  }

  return TRUE;
}

BOOL CALLBACK EnumResNames(
  HMODULE hModule,
  LPCTSTR lpszType,
  LPTSTR  lpszName,
  DWORD   lParam
)
{

#if 0
  if (IS_INTRESOURCE(lpszName))
    printf ("  name: %d\n", lpszName);
  else
    printf ("  name: %s\n", lpszName);
#endif

  EnumResourceLanguages (hModule, lpszType, lpszName, (ENUMRESLANGPROC)EnumResLang, 0);
  return TRUE;
}

BOOL CALLBACK EnumResType (
  HMODULE hModule,
  LPTSTR lpszType,
  DWORD  lParam
)
{
#if 0
  if (IS_INTRESOURCE(lpszType))
    printf ("type: %d\n", lpszType);
  else
    printf ("type: %s\n", lpszType);
#endif

  EnumResourceNames (hModule, lpszType, (ENUMRESNAMEPROC)EnumResNames, 0);
  return TRUE;
}

int main (int argc, char *argv[])
{
  HMODULE h;

  if (argc != 2)
  {
    printf ("extract resources from an EXE file\n");
    printf ("extract <file.exe>\n");
    return -1;
  }

  h = LoadLibrary (argv[1]);
  if (h == 0)
  {
    printf ("error: cannot open %s\n", argv[1]);
    return -1;
  }


  EnumResourceTypes (h, (ENUMRESTYPEPROC)EnumResType, 0);

  return 0;
}
