
// glibpaths.c : global library paths

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <malloc.h>
#include <io.h>

#include "inifile.h"

#include "types.h"
#include "strings.h"

#include "error.h"

//=================================================================================

static char **library_path;
static int  nb_lib_paths;

//=================================================================================

static char *new_string (string str)
{
  char *s;
  s = malloc (strlen(str)+1);
  if (s == null)
    fatal_out_of_memory_error ("new_string() failed");
  strcpy (s, str);
  return s;
}

//=================================================================================

void load_global_library_paths (string ini_filename)
{
  int   nth;
  char  *section = "library";
  char  key[128], value[128], str[128];

  nb_lib_paths = 0;
  for (nth=0; ; nth++)
  {
    if (get_nth_key (ini_filename, section, nth, key, sizeof(key)) < 0 ||
        get_nth_parameter (ini_filename, section, nth, value, sizeof(value)) < 0)
      break;

    if (strcmp (key, "dir") == 0)
    {
      if (value[0] == '\0' || value[strlen(value)-1] != '/')
      {
        sprintf (str, "%.32s : [library] dir path '%s' must end with /", ini_filename, value);
        config_error (str);
      }
      else
      {
        nb_lib_paths++;
      }
    }
    else
    {
      sprintf (str, "%.32s : illegal compiler [library] key %.64s", ini_filename, key);
      config_error (str);
    }
  }

  library_path = (char **)malloc (sizeof(char *) * nb_lib_paths);
  if (library_path == null)
    fatal_out_of_memory_error ("load_global_library_paths(1)");

  for (nth=0; ; nth++)
  {
    if (get_nth_key (ini_filename, section, nth, key, sizeof(key)) < 0 ||
        get_nth_parameter (ini_filename, section, nth, value, sizeof(value)) < 0)
      break;
    library_path[nth] = new_string (value);
  }
}

//=================================================================================

void find_library (string unit_name,
                   string library_filename)   // out
{
  char filename[260*2];
  int  i;

  sprintf (filename, "%s%s", unit_name, ".lib");
  if (access (filename, 0) == 0)
  {
    strcpy (library_filename, filename);
    return;
  }

  for (i=0; i<nb_lib_paths; i++)
  {
    sprintf (filename, "%s%s%s", library_path[i], unit_name, ".lib");
    if (access (filename, 0) == 0)
    {
      strcpy (library_filename, filename);
      return;
    }
  }

  sprintf (filename, "cannot find library %s.lib", unit_name);
  config_error (filename);

  stop_compilation ();
}

//=================================================================================
