
// goptions.c : global compiler options

#include <stdio.h>
#include <string.h>
#include <malloc.h>

#include "inifile.h"

#include "types.h"
#include "strings.h"

#include "error.h"

#include "goptions.h"

/**********************************************************************************/

int address_size = 8;      // 4 or 8 bytes
bool big_endian = false;   // false=Intel, true=Motorola
bool pointer_checks_enabled = true;
bool array_checks_enabled = true;
bool assertion_checks_enabled = true;
bool option_check_stack_M16_alignment = false;  // true for test, false for prod
bool g_tracing;

/**********************************************************************************/

void load_global_options (string ini_filename)
{
  int   nth;
  char  *section = "options";
  char  key[128], value[128], str[128];

  for (nth=0; ; nth++)
  {
    if (get_nth_key (ini_filename, section, nth, key, sizeof(key)) < 0 ||
        get_nth_parameter (ini_filename, section, nth, value, sizeof(value)) < 0)
      break;

    if (strcmp (key, "memory_model") == 0)
    {
      if (strcmp (value, "32") == 0)
      {
        address_size = 4;
      }
      else if (strcmp (value, "64") == 0)
      {
        address_size = 8;
      }
      else
      {
        sprintf (str, "%.32s : illegal compiler option %.64s : only 32 or 64 supported", ini_filename, key);
        config_error (str);
      }
    }
    else if (strcmp (key, "pointer_check") == 0)
    {
      if (strcmp (value, "yes") == 0)
        pointer_checks_enabled = true;
      else if (strcmp (value, "no") == 0)
        pointer_checks_enabled = false;
      else
      {
        sprintf (str, "%.32s : illegal compiler option %.64s : yes or no expected", ini_filename, key);
        config_error (str);
      }
    }
    else if (strcmp (key, "array_check") == 0)
    {
      if (strcmp (value, "yes") == 0)
        array_checks_enabled = true;
      else if (strcmp (value, "no") == 0)
        array_checks_enabled = false;
      else
      {
        sprintf (str, "%.32s : illegal compiler option %.64s : yes or no expected", ini_filename, key);
        config_error (str);
      }
    }
    else if (strcmp (key, "assertion_check") == 0)
    {
      if (strcmp (value, "yes") == 0)
        assertion_checks_enabled = true;
      else if (strcmp (value, "no") == 0)
        assertion_checks_enabled = false;
      else
      {
        sprintf (str, "%.32s : illegal compiler option %.64s : yes or no expected", ini_filename, key);
        config_error (str);
      }
    }
    else
    {
      sprintf (str, "%.32s : illegal compiler option %.64s", ini_filename, key);
      config_error (str);
    }
  }
}

/**********************************************************************************/
