
// gsymb.c : global symbols

#include <stdio.h>
#include <string.h>
#include <malloc.h>

#include "inifile.h"

#include "types.h"
#include "strings.h"

#include "lex.h"
#include "error.h"
#include "symbols.h"

#include "goptions.h"

#include "gsymb.h"

/**********************************************************************************/

BINARY_TREE global_symbols;      // referenced in lex.c

/**********************************************************************************/

void load_global_symbols (string ini_filename)
{
  int   nth, rc;
  char  *section = "symbols";
  char  key[128], value[128], str[128];
  wchar wkey[128];
  bool  b;

  create_symbol_btree (&global_symbols);

  // predefined symbols
#if WIN
  insert_symbol (&global_symbols, L"WINDOWS", true);
#else
  insert_symbol (&global_symbols, L"WINDOWS", false);
#endif

#if unix
  insert_symbol (&global_symbols, L"UNIX",    true);
#else
  insert_symbol (&global_symbols, L"UNIX",    false);
#endif

  insert_symbol (&global_symbols, L"MEM32", (bool)(address_size == 4));
  insert_symbol (&global_symbols, L"MEM64", (bool)(address_size == 8));


  for (nth=0; ; nth++)
  {
    if (get_nth_key (ini_filename, section, nth, key, sizeof(key)) < 0 ||
        get_nth_parameter (ini_filename, section, nth, value, sizeof(value)) < 0)
      break;

    string_to_wstring (key, wkey);

    if (!is_all_in_uppercase (wkey))
    {
      sprintf (str, "%.32s : symbol '%.32S' must be all in UPPER CASE",
               ini_filename, wkey);
      config_error (str);
    }

    if (!has_at_least_one_letter (wkey))
    {
      sprintf (str, "%.32s : symbol '%.32S' must contain at least one letter A-Z",
               ini_filename, wkey);
      config_error (str);
    }

    if (strcmp (value, "0") != 0 && strcmp (value, "1") != 0)
    {
      sprintf (str, "%.32s : symbol '%.32S' must have value 0 or 1", ini_filename, wkey);
      config_error (str);
    }

    b = (strcmp (value, "1") == 0);

    rc = insert_symbol (&global_symbols, wkey, b);

    if (rc < 0)
    {
      fatal_out_of_memory_error ("load_global_symbols() : insert_symbol() failed");
    }
    else if (rc == +1)
    {
      sprintf (str, "%.32s : symbol '%.32S' is defined twice", ini_filename, wkey);
      config_error (str);
    }
  }
}

/**********************************************************************************/
