#define BLOCK_H

/* block.h : allocation/deallocation of memory blocks on various devices */

/**********************************************************************/

typedef struct {
  unsigned char kind;   /* 0 = null, 1 = memory, 2 = virtual memory, ... */
  unsigned char filler;
  union {
    char         *pointer;   /* heap handle           */
    long         hnd;        /* virtual memory handle */
    unsigned int whnd;       /* windows handle        */
  } o;
  int           size;    /* size of the block's data */
} BLOCK_HANDLE;

/**********************************************************************/

/* initialize a null handle */
void set_null_handle (BLOCK_HANDLE *handle);

/* check if a handle denotes a null handle */
int is_null_handle (BLOCK_HANDLE *handle);

/* returns 0 if OK, RC_OUT_OF_SPACE if no space left on any device */
int allocate_block (BLOCK_HANDLE *handle,
                    char         *buffer,
                    int          size);

/* if the block's data is larger than buffer_size, it is truncated */
int retrieve_block (BLOCK_HANDLE *handle,
                    char         *buffer,
                    unsigned int buffer_size,
                    int          *size);

/* 'size' must equal exactly the size of the initially allocated block */
int update_block (BLOCK_HANDLE *handle,
                  char         *buffer,
                  int          size);

/* after the call, the handle will be set to null */
int deallocate_block (BLOCK_HANDLE *handle);

/* returns the size of the allocated block.       */
/* an assertion error occurs if 'handle' is null. */
int size_of_block (BLOCK_HANDLE *handle);

/**********************************************************************/
#define RC_OUT_OF_SPACE     (-1)
#define RC_ILLEGAL_HANDLE   (-2)
#define RC_ILLEGAL_SIZE     (-3)
#define RC_DEVICE_ERROR     (-4)
/**********************************************************************/
