
/* calendar.h */

typedef struct {
  unsigned int  year;     /* 1901 to 9999 */
  unsigned char month;    /*    1 to   12 */
  unsigned char day;      /*    1 to   31 */
  unsigned char hour;     /*    0 to   23 */
  unsigned char min;      /*    0 to   59 */
  unsigned char sec;      /*    0 to   59 */
} DATE_TIME;

void get_datetime (DATE_TIME *datetime);

/* assertion: year is in range 1901 .. 9999 */
int max_days_in_month (int month, int year);

/* returns: 1 if date is valid, 0 otherwise. */
/* note: only years in range 1901 .. 9999 are considered as valid */
int is_valid_date (int day, int month, int year);

/* returns: 1=Monday, ..., 7=Sunday */
/* assertion: year is in range 1901 .. 9999 */
int day_of_week (int day, int month, int year);

/* returns number of days since 1/1/1901.   */
/* assertion: year is in range 1901 .. 9999 */
long nb_days_since_1901 (int day, int month, int year);

/* add 'days' to date; 'days' can be negative. */
void add_days (DATE_TIME *date, long days);

/* add 'secs' to date; 'secs' can be negative. */
void add_seconds (DATE_TIME *date, long secs);

/* returns the time offset of the local computer compared to GMT time.     */
/* the value is expressed in minutes and can be in range -12*60 .. +12*60. */
/* the value can fluctuate +/-60 min. during summer/winter time in Europe. */
/* some examples: London = 0, Brussels = +60 or +120, Mexico = -6*60.      */
int get_time_zone (void);
