
/* crc.c */

#include <stdio.h>
#include <assert.h>
#include <malloc.h>
#include "crc.h"

/************************************************************************/

void update_adler (unsigned long *adler,
                   unsigned char *buf,
                   int           len)
{
  unsigned long s1 = (*adler) & 0xffff;
  unsigned long s2 = ((*adler) >> 16) & 0xffff;

  while (len--)
  {
                           /* s1 in range 0 .. 65520 */
    s1 += (*buf);
                           /* s1 in range 0 .. 65520 + 255 */
    if (s1 >= 65521)
      s1 -= 65521;
                           /* s1, s2 in range 0 .. 65520 */
    s2 += s1;
                           /* s2 in range 0 .. 65520 + 65520 */
    if (s2 >= 65521)
    {
      s2 -= 65521;
      if (s2 >= 65521)
        s2 -= 65521;
    }
                           /* s2 in range 0 .. 65520 */

    buf++;
  }

  (*adler) = (s2 << 16) + s1;
}

/************************************************************************/

void update_crc (unsigned long *crc,
                 unsigned char *buf,
                 int           len)
{
  static unsigned long *crc_table = NULL;/* ptr to 256 unsigned long values */
  unsigned long        c;
  int                  n, k;

  if (crc_table == NULL)
  {
    crc_table = malloc (256 * sizeof(unsigned long));
    assert (crc_table != NULL);

    for (n=0; n<256; n++)
    {
      c = (unsigned long)n;

      for (k=0; k<8; k++)
      {
        if (c & 1)
          c = 0xedb88320L ^ (c >> 1);
        else
          c = c >> 1;
      }
      crc_table[n] = c;
    }
  }

  c = (*crc) ^ 0xffffffffL;

  for (n=0; n<len; n++)
    c = crc_table[(c ^ buf[n]) & 0xff] ^ (c >> 8);

  (*crc) = c ^ 0xffffffffL;
}

/***********************************************************************/
