
/* crc.h */

/************************************************************************/

/* Updates an Adler-32 checksum with the bytes buf[0..len-1] */
/* The Adler-32 checksum must be initialized to 1L (!!!!!).  */

void update_adler (unsigned long *adler,
                   unsigned char *buf,
                   int           len);

/************************************************************************/

/* Updates a CRC checksum with the bytes buf[0..len-1] */
/* The CRC checksum must be initialized to 0L (!!!!!). */

void update_crc (unsigned long *crc,
                 unsigned char *buf,
                 int           len);

/************************************************************************/

/* note: the Adler-32 should be preferred as it is faster, safer */
/*       and uses less memory than the CRC.                      */

/************************************************************************/

/*
 Examples:

 unsigned long adler = 1L;
 while (read_buffer (buffer, length) != EOF)
   update_adler (&adler, buffer, length);
 if (adler != original_adler) error();

 unsigned long crc = 0L;
 while (read_buffer (buffer, length) != EOF)
   update_crc (&crc, buffer, length);
 if (crc != original_crc) error();
*/

/************************************************************************/
