
/* files.c : portable file access */

#include <stdio.h>
#include <fcntl.h>
#include <io.h>

#ifdef MSDOS
 #include <dos.h>
 #include <ctype.h>
 #include <direct.h>
#endif

#ifdef _WIN32
 #include <windows.h>
 #include <io.h>
#endif

#ifdef unix
 #include <sys/types.h>
 #include <sys/stat.h>
 #include <dirent.h>
 #include <unistd.h>
 #include <limits.h>
#endif

#include <string.h>
#include <malloc.h>
#include <assert.h>
#include "files.h"

/**************************************************************************/

#ifdef MSDOS
  static struct find_t info;     /* must be kept from a call to the next */
#endif

/**************************************************************************/

int open_directory (FILE_INFO *file_info, char *directory_path)
{
#ifdef MSDOS
  unsigned rc;

  rc = _dos_findfirst (directory_path, _A_RDONLY | _A_HIDDEN | _A_SYSTEM | _A_SUBDIR, &info);
  if (rc == 0)
  {
    strncpy (file_info->name, info.name, sizeof(file_info->name));
    file_info->name[sizeof(file_info->name)-1] = (char)0;

    if (info.attrib == _A_SUBDIR)                    /* a directory */
      file_info->type = TYPE_DIRECTORY;
    else
      file_info->type = TYPE_REGULAR_FILE;

    memset (&file_info->time, 0x00, sizeof(FILE_TIME));
    file_info->size = 0;
    file_info->size_high = 0;

    return 0;
  }
  else
  {
    return RC_NO_MORE_FILES;
  }
#endif

#ifdef _WIN32
  WIN32_FIND_DATA info;
  HANDLE          h;
  DWORD           rc;
  FILETIME        localtime;
  SYSTEMTIME      systemtime;

  h = FindFirstFile (directory_path, &info);
  if (h != INVALID_HANDLE_VALUE)                /* the function succeeded */
  {
    strncpy (file_info->name, info.cFileName, sizeof(file_info->name));
    file_info->name[sizeof(file_info->name)-1] = (char)0;

    if (info.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)  /* a directory */
      file_info->type = TYPE_DIRECTORY;
    else
      file_info->type = TYPE_REGULAR_FILE;

    memset (&file_info->time, 0x00, sizeof(FILE_TIME));
    if (FileTimeToLocalFileTime (&info.ftLastWriteTime, &localtime) == TRUE &&
        FileTimeToSystemTime (&localtime, &systemtime) == TRUE)
    {
      file_info->time.year   = (unsigned int)systemtime.wYear;
      file_info->time.month  = (unsigned char)systemtime.wMonth;
      file_info->time.day    = (unsigned char)systemtime.wDay;
      file_info->time.hour   = (unsigned char)systemtime.wHour;
      file_info->time.min    = (unsigned char)systemtime.wMinute;
      file_info->time.sec    = (unsigned char)systemtime.wSecond;
    }

    file_info->size = info.nFileSizeLow;
    file_info->size_high = info.nFileSizeHigh;

    file_info->handle = (int)h;

    return 0;
  }
  else
  {
    rc = GetLastError ();

    file_info->handle = (int)INVALID_HANDLE_VALUE;

    switch (rc)
    {
      case ERROR_FILE_NOT_FOUND:
        return RC_NO_MORE_FILES;

      default:
        return RC_SYSTEM_ERROR;
    }
  }
#endif

#ifdef unix
  DIR           *dp;
  struct dirent *info;
  char          buffer[MAX_FILENAME_LENGTH];
  struct stat   st;

  strcpy (buffer, directory_path);
  if (strlen(buffer) > 1 && buffer[strlen(buffer)-1] == '/')   /* it's not "/" but ending with '/' */
    buffer[strlen(buffer)-1] = 0;                              /* remove trailing '/'              */

  dp = opendir (buffer);
  if (dp != (DIR *)0)
  {
    info = readdir (dp);

    if (info == (struct dirent *)0)
    {
      closedir (dp);
      return RC_NO_MORE_FILES;
    }

    strncpy (file_info->name, info->d_name, sizeof(file_info->name));
    file_info->name[sizeof(file_info->name)-1] = (char)0;

    if (stat (buffer, &st) < 0)
      file_info->type = 0;
    else if (st.st_mode & S_IFDIR)           /* directory */
      file_info->type = TYPE_DIRECTORY;
    else if (st.st_mode & S_IFREG)           /* regular file */
      file_info->type = TYPE_REGULAR_FILE;
    else
      file_info->type = TYPE_DEVICE;

    memset (&file_info->time, 0x00, sizeof(FILE_TIME));
    file_info->size = 0;
    file_info->size_high = 0;

    file_info->handle = (char)dp;

    return 0;
  }
  else              /* something went wrong */
  {
    return RC_SYSTEM_ERROR;
  }
#endif
}

/**************************************************************************/

int read_directory (FILE_INFO *file_info)
{
#ifdef MSDOS
  unsigned rc;

  rc = _dos_findnext (&info);
  if (rc == 0)
  {
    strncpy (file_info->name, info.name, sizeof(file_info->name));
    file_info->name[sizeof(file_info->name)-1] = (char)0;

    if (info.attrib == _A_SUBDIR)                    /* a directory */
      file_info->type = TYPE_DIRECTORY;
    else
      file_info->type = TYPE_REGULAR_FILE;

    memset (&file_info->time, 0x00, sizeof(FILE_TIME));
    file_info->size = 0;
    file_info->size_high = 0;

    return 0;
  }
  else
  {
    return RC_NO_MORE_FILES;
  }
#endif

#ifdef _WIN32
  WIN32_FIND_DATA info;
  DWORD           rc;
  FILETIME        localtime;
  SYSTEMTIME      systemtime;

  if (FindNextFile ((HANDLE)file_info->handle, &info) == TRUE)       /* the function succeeded */
  {
    strncpy (file_info->name, info.cFileName, sizeof(file_info->name));
    file_info->name[sizeof(file_info->name)-1] = (char)0;

    if (info.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)       /* a directory */
      file_info->type = TYPE_DIRECTORY;
    else
      file_info->type = TYPE_REGULAR_FILE;

    memset (&file_info->time, 0x00, sizeof(FILE_TIME));
    if (FileTimeToLocalFileTime (&info.ftLastWriteTime, &localtime) == TRUE &&
        FileTimeToSystemTime (&localtime, &systemtime) == TRUE)
    {
      file_info->time.year   = (unsigned int)systemtime.wYear;
      file_info->time.month  = (unsigned char)systemtime.wMonth;
      file_info->time.day    = (unsigned char)systemtime.wDay;
      file_info->time.hour   = (unsigned char)systemtime.wHour;
      file_info->time.min    = (unsigned char)systemtime.wMinute;
      file_info->time.sec    = (unsigned char)systemtime.wSecond;
    }

    file_info->size = info.nFileSizeLow;
    file_info->size_high = info.nFileSizeHigh;

    return 0;
  }
  else
  {
    rc = GetLastError ();
    switch (rc)
    {
      case ERROR_NO_MORE_FILES:
        return RC_NO_MORE_FILES;

      default:
        return RC_SYSTEM_ERROR;
    }
  }
#endif

#ifdef unix
  struct dirent *info;
  struct stat   st;

  info = readdir ((DIR *)file_info->handle);
  if (info == (struct dirent *)0)
    return RC_NO_MORE_FILES;

  strncpy (file_info->name, info->d_name, sizeof(file_info->name));
  file_info->name[sizeof(file_info->name)-1] = (char)0;

  if (stat (buffer, &st) < 0)
    file_info->type = 0;
  else if (st.st_mode & S_IFDIR)           /* directory */
    file_info->type = TYPE_DIRECTORY;
  else if (st.st_mode & S_IFREG)           /* regular file */
    file_info->type = TYPE_REGULAR_FILE;
  else
    file_info->type = TYPE_DEVICE;

  memset (&file_info->time, 0x00, sizeof(FILE_TIME));
  file_info->size = 0;
  file_info->size_high = 0;

  return 0;
#endif
}

/**************************************************************************/

int close_directory (FILE_INFO *file_info)
{
#ifdef MSDOS
  return 0;
#endif

#ifdef _WIN32
  if (file_info->handle == (int)INVALID_HANDLE_VALUE) /* already closed */
    return 0;

  if (FindClose ((HANDLE)file_info->handle) == TRUE)
    return 0;

  return RC_SYSTEM_ERROR;
#endif

#ifdef unix
  if (closedir ((DIR *)file_info->dir) == 0)
    return 0;
  else
    return RC_SYSTEM_ERROR;
#endif
}

/**************************************************************************/

#ifndef O_BINARY
  #define O_BINARY  0
#endif

int copy_file (char *source, char *target)
{
  int  rc, in, out, size;
  char buffer[64*1024];

  in = _open (source, O_RDONLY|O_BINARY);
  if (in < 0)
    return -1;

  out = _open (target, O_RDWR|O_CREAT|O_TRUNC|O_BINARY, 0660);
  if (out < 0)
  {
    _close (in);
    return -1;
  }

  for (;;)
  {
    rc = _read (in, buffer, sizeof(buffer));
    if (rc < 0)
    {
      _close (in);
      _close (out);
      _unlink (target);
      return -1;
    }

    size = rc;

    rc = _write (out, buffer, size);
    if (rc != size)
    {
      _close (in);
      _close (out);
      _unlink (target);
      return -1;
    }

    if (size < sizeof(buffer))
      break;
  }

  if (_close (in) < 0)
  {
    _close (out);
    _unlink (target);
    return -1;
  }

  if (_close (out) < 0)
  {
    _unlink (target);
    return -1;
  }

  return 0;
}

/**************************************************************************/

/* filename  ::= [device] [pathname]                  */
/* device    ::= drive-letter: | \\computer | (empty) */
/* pathname  ::= {[/|\] [name]}                       */

static void split_filename (char *filename,   /* in  */
                            char *device,     /* out */
                            char *pathname)   /* out */
{
  int i;

  if (filename[0] != '\0' && filename[1] == ':')        /* drive letter */
  {
    sprintf (device, "%.2s", filename);
    strcpy (pathname, &filename[2]);
  }
  else if (filename[0] == '\\' && filename[1] == '\\')  /* computer name */
  {
    i = 2;
    while (filename[i] != '\0' && filename[i] != '\\')
      i++;
    memcpy (device, filename, i);
    device[i] = '\0';
    strcpy (pathname, &filename[i]);
  }
  else
  {
    device[0] = '\0';
    strcpy (pathname, filename);
  }
}

/**************************************************************************/

/* filename  ::= [device] [pathname]                        */
/* device    ::= drive-letter: | \\computer | (empty)       */
/* pathname  ::= {[/|\] [name]}                             */
/*                                                          */
/* replace "\" by "/" except first one after \\computer     */
/* replace "//" by "/".                                     */
/* normalize ".", "..", "...", .. path components.          */
/*                                                          */
/* the device parameter is provided for read-only.          */

static void normalize_device_pathname (char *device,     /* in     */
                                       char *pathname)   /* in out */
{
  char *s, *t;

  /* replace all '\' by '/' */

  for (s=pathname; *s!='\0'; s++)
  {
    if (*s == '\\')
      *s = '/';
  }


  /* remove duplicate '/' */

  for (s=pathname,t=pathname; *s!='\0'; )
  {
    if (*s == '/')
    {
      *t++ = *s++;
      while (*s == '/')    /* swallow duplicate '/' */
        s++;
    }
    else
    {
      *t++ = *s++;
    }
  }
  *t = '\0';


  /* normalize ".", "..", "...", .. path components */

  {
    char *p, *start;
    char *begin_relative_path, *previous_component;
    int  len, only_dots, count;

    p = pathname;

    if (*p == '/')     /* skip separator (in case of absolute path) */
      p++;

    begin_relative_path = p;  /*no change of the filename before this point*/

    for (;;)
    {
      start = p;      /* start of path component */

      /* parse component name */

      only_dots = 1;
      while (*p != '\0' && *p != '/')
      {
        if (*p != '.')
          only_dots = 0;
        p++;
      }

      len = p - start;

      if (len == 0)   /* no more components */
      {
        break;
      }
      else if (only_dots)    /* "." or ".." or "...", .. */
      {
        /* we must swallow this component and the (len-1) previous ones */

        previous_component = start;
        count = len;

        for (;;)
        {
          if (previous_component == begin_relative_path)
            break;

          previous_component--;

          if (*previous_component == '/')
          {
            count--;
            if (count == 0)
            {
              previous_component++;   /* skip '/' */
              break;
            }
          }
        }

        if (*p == '\0')   /* path ends with dot */
        {
          strcpy (previous_component, ".");
          break;
        }
        else              /* "/" */
        {
          memmove (previous_component, &p[1], strlen(&p[1])+1);
          p = previous_component;
        }
      }
      else   /* normal path component */
      {
        if (*p == '\0')
          break;
        p++;               /* skip '/' */
      }
    }
  }


  /* restore '\' for a device that is a computer name */

  if (device[0] == '\\' && pathname[0] == '/')
    pathname[0] = '\\';
}

/**************************************************************************/

/* compute a filename from :                                      */
/* - a current directory,                                         */
/* - an absolute or relative filename.                            */
/* normalizes the filename.                                       */
/* note: source and result filenames can denote the same buffer.  */
/* returns 0 if OK, -1 if the filenames are too large.            */

int expand_pathname (char *current_directory,
                     char *source_filename,
                     char *result_filename,
                     int  size_of_result_filename)
{
  int  current_length, source_length, result_length;
  char *current_device, *current_pathname;
  char *source_device,  *source_pathname;
  char *result_device,  *result_pathname;

  current_length = strlen(current_directory) + 2;
  source_length  = strlen(source_filename)   + 2;
  result_length  = current_length + source_length;

  current_device   = alloca (current_length);
  current_pathname = alloca (current_length);
  source_device    = alloca (source_length);
  source_pathname  = alloca (source_length);
  result_device    = alloca (result_length);
  result_pathname  = alloca (result_length);

  if (current_device == NULL || current_pathname == NULL ||
      source_device  == NULL || source_pathname  == NULL ||
      result_device  == NULL || result_pathname  == NULL)
  {
    return -1;
  }

  split_filename (current_directory, current_device, current_pathname);
  split_filename (source_filename,   source_device,  source_pathname);

  if (source_device[0] != '\0')
    strcpy (result_device, source_device);
  else
    strcpy (result_device, current_device);

  if (source_pathname[0]  == '/'  ||
      source_pathname[0]  == '\\' ||
      current_pathname[0] == '\0')
    strcpy (result_pathname, source_pathname);
  else
    sprintf (result_pathname, "%s/%s", current_pathname, source_pathname);

  normalize_device_pathname (result_device, result_pathname);

  if (strlen(result_device) + strlen(result_pathname) + 1 > (unsigned)size_of_result_filename)
    return -1;

  sprintf (result_filename, "%s%s", result_device, result_pathname);

  return 0;
}

/************************************************************************/

/* filename  ::= [device] [pathname]                        */
/* device    ::= drive-letter: | \\computer | (empty)       */
/* pathname  ::= {[/|\] [name]}                             */
/*                                                          */
/* replace "\" by "/" except first one after \\computer     */
/* replace "//" by "/".                                     */
/* normalize ".", "..", "...", .. path components.          */

void normalize_pathname (char *pathname)
{
  (void)expand_pathname ("", pathname, pathname, strlen(pathname) + 2);
}

/************************************************************************/

#ifdef _WIN32

int get_file_size (int           fd,
                   unsigned long *low,
                   unsigned long *high)
{
  int   h;
  DWORD lo, hi;

  h = _get_osfhandle (fd);
  if (h == -1)
    return -1;

  lo = GetFileSize ((HANDLE)h, &hi);
  if (lo == INVALID_FILE_SIZE && GetLastError() != NO_ERROR)
    return -1;

  *low  = lo;
  *high = hi;

  return 0;
}

#endif

/**************************************************************************/

#ifdef _WIN32

int get_file_pos (int           fd,
                  unsigned long *low,
                  unsigned long *high)
{
  long  h;
  DWORD lo, hi;

  h = _get_osfhandle (fd);
  if (h == -1)
    return -1;

  hi = 0;
  lo = SetFilePointer ((HANDLE)h, 0, &hi, FILE_CURRENT);
  if (lo == (unsigned long)-1L && GetLastError() != NO_ERROR)
    return -1;

  *low = lo;
  *high = hi;

  return 0;

}

#endif

/**************************************************************************/

#ifdef _WIN32

int set_file_pos (int           fd,
                  unsigned long low,
                  unsigned long high)
{
  long  h;
  DWORD lo, hi;

  h = _get_osfhandle (fd);
  if (h == -1)
    return -1;

  hi = high;
  lo = SetFilePointer ((HANDLE)h, low, &hi, FILE_BEGIN);
  if (lo == (unsigned long)-1L && GetLastError() != NO_ERROR)
    return -1;

  return 0;
}

/**************************************************************************/

/* add signed 'offset' to file position (low,high) */

void add_file_pos (unsigned long *low,
                   unsigned long *high,
                   long          offset)
{
  unsigned long low_before;

  low_before = *low;
  *low += (unsigned long)offset;

  if (offset >= 0)
  {
    if (*low < low_before)   /* has become smaller : overflow occured */
      (*high)++;
  }
  else
  {
    if (*low > low_before)   /* has become larger : underflow occured */
      (*high)--;
  }
}

/**************************************************************************/

/* add unsigned filesize to file position (low,high) */

void add_file_size (unsigned long *low,
                    unsigned long *high,
                    unsigned long size_low,
                    unsigned long size_high)
{
  unsigned long before;

  before = *low;
  *low += size_low;

  if (*low < before)   /* has become smaller : overflow occured */
    (*high)++;

  *high += size_high;
}

#endif

/**************************************************************************/

static int _intern_name_match_pattern (char *name,
                                       char *pattern,
                                       char replace_one,
                                       char replace_many,
                                       int  name_first,
                                       int  name_last,
                                       int  pattern_first,
                                       int  pattern_last)
{
  int i;

  while (name_first <= name_last && pattern_first <= pattern_last)
  {
    if ((pattern[pattern_first] == name[name_first] &&
         pattern[pattern_first] != replace_many)
        || pattern[pattern_first] == replace_one)
    {
      /* eliminate common first characters */
      name_first++;
      pattern_first++;
    }
    else if ((pattern[pattern_last] == name[name_last] &&
              pattern[pattern_last] != replace_many)
             || pattern[pattern_last] == replace_one)
    {
      /* eliminate common last characters */
      name_last--;
      pattern_last--;
    }
    else if (pattern[pattern_first] == replace_many)
    {
      /* suppress all 'replace many' characters */

      while (pattern_first <= pattern_last &&
             pattern[pattern_first] == replace_many)
      {
        pattern_first++;
      }

      for (i=name_last+1; i>=name_first; i--)
      {
        if (_intern_name_match_pattern (name, pattern,
                                        replace_one, replace_many,
                                        i, name_last,
                                        pattern_first, pattern_last))
        {
          return 1;  /* TRUE */
        }
      }

      return 0;  /* FALSE */
    }
    else   /* in all other cases there's no match */
    {
      return 0;  /* FALSE */
    }
  }

  /* here, either the name or the pattern is empty, or both */

  /* suppress all 'replace many' characters */

  while (pattern_first <= pattern_last &&
         pattern[pattern_first] == replace_many)
  {
    pattern_first++;
  }

  return (name_first > name_last && pattern_first > pattern_last);
}

/**************************************************************************/

int name_matches_pattern (char *name,
                          char *pattern,
                          char replace_one,     /* should be '?' */
                          char replace_many)    /* should be '*' */
{
  return _intern_name_match_pattern (name, pattern,
                                     replace_one, replace_many,
                                     0, strlen(name)-1,
                                     0, strlen(pattern)-1);
}

/**************************************************************************/

#ifdef _WIN32

int get_disk_free_space (char          *directory,
                         unsigned long *free_low,
                         unsigned long *free_high,
                         unsigned long *total_low,
                         unsigned long *total_high)
{
  FARPROC pGetDiskFreeSpaceEx;

  struct {
    unsigned long low, high;
  } total64, free64;

  pGetDiskFreeSpaceEx = GetProcAddress (GetModuleHandle ("kernel32.dll"),
                                        "GetDiskFreeSpaceExA");
  if (pGetDiskFreeSpaceEx == NULL)
    return -2;

  if ((*pGetDiskFreeSpaceEx)(directory,
                             &free64, &total64, NULL) == FALSE)
    return -1;

  *free_low   = free64.low;
  *free_high  = free64.high;
  *total_low  = total64.low;
  *total_high = total64.high;

  return 0;
}

#endif

/**************************************************************************/

#ifdef _WIN32

/* get date/time of last write access */

int get_file_datetime (int       fd,
                       FILE_TIME *datetime)
{
  long       h;
  FILETIME   ft;
  SYSTEMTIME st;

  memset (datetime, 0x00, sizeof(FILE_TIME));

  h = _get_osfhandle (fd);
  if (h == -1)
    return -1;

  if (GetFileTime ((HANDLE)h, NULL, NULL, &ft) == FALSE)
    return -1;

  if (FileTimeToLocalFileTime (&ft, &ft) == FALSE)
    return -1;

  if (FileTimeToSystemTime (&ft, &st) == FALSE)
    return -1;

  datetime->year  = st.wYear;
  datetime->month = (unsigned char)st.wMonth;
  datetime->day   = (unsigned char)st.wDay;
  datetime->hour  = (unsigned char)st.wHour;
  datetime->min   = (unsigned char)st.wMinute;
  datetime->sec   = (unsigned char)st.wSecond;

  return 0;
}

#endif

/**************************************************************************/
