
/* inifile.h : read .ini file */

/************************************************************************/

/* retrieve parameter from .ini file                     */
/* returns 0 or a negative error code.                   */
/* in case of error, the default value (if any) is used. */

int get_parameter (char         *filename,        /* param.ini     */
                   char         *section,         /* section name  */
                   char         *key,             /* key name      */
                   char         *parameter,       /* out buffer    */
                   unsigned int parameter_size,   /* sizeof buffer */
                   char         *default_value);  /* can be NULL   */

/************************************************************************/

/* retrieve nth section of the file.   */
/* returns 0 or a negative error code. */

int get_nth_section (char         *filename,        /* ini filename  */
                     int          nth,              /* 0 .. ?        */
                     char         *section,         /* out buffer    */
                     unsigned int section_size);    /* sizeof buffer */

/************************************************************************/

/* retrieve nth key within given section of the file. */
/* returns 0 or a negative error code.                */

int get_nth_key (char         *filename,  /* ini filename  */
                 char         *section,   /* section name  */
                 int          nth,        /* 0 .. ?        */
                 char         *key,       /* out buffer    */
                 unsigned int key_size);  /* sizeof buffer */

/************************************************************************/

/* retrieve nth parameter within given section of the file. */
/* returns 0 or a negative error code.                      */

int get_nth_parameter (char         *filename,  /* ini filename  */
                       char         *section,   /* section name  */
                       int          nth,        /* 0 .. ?        */
                       char         *parameter, /* out buffer    */
                       unsigned int parameter_size);  /* sizeof buffer */

/************************************************************************/

/* signal that initialization file was changed and must be reloaded. */

void signal_ini_file_changed (void);

/************************************************************************/

#define E_INI_FILE_NOT_FOUND    (-1)  /* cannot open .ini file */
#define E_INI_SECTION_NOT_FOUND (-2)
#define E_INI_KEY_NOT_FOUND     (-3)
#define E_INI_BUFFER_TOO_SMALL  (-4)  /* output value was truncated   */
#define E_INI_BAD_INDEX         (-5)  /* parameter 'nth' out of range */

/************************************************************************/

/* format of .ini file :

# this is a comment !

[section1]
key1 = parameter1
key2 = parameter2   ; comment

[section2]
key1 =             ; empty
test = a;;b        ; value "a;b"

Notes:
.  max line length in the file is 4096 chars.
. '#' or ';' characters are used to start a comment.
. '#' or ';' characters must be doubled within a parameter value
             to indicate the corresponding single character.
. keys are case insensitive, parameter values are not.

*/

/************************************************************************/
