
/* pack.h : packing/unpacking using Deflate+Huffman Algorithm */

/***********************************************************************/

/* user-provided read/write function.           */
/* must return the number of bytes transferred, */
/* zero in case of end-of-data (read),          */
/* or a negative value in case of error.        */

typedef int IO_FUNC (void          *user_info,
                     unsigned char *buffer,
                     unsigned int  size);

/***********************************************************************/

#define UNPACK_MAX_READ_AHEAD  32768

typedef struct {
  unsigned int  size;      /* nb bytes in 'buffer' */
  unsigned char buffer[UNPACK_MAX_READ_AHEAD];
} READ_AHEAD;

/***********************************************************************/

/* 'user_info' can be used to pass data to the user's i/o functions */

int pack (void *user_info, IO_FUNC *user_read, IO_FUNC *user_write);

/***********************************************************************/

/* 'user_info' can be used to pass data to the user's i/o functions. */

/* note: unpack() reads data in large blocks and will probably read  */
/*       past the compression part. The data read but not used by    */
/*       unpack can be returned in the structure 'extra'.            */
/*       'extra' can also be set to NULL if the data is not wanted.  */

int unpack (void *user_info, IO_FUNC *user_read, IO_FUNC *user_write,
            READ_AHEAD *extra);

/***********************************************************************/

/* pack error codes: */
#define PK_READ_ERROR   (-1)       /* error in user_read */
#define PK_WRITE_ERROR  (-2)       /* error in user_write */
#define PK_STACK_ERROR  (-3)       /* out of stack space */
#define PK_INTERN_1     (-4)       /* intern error (should never occur) */

/***********************************************************************/

/* unpack error codes: */
#define UNPK_READ_ERROR      (-1)       /* error in user_read */
#define UNPK_WRITE_ERROR     (-2)       /* error in user_write */
#define UNPK_STACK_ERROR     (-3)       /* out of stack space */
#define UNPK_BAD_BTYPE       (-4)       /* bad field BTYPE */
#define UNPK_BAD_NLEN        (-5)       /* bad field NLEN */
#define UNPK_HUFF_OVERFLOW   (-6)       /* bad format */
#define UNPK_HUFF_BAD_STRUCT (-7)       /* bad format */
#define UNPK_HUFF_BAD_CODE   (-8)       /* bad format */
#define UNPK_NO_PREV_CODE    (-9)       /* bad format */
#define UNPK_TOO_MANY       (-10)       /* bad format */
#define UNPK_BAD_CODE       (-11)       /* bad format */

/***********************************************************************/
