
/* pack3.c : output bit stream */

#define DEBUG_COUNT  0     /* printf nb of bits output */

/**************************************************************************/

#if DEBUG_COUNT
  static long global_bit_count;
#endif

/**************************************************************************/

static void init_output (OUTPUT_INFO *info,
                         IO_FUNC     *user_write,
                         void        *user_info)  /* passed to user_write() */
{
  info->count      = 0;
  info->by         = 0;
  info->power      = 1;
  info->user_write = user_write;
  info->user_info  = user_info;

#if DEBUG_COUNT
  global_bit_count = 0L;
#endif
}

/**************************************************************************/

static int send_code (OUTPUT_INFO *info, int code, int nb_bits, int bit_order)
{
  int           i, power, rc;
  unsigned char bit;

  assert (nb_bits >= 1 && nb_bits <= 15);
  assert (code >= 0 && code < (1 << nb_bits));

  if (bit_order == LSB_FIRST)
    power = 1;
  else if (bit_order == MSB_FIRST)
    power = (1 << (nb_bits-1));
  else
    assert (0);

  for (i=0; i<nb_bits; i++)
  {
    bit = ((code & power) != 0);

    /* send 'bit' to output stream */

    if (bit)
      info->by |= info->power;
    info->power <<= 1;

    if (info->power > 128)
    {
      if (info->count == OUTPUT_BUFFER_SIZE)  /* buffer full */
      {
        rc = (*info->user_write) (info->user_info, info->buffer, info->count);
        if (rc != info->count)
          return PK_WRITE_ERROR;

        info->count = 0;
      }

      info->buffer[info->count++] = info->by;

      info->by = 0;
      info->power = 1;
    }

    /* advance to next power */

    if (bit_order == LSB_FIRST)
      power <<= 1;
    else
      power >>= 1;

#if DEBUG_COUNT
  global_bit_count++;
#endif
  }

  return 0;
}

/**************************************************************************/

/* flush the remaining bits (if any) of the current byte       */
/* in order to be on a byte boundary, then send the data block */

static int send_block (OUTPUT_INFO *info, unsigned char *data, int length)
{
  int rc, i;

  assert (length >= 0);

  /* flush remaining bits of current byte (if any) */

  while (info->power != 1)
  {
    rc = send_code (info, 0, 1, MSB_FIRST);
    if (rc)
      return rc;

#if DEBUG_COUNT
  global_bit_count--;
#endif
  }

  for (i=0; i<length; i++)
  {
    if (info->count == OUTPUT_BUFFER_SIZE)  /* buffer full */
    {
      rc = (*info->user_write) (info->user_info, info->buffer, info->count);
      if (rc != info->count)
        return PK_WRITE_ERROR;

      info->count = 0;
    }

    info->buffer[info->count++] = data[i];

#if DEBUG_COUNT
  global_bit_count+=8;
#endif
  }

  return 0;
}

/**************************************************************************/

/* send all remaining data to output device */

static int flush_output (OUTPUT_INFO *info)
{
  int rc;

  /* flush remaining bits of current byte (if any) */

  while (info->power != 1)
  {
    rc = send_code (info, 0, 1, MSB_FIRST);
    if (rc)
      return rc;

#if DEBUG_COUNT
  global_bit_count--;
#endif
  }

  if (info->count != 0)    /* buffer contains something */
  {
    rc = (*info->user_write) (info->user_info, info->buffer, info->count);
    if (rc != info->count)
      return PK_WRITE_ERROR;

    info->count = 0;
  }

#if DEBUG_COUNT
  printf ("global_bit_count = %ld\n", global_bit_count);
#endif
  return 0;
}

/**************************************************************************/
