
/* unpack0.c */

#define DEBUG  0

/**************************************************************************/

/* the maximum backward distance of a reference: */
#define CHUNK_SIZE   32768      /* must be a power of 2 !!! */

#define LOOK_AHEAD     256

#define METHOD_UNCOMPRESSED 0
#define METHOD_FIX_HUFF     1
#define METHOD_DYN_HUFF     2

#define MAX_NODES_A  286   /* literal and lengths */
#define MAX_NODES_B   30   /* distance */
#define MAX_NODES_C   19   /* code_len */

static short alpha_tab[MAX_NODES_C] =
  {16, 17, 18, 0, 8, 7, 9, 6, 10, 5, 11, 4, 12, 3, 13, 2, 14, 1, 15};

static struct {
  short extra_bits;
  short min_length;
} length_info [29] =
{{0,   3},     /* 4 */
 {0,   4},
 {0,   5},
 {0,   6},
 {0,   7},     /* 4 */
 {0,   8},
 {0,   9},
 {0,  10},
 {1,  11},    /* 8 */
 {1,  13},
 {1,  15},
 {1,  17},
 {2,  19},    /* 16 */
 {2,  23},
 {2,  27},
 {2,  31},
 {3,  35},    /* 32 */
 {3,  43},
 {3,  51},
 {3,  59},
 {4,  67},    /* 64 */
 {4,  83},
 {4,  99},
 {4, 115},
 {5, 131},   /* 127 !! */
 {5, 163},
 {5, 195},
 {5, 227},
 {0, 258}};  /*   1 */


static struct {
  short extra_bits;
  int   min_distance;
} distance_info [30] =
{{ 0,     1 },
 { 0,     2 },
 { 0,     3 },
 { 0,     4 },
 { 1,     5 },
 { 1,     7 },
 { 2,     9 },
 { 2,    13 },
 { 3,    17 },
 { 3,    25 },
 { 4,    33 },
 { 4,    49 },
 { 5,    65 },
 { 5,    97 },
 { 6,   129 },
 { 6,   193 },
 { 7,   257 },
 { 7,   385 },
 { 8,   513 },
 { 8,   769 },
 { 9,  1025 },
 { 9,  1537 },
 {10,  2049 },
 {10,  3073 },
 {11,  4097 },
 {11,  6145 },
 {12,  8193 },
 {12, 12289 },
 {13, 16385 },
 {13, 24577 }};

/**************************************************************************/

int unpack (void *user_info, IO_FUNC *user_read, IO_FUNC *user_write,
            READ_AHEAD *extra)
{
  INPUT_INFO    in;
  OUTPUT_INFO   out;
  int           final, method, rc;
  unsigned char output [CHUNK_SIZE];
  int           output_i = 0;


  /* init input/output modules */
  init_input (&in, user_read, user_info);
  init_output (&out, user_write, user_info);

  for (;;)
  {
    /* is it the final block ? */
    final = receive_code (&in, 1);
    if (final < 0)
      return final;

#if DEBUG
    printf ("BFINAL : %d\n", final);
#endif

    /* get method */
    method = receive_code (&in, 2);
    if (method < 0)
      return method;

#if DEBUG
    printf ("BTYPE : %d\n", method);
#endif

    switch (method)
    {
      case METHOD_UNCOMPRESSED:
        {
          unsigned char  len[4];
          unsigned short length, nlength, cmp;


          /* receive LEN/NLEN */
          rc = receive_block (&in, len, 4);
          if (rc)
            return rc;

          length  = len[0] + (((unsigned short)len[1]) << 8);
          nlength = len[2] + (((unsigned short)len[3]) << 8);

#if DEBUG
    printf ("LENGTH : 0x%04X (NLEN=0x%04X)\n", length, nlength);
#endif

          /* check validity of NLEN */
          cmp = ~nlength;       /* avoid converting to int ! */
          if (length != cmp)
            return UNPK_BAD_NLEN;


          {
            unsigned char  buf[256];
            unsigned short chunk, i;

            while (length)
            {
              chunk = sizeof(buf);
              if (length < chunk)
                chunk = length;

              rc = receive_block (&in, buf, chunk);
              if (rc)
                return rc;

              rc = send_block (&out, buf, chunk);
              if (rc)
                return rc;

              for (i=0; i<chunk; i++)
              {
                output[output_i] = buf[i];
                output_i = (output_i + 1) & (CHUNK_SIZE - 1);
              }

              length -= chunk;
            }
          }
        }
        break;


      case METHOD_FIX_HUFF:
      case METHOD_DYN_HUFF:
        {
          NODE node_a[2 * MAX_NODES_A];  /* info for literals and lengths */
                                         /* + place for intermediate huffman nodes */
          NODE node_b[2 * MAX_NODES_B];  /* info for distances */
                                         /* + place for intermediate huffman nodes */

          if (method == METHOD_DYN_HUFF)
          {
            int   count_a, count_b, hclen, i, rc;

            short code_alpha[MAX_NODES_C];    /* 'hclen' values */

            NODE  node_c[2 * MAX_NODES_C];  /* info for code_lengths */
                                            /* + place for intermediate huffman nodes */

            unsigned char code_length[MAX_NODES_A + MAX_NODES_B];
            short         count_code_length, max_count_code_length;


            /* input HLIT (5 bits) / HDIST (5 bits) / HCLEN (4 bits) */

            count_a = receive_code (&in, 5);
            if (count_a < 0)
              return count_a;

            count_b = receive_code (&in, 5);
            if (count_b < 0)
              return count_b;

            hclen = receive_code (&in, 4);
            if (hclen < 0)
              return hclen;

            count_a += 257;
            count_b += 1;
            hclen   += 4;

#if DEBUG
  printf ("HLIT=%d HDIST=%d HCLEN=%d\n", count_a, count_b, hclen);
#endif

            /* read-in alpha table */

#if DEBUG
  printf ("alpha table : ");
#endif
            /* make sure unused codes are set to zero */
            memset (code_alpha, 0x00, sizeof(code_alpha));

            for (i=0; i<hclen; i++)
            {
              rc = receive_code (&in, 3);
              if (rc < 0)
                return rc;
              code_alpha[i] = (short)rc;
#if DEBUG
  printf ("%d ", rc);
#endif
            }

#if DEBUG
  printf ("\n");
#endif

            /* convert back to node_c[] using alpha_tab[] */

            memset (node_c, 0x00, sizeof(node_c));
            for (i=0; i<MAX_NODES_C; i++)
              node_c[alpha_tab[i]].code_length = code_alpha[i];


            /* build huffman tree */

            rc = compute_huffman (node_c, MAX_NODES_C);
            if (rc)
              return rc;


            /* read-in the code lengths for node_a[] and node_b[] */

            count_code_length = 0;
            max_count_code_length = count_a + count_b;

            {
              int previous_code, code, code_len, repeat;

              previous_code = -1;

              while (count_code_length < max_count_code_length)
              {
                /* read-in a code length */
                rc = receive_huffman_code (&in, node_c, MAX_NODES_C, &code);
                if (rc)
                  return rc;

                assert (code >= 0 && code <= 18);

                if (code <= 15)
                {
                  code_len = code;
                  repeat = 1;
                }
                else if (code == 16)
                {
                  rc = receive_code (&in, 2);
                  if (rc < 0)
                    return rc;

                  code_len = previous_code;
                  repeat = 3 + rc;
                }
                else if (code == 17)
                {
                  rc = receive_code (&in, 3);
                  if (rc < 0)
                    return rc;

                  code_len = 0;
                  repeat = 3 + rc;
                }
                else if (code == 18)
                {
                  rc = receive_code (&in, 7);
                  if (rc < 0)
                    return rc;

                  code_len = 0;
                  repeat = 11 + rc;
                }
                else
                {
                  assert (0);
                }

#if DEBUG
  if (code < 16)
    printf ("RLE CODE %d\n", code);
  else
    printf ("RLE CODE %d (repeat = %d)\n", code, repeat);
#endif

                if (code_len == -1)
                  return UNPK_NO_PREV_CODE;

                previous_code = code_len;

                if (count_code_length + repeat > max_count_code_length)
                  return UNPK_TOO_MANY;

                while (repeat--)
                  code_length [count_code_length++] = code_len;
              }
            }


            /* copy the code lengths into node_a[] and node_b[] */

            {
              int i;

              memset (node_a, 0x00, sizeof(node_a));
              memset (node_b, 0x00, sizeof(node_b));

              for (i=0; i<count_a; i++)
                node_a[i].code_length = code_length[i];

              for (i=0; i<count_b; i++)
                node_b[i].code_length = code_length[count_a + i];
            }


            /* build huffman trees */

            rc = compute_huffman (node_a, MAX_NODES_A);
            if (rc)
              return rc;

            rc = compute_huffman (node_b, MAX_NODES_B);
            if (rc)
              return rc;
          }
          else     /* method == METHOD_FIX_HUFF */
          {
            int i;

            for (i=0; i<=143; i++)
            {
              node_a[i].code_length = 8;
              node_a[i].code_value  = 0x30 + (i - 0);
            }

            for (i=144; i<=255; i++)
            {
              node_a[i].code_length = 9;
              node_a[i].code_value  = 0x190 + (i - 144);
            }

            for (i=256; i<=279; i++)
            {
              node_a[i].code_length = 7;
              node_a[i].code_value  = 0x00 + (i - 256);
            }

            for (i=280; i<MAX_NODES_A; i++)
            {
              node_a[i].code_length = 8;
              node_a[i].code_value  = 0xC0 + (i - 280);
            }

            for (i=0; i<MAX_NODES_B; i++)
            {
              node_b[i].code_length = 5;
              node_b[i].code_value  = i;
            }

            rc = build_huffman_tree (node_a, MAX_NODES_A);
            if (rc)
              return rc;

            rc = build_huffman_tree (node_b, MAX_NODES_B);
            if (rc)
              return rc;
          }


          /* receive the actual data */

          for (;;)
          {
            int code, extra_bits, length, distance;

            /* read-in a length code */
            rc = receive_huffman_code (&in, node_a, MAX_NODES_A, &code);
            if (rc)
              return rc;

            if (code == 256)     /* end of block */
              break;

            if (code <= 255)     /* a literal */
            {
#if DEBUG
              printf ("literal : %d (%c)\n", code, isprint(code) ? code : '.');
#endif
              {
                unsigned char ch;

                ch = (unsigned char)code;

                output [output_i] = ch;
                output_i = (output_i + 1) & (CHUNK_SIZE - 1);

                rc = send_block (&out, &ch, 1);
                if (rc)
                  return rc;
              }
            }
            else                 /* a length */
            {
              /* read the length code suffix (if any) */

              assert (code >= 257 && code < MAX_NODES_A);

              length = length_info [code-257] . min_length;

              extra_bits = length_info[code-257].extra_bits;
              if (extra_bits)
              {
                rc = receive_code (&in, extra_bits);
                if (rc < 0)
                  return rc;

                if (code == 284 && rc == 31)   /* not allowed value */
                  return UNPK_BAD_CODE;

                length += rc;
              }


              /* read-in a distance code */

              rc = receive_huffman_code (&in, node_b, MAX_NODES_B, &code);
              if (rc)
                return rc;

              /* read the distance code suffix (if any) */

              assert (code >= 0 && code < MAX_NODES_B);

              distance = distance_info [code] . min_distance;

              extra_bits = distance_info[code].extra_bits;
              if (extra_bits)
              {
                rc = receive_code (&in, extra_bits);
                if (rc < 0)
                  return rc;
                distance += rc;
              }


              /* length (3..258); distance (1..32768) */
#if DEBUG
              printf ("reference=(%d,%d)\n", -distance, length);
#endif

              {
                int           from, i;
                unsigned char ch;

                from = output_i - distance;
                if (from < 0)
                  from += CHUNK_SIZE;

                if (from     + length >= CHUNK_SIZE ||   /* source clip */
                    output_i + length >= CHUNK_SIZE ||   /* target clip */
                    length > distance)                   /* overlapping */
                {
                  /* use slow byte-wise copy */

                  for (i=0; i<length; i++)
                  {
                    ch = output[from];
                    from = (from + 1) & (CHUNK_SIZE - 1);

                    output [output_i] = ch;
                    output_i = (output_i + 1) & (CHUNK_SIZE - 1);

                    rc = send_block (&out, &ch, 1);
                    if (rc)
                      return rc;
                  }
                }
                else    /* optimized version : copy entire block */
                {
                  rc = send_block (&out, &output[from], length);
                  if (rc)
                    return rc;

                  memcpy (&output[output_i], &output[from], length);
                  output_i += length;
                  from     += length;
                }
              }
            }
          }
        }
        break;


      default:
        return UNPK_BAD_BTYPE;
    }

    if (final)    /* was this the final block ? */
      break;
  }

  rc = flush_output (&out);
  if (rc)
    return rc;

  if (extra != NULL)
    copy_to_read_ahead (&in, extra);

  return 0;
}

/**************************************************************************/
