
/* unpack3.h */

/**************************************************************************/

#define INPUT_BUFFER_SIZE 32768  /* must match with UNPACK_MAX_READ_AHEAD ! */

typedef struct {
  unsigned char buffer[INPUT_BUFFER_SIZE];
  int           first;   /* index of byte currently being read bit per bit */
  int           count;   /* nb of bytes present in buffer[] */
  int           power;   /* power of 2 that will be extracted from   */
                         /* buffer[first] next time. when power == 1 */
                         /* was extracted, 'first' is incremented.   */
  IO_FUNC       *user_read;  /* function used to read a buffer */
  void          *user_info;
} INPUT_INFO;

/**************************************************************************/

static void init_input (INPUT_INFO *info,
                        IO_FUNC    *user_read,
                        void       *user_info); /* passed to user_read() */

/**************************************************************************/

/* returns a fixed-length code in LSB order, or a negative error. */
/* nb_bits must not exceed 15.                                    */

static int receive_code (INPUT_INFO *info, int nb_bits);

/**************************************************************************/

/* skip remaining bits of current byte and receive 'length' bytes */

static int receive_block (INPUT_INFO    *info,
                          unsigned char *buf,
                          unsigned int  length);

/**************************************************************************/

/* analyzes a varying-length bit sequence in MSB order.     */
/* returns 0 and the corresponding huffman code in '*code'  */
/* or an error code if it does not exist.                   */

static int receive_huffman_code (INPUT_INFO *info,
                                 NODE       *node,      /* Huffman tree  */
                                 int        max_nodes,  /* nb leaf nodes */
                                 int        *code);     /* result code   */

/**************************************************************************/

static void copy_to_read_ahead (INPUT_INFO *info,
                                READ_AHEAD *extra);

/**************************************************************************/
