
// symbols.c

#include <stdio.h>
#include <string.h>
#include <malloc.h>

#include "bintree.h"

#include "types.h"
#include "strings.h"

#include "symbols.h"

/**********************************************************************************/

static int compare (void               *user,
                    const SYMBOL_ENTRY *a,
                    const unsigned int size1,
                    const SYMBOL_ENTRY *b,
                    const unsigned int size2)
{
  return wstrcmp (a->symbol, b->symbol);
}

/**********************************************************************************/

void create_symbol_btree (BINARY_TREE *tree)
{
  create_btree (tree, NULL, compare);
}

/**********************************************************************************/

// returns 0 if OK, -1 if error, +1 if already exists

int insert_symbol (BINARY_TREE *tree, wstring s, bool value)
{
  SYMBOL_ENTRY e;
  int          rc;

  e.symbol = malloc ((wstrlen(s) + 1) * 2);
  if (e.symbol == NULL)
    return -1;

  wstrcpy (e.symbol, s);
  e.value = value;

  rc = insert_btree (tree, &e, sizeof(e));
  if (rc == 0)
    return 0;

  free (e.symbol);

  if (rc == BT_DUPLICATE_KEY)
    return +1;

  return -1;
}

/**********************************************************************************/

static int operate (void               *user,
                    const SYMBOL_ENTRY *data,
                    const unsigned int size)
{
  free (data->symbol);
  return 0;
}

/**********************************************************************************/

void free_symbol_btree (BINARY_TREE *tree)
{
  traverse_btree (tree, operate, +1);
  close_btree (tree);
}

/**********************************************************************************/

// returns 0 if OK, -1 if error, +1 if not found

int get_symbol_value (BINARY_TREE *tree, wstring s, bool *presult)
{
  SYMBOL_ENTRY e;
  int          rc;
  unsigned int dummy;
  wchar        *p;

  p = malloc ((wstrlen(s) + 1) * 2);
  if (p == NULL)
    return -1;

  wstrcpy (p, s);

  e.symbol = p;
  e.value = false;

  rc = retrieve_btree (tree, &e, sizeof(e), sizeof(e), &dummy, BT_EQUAL);

  free (p);

  if (rc == 0)
  {
    *presult = e.value;
    return 0;
  }
  else
  {
    *presult = false;
    if (rc == BT_KEY_NOT_FOUND)
      return +1;
    return -1;
  }
}

/**********************************************************************************/
