
// exception.c

use calendar, thread, files, strings, system, win/windows;

bool _gf_create_crash_report_file;
bool _gf_display_fatal_error_message_box;

#begin unsafe

//----------------------------------------------------------------------------------------

const string CRASH_REPORT = "CRASH-REPORT.TXT";

uint g_var;

struct EXCEPTION_TYPE
{
  uint   ExceptionCode;
  string msg;
}

const EXCEPTION_TYPE exception_type[] =
  {{0x80000001, "GUARD_PAGE_VIOLATION"},
   {0x80000002, "DATATYPE_MISALIGNMENT"},
   {0xC0000005, "ACCESS_VIOLATION"},
   {0xC0000017, "OUT_OF_HEAP_MEMORY"},
   {0xC000001D, "ILLEGAL_INSTRUCTION"},
   {0xC0000094, "DIVIDE_BY_ZERO"},
   {0xC0000096, "PRIV_INSTRUCTION"},
   {0xC00000FD, "STACK_OVERFLOW"}};

//----------------------------------------------------------------------------------------

struct RIP_REPEAT_INFO
{
  int     count, index, repeat;
  INT_PTR history[128];
}

//----------------------------------------------------------------------------------------

package Read_DBG

  packed struct Couple
  {
    uint a;
    uint b;
  }

  Couple cpl;
  uint   v;
  char   ch;
  int    fd;

  void r();
  void gv();
  void gch();

end Read_DBG;

//----------------------------------------------------------------------------------------

package body Read_DBG

  public void r()
  {
    if (read (fd, out cpl) != (int)cpl'size)
      cpl.b = 0x80000000;
  }

  public void gv()
  {
    if (read (fd, out v) != (int)v'size)
      v = 0;
  }

  public void gch()
  {
    if (read (fd, out ch) != (int)ch'size)
      ch = nul;
  }

end Read_DBG;

//----------------------------------------------------------------------------------------

void retrieve_location (long ip, out char[32] location, out int line_nr)
{
  char filename[MAX_EXECUTABLE_FILENAME_LENGTH];
  uint len, unit, line, current_unit, i;
  bool found;

  clear location;
  line_nr = 0;

  get_executable_filename (out filename);
  len = (uint)strlen(filename);
  if (len < 4)
    return;

  filename[len-4] = nul;
  strcat (ref filename, ".dbg");

  Read_DBG.fd = open (filename);
  if (Read_DBG.fd < 0)
    return;

  r();

  if (cpl.a != 0xF2761287 || cpl.b != 0xD5016423)   // check header
    return;

  r();
  found = false;
  unit = 0;
  line = 0;
  current_unit = 0;

  while (cpl.b != 0x80000000)
  {
    if (!found)
    {
      if (cpl.a == 0)  // new unit (file)
        current_unit = cpl.b;
      else if (ip >= cpl.a && cpl.b != 0)
      {
        unit = current_unit;
        line = cpl.b;
        r();
        if (ip < cpl.a)   // ip between 2 records
          found = true;
        else
          continue;
      }
    }
    r();
  }

  if (!found)
  {
    close (Read_DBG.fd);
    return;
  }

  clear filename;

  gv();
  while (v != 0)
  {
    current_unit = v;

    gv();  // get length of library
    len = v;

    for (i=0; i<len; i++)
      gch();

    gv();  // get length of source file
    len = v;

    for (i=0; i<len; i++)
    {
      gch();
      if (current_unit == unit && i < (uint)filename'length)
        filename[i] = ch;
    }

    gv();
  }

  close (Read_DBG.fd);

  {
    int length, offset;

    // cut filename
    length = strlen(filename);
    offset = length - location'length;
    if (offset < 0)   // short filename
    {
      offset = 0;
    }
    else   // long filename
    {
      offset = length - location'length;
      length = location'length;
    }

    sprintf (out location, "%s", filename[offset : length]);
  }

  line_nr = (int)line;
}

//----------------------------------------------------------------------------------------

void trace_newline (int fd)
{
  write (fd, "\r\n");
}

//----------------------------------------------------------------------------------------

void trace_register_32 (int fd, char[3] register_name, int value)
{
  char[32]   str;
  char[32+2] str2;

  sprintf (out str, "%s=%8x", register_name, value);
  if (value != 0)
    strcatf (ref str, "   (= %d)", value);
  sprintf (out str2, "  %-32.32s", str);

  write (fd, str2);
}

//----------------------------------------------------------------------------------------

void trace_register (int fd, char[3] register_name, long value)
{
  char[48]   str;
  char[48+4] str2;
  long v = value;

  if (value >= 0 && value <= uint'max)  // seems int4 value zero-extended to int8
    v = (int)v;  // sign-extend

  sprintf (out str, "%s=%16x", register_name, value);
  if (v != 0)
    strcatf (ref str, "   (= %d)", v);
  sprintf (out str2, "    %-48.48s", str);

  write (fd, str2);
}

//----------------------------------------------------------------------------------------

void load_module_name (long ip, out char[32] location)
{
  HANDLE        hModuleSnap;
  MODULEENTRY32 me32;

  clear location;

  hModuleSnap = CreateToolhelp32Snapshot (TH32CS_SNAPMODULE | TH32CS_SNAPMODULE32, 0);
  if (hModuleSnap == INVALID_HANDLE_VALUE)
    return;

  me32.dwSize = MODULEENTRY32'size;

  if (Module32First (hModuleSnap, &me32) == 0)
  {
    CloseHandle (hModuleSnap);
    return;
  }

  for (;;)
  {
    INT_PTR base;
    base'byte = me32.modBaseAddr'byte;

    if (ip >= base && ip < base + (long)me32.modBaseSize)
    {
      strncpy (out location, me32.szModule, location'length);
      CloseHandle (hModuleSnap);
      return;
    }

    if (Module32Next (hModuleSnap, &me32) == 0)
      break;
  }

  CloseHandle (hModuleSnap);
}

//----------------------------------------------------------------------------------------

void list_all_DLLs (int fd)
{
  HANDLE        hModuleSnap;
  MODULEENTRY32 me32;

  hModuleSnap = CreateToolhelp32Snapshot (TH32CS_SNAPMODULE | TH32CS_SNAPMODULE32, 0);
  if (hModuleSnap == INVALID_HANDLE_VALUE)
    return;

  me32.dwSize = MODULEENTRY32'size;

  if (Module32First (hModuleSnap, &me32) == 0)
  {
    CloseHandle (hModuleSnap);
    return;
  }

  trace_newline (fd);
  write (fd, "Memory modules:\r\n");

  for (;;)
  {
    char str[512];
    INT_PTR base;
    base'byte = me32.modBaseAddr'byte;

#if MEM32
    sprintf (out str, "  %8x to %8x %s\r\n", base, base + (long)me32.modBaseSize, me32.szModule);
#else
    sprintf (out str, "%16x to %16x %s\r\n", base, base + me32.modBaseSize, me32.szModule);
#endif

    write (fd, str[0:strlen(str)]);

    if (Module32Next (hModuleSnap, &me32) == 0)
      break;
  }

  CloseHandle (hModuleSnap);
}

//----------------------------------------------------------------------------------------

bool in_main_module (INT_PTR address)
{
  byte*   data_ptr = (byte*)&_gf_create_crash_report_file;  // address of some global
  INT_PTR last;

  last'byte = data_ptr'byte;

  return address >= 0x401000 && address < last;
}

//----------------------------------------------------------------------------------------

void compute_address_details (    INT_PTR  address,
                              out char[32] location,
                              out int      line_nr)
{
  if (in_main_module (address))
    retrieve_location (address, out location, out line_nr);
  else
  {
    load_module_name (address, out location);
    line_nr = 0;
  }
}

//----------------------------------------------------------------------------------------

void build_short_address_details (    INT_PTR  address,
                                      char[32] location,
                                      int      line_nr,
                                  out char[64] buffer)
{
  if (line_nr > 0)   // translated by .dbg file
    sprintf (out buffer, "%s:%d", location, line_nr);
  else if (location[0] > nul)  // in DLL
    sprintf (out buffer, "%x (%s)", address, location);
  else   // untranslated
    sprintf (out buffer, "%x", address);
}

//----------------------------------------------------------------------------------------

void build_medium_address_details (    INT_PTR  address,
                                       char[32] location,
                                       int      line_nr,
                                   out char[80] buffer)
{
  if (line_nr > 0)
    sprintf (out buffer, "%x  %s line %d", address, location, line_nr);
  else if (location[0] > nul)
    sprintf (out buffer, "%x (%s)", address, location);
  else
    sprintf (out buffer, "%x", address);
}

//----------------------------------------------------------------------------------------

void build_large_address_details (    INT_PTR  address,
                                      char[32] location,
                                      int      line_nr,
                                  out char[80] buffer)
{
  if (line_nr > 0)
  {
    #if MEM32
      sprintf (out buffer, "  %8x  %s line %d", address, location, line_nr);
    #else
      sprintf (out buffer, "%16x  %s line %d", address, location, line_nr);
    #endif
  }
  else if (location[0] > nul)
  {
    #if MEM32
      sprintf (out buffer, "  %8x  (%s)", address, location);
    #else
      sprintf (out buffer, "%16x  (%s)", address, location);
    #endif
  }
  else
  {
    #if MEM32
      sprintf (out buffer, "  %8x", address);
    #else
      sprintf (out buffer, "%16x", address);
    #endif
  }
}

//----------------------------------------------------------------------------------------

void output_rip (INT_PTR rip, int fd)
{
  char[32] location;
  int      line_nr;
  char[80] buffer;

  compute_address_details (rip, out location, out line_nr);
  build_large_address_details (rip, location, line_nr, out buffer);

  write (fd, buffer[0:strlen(buffer)]);
  trace_newline (fd);
}

//----------------------------------------------------------------------------------------

void terminate_repeating_rip (    int             fd,
                              ref RIP_REPEAT_INFO r)
{
  if (r.repeat > 1)
  {
    char[128] line_buffer;

    if (r.repeat == 2 && r.count == 1)
    {
      output_rip (r.history[0], fd);
    }
    else
    {
      strcpy (out line_buffer, "    (previous ");

      if (r.count == 1)
        strcat (ref line_buffer, "line");
      else
        strcatf (ref line_buffer, "%d lines", r.count);

      strcatf (ref line_buffer, " repeated %d time", r.repeat-1);

      if (r.repeat > 2)
        strcat (ref line_buffer, "s");

      strcat (ref line_buffer, ")");

      write (fd, line_buffer[0:strlen(line_buffer)]);
      trace_newline (fd);
    }
  }

  if (r.index > 0)  // last part was not fully repeated : output the remaining lines
  {
    int k;
    for (k=0; k<r.index; k++)
      output_rip (r.history[k], fd);
  }
}

//----------------------------------------------------------------------------------------

// display rip or record it in repeat structure

void display_or_record_rip (    INT_PTR         rip,
                                int             fd,
                            ref int             callstack_lines_displayed,
                            ref RIP_REPEAT_INFO r)
{
  if (r.repeat == 0)   // no recursive call history active
  {
    int j;

    // check if repeated rip
    for (j=0; j<r.count; j++)
    {
      if (rip == r.history[j])
        break;
    }

    if (j == r.count)   // not found
    {
      // store rip
      if (r.count == r.history'length)
      {
        r.history[0 : r.history'length-1] = r.history[1 : r.history'length-1];
        r.count--;
      }
      r.history[r.count++] = rip;

      output_rip (rip, fd);
    }
    else   // found a repeating rip : go in repeat mode
    {
      r.count -= j;
      r.history [0 : r.count] = r.history [j : r.count];
      r.index = 1;
      r.repeat = 1;

      if (r.count == 1)  // it's 1 line repeat sequence, it was already repeated twice now
      {
        r.index = 0;
        r.repeat = 2;
        callstack_lines_displayed--;
      }
    }
  }
  else  // we are repeating a sequence of recursive calls
  {
    if (rip == r.history [r.index])  // match
    {
      r.index++;

      if (r.index == r.count)  // whole sequence repeated
      {
        r.index = 0;
        r.repeat++;
        callstack_lines_displayed -= r.count;
      }
    }
    else  // sequence does not match anymore : we need to produce some output
    {
      terminate_repeating_rip (fd, ref r);

      // leave repeat mode
      clear r;

      // output the new sequence
      output_rip (rip, fd);

      // store as new initial history
      r.history[r.count++] = rip;
    }
  }
}

//----------------------------------------------------------------------------------------

void handle_the_exception (EXCEPTION_POINTERS e)
{
  // we have a guard page (4K) of stack space available in case of exception STACK OVERFLOW.
  char[MAX_EXECUTABLE_FILENAME_LENGTH]      exe_filename;
  char[MAX_EXECUTABLE_FILENAME_LENGTH + 32] crash_report_filename;
  char[128]                                 str, dialog_box_buffer, line_buffer;

  RIP_REPEAT_INFO          rip_info;  // 1K
  MEMORY_BASIC_INFORMATION mem;

  ref EXCEPTION_RECORD  r = *e.ExceptionRecord;
  ref CONTEXT           c = *e.ContextRecord;

  int              i, callstack_lines_displayed;
  INT_PTR          rbp, rip;
  uint             pu*, u;
  int              fd;
  DATE_TIME        now;
  char             *pstr;
  bool             dialog_full, crash_in_dll;

  // report only the first crash
  if (InterlockedExchange (ref g_var, 1) != 0)
    sleep 86400;  // other threads wait forever

  get_executable_filename (out exe_filename);
  strcpy (out crash_report_filename, exe_filename);
  i = strrchr (crash_report_filename, '\\');
  crash_report_filename[i+1] = nul;
  strcat (ref crash_report_filename, CRASH_REPORT);

  if (_gf_create_crash_report_file)
  {
    fd = open (crash_report_filename, READ+WRITE, 0);
    if (fd > 0)   // exists
    {
      lseek (fd, 0L, SEEK_END);
    }
    else  // new file
    {
      fd = create (crash_report_filename, WRITE, 0);
    }
  }
  else
  {
    fd = -1;
  }

  get_datetime (out now);
  sprintf (out str, "crash date   : %02d/%02d/%04d %02d:%02d:%02d\r\napplication  : ", now.day, now.month, now.year, now.hour, now.min, now.sec);
  write (fd, str[0:strlen(str)]);

  write (fd, exe_filename[0:strlen(exe_filename)]);

  write (fd, "\r\ncommand line : ");
  pstr = GetCommandLineA();
  write (fd, pstr[0:strlen(pstr[0:260])]);

  rip = 0x400088;
  pu'byte = rip'byte;
  u = *pu;

  now.day = 1;
  now.month = 1;
  now.year = 1970;
  add_days (ref now, (int)(u / 86400));
  u %= 86400;
  now.hour = (tiny)(u / 3600);
  u -= 3600 * (uint)now.hour;
  now.min = (tiny)(u / 60);
  u -= 60 * (uint)now.min;
  now.sec = (tiny)u;
  sprintf (out str, "\r\ncompile date : %02d/%02d/%04d %02d:%02d:%02d\r\n\r\n", now.day, now.month, now.year, now.hour, now.min, now.sec);
  write (fd, str[0:strlen(str)]);

  for (i=0; i<exception_type'length; i++)
  {
    if (exception_type[i].ExceptionCode == r.ExceptionCode)
      break;
  }

  if (i < exception_type'length)
    sprintf (out str, "%s", exception_type[i].msg);
  else
    sprintf (out str, "Ex %08x", r.ExceptionCode);

#if MEM32
  strcat (ref str, " at ");
#else
  strcat (ref str, " at ");
#endif

  strcpy (out dialog_box_buffer, str);
  strcpy (out line_buffer, str);

#if MEM32
  strcat (ref line_buffer, "EIP=");
#else
  strcat (ref line_buffer, "RIP=");
#endif

  {
    char[32] location;
    int      line_nr;
    char[64] short_buffer;
    char[80] medium_buffer;

    compute_address_details (r.ExceptionAddress, out location, out line_nr);
    build_short_address_details (r.ExceptionAddress, location, line_nr, out short_buffer);
    build_medium_address_details (r.ExceptionAddress, location, line_nr, out medium_buffer);

    crash_in_dll = (location[0] != nul) && (line_nr == 0);

    strcat (ref dialog_box_buffer, short_buffer);
    strcat (ref line_buffer, medium_buffer);
  }

  write (fd, line_buffer[0:strlen(line_buffer)]);
  trace_newline (fd);

  if (r.ExceptionCode == 0xC0000005)  // ACCESS_VIOLATION
  {
    sprintf (out line_buffer, "  attempted ");

    switch ((int)r.ExceptionInformation[0])
    {
      case 0:
        strcat (ref line_buffer, "READ");
        break;

      case 1:
        strcat (ref line_buffer, "WRITE");
        break;

      case 8:
        strcat (ref line_buffer, "EXECUTE");
        break;

      default:
        strcat (ref line_buffer, "ACCESS");
        break;
    }

    strcatf (ref line_buffer, " at address %x\r\n", r.ExceptionInformation[1]);
    write (fd, line_buffer[0:strlen(line_buffer)]);
  }
  trace_newline (fd);

  write (fd, "Call stack:\r\n");

  // determine range of valid stack pages
  VirtualQuery (c.Rsp, &mem, mem'size);

  dialog_full = false;
  clear rip_info;

  rbp = c.Rbp;

  callstack_lines_displayed = 0;

  if (crash_in_dll)   // find some intermediate return addresses
  {
    INT_PTR Rsp = c.Rsp;

    if (rbp < mem.BaseAddress || rbp >= mem.BaseAddress + (INT_PTR)mem.RegionSize)
      rbp = mem.BaseAddress + (INT_PTR)mem.RegionSize - (int)(2*Rsp'size);

    #if MEM64
      Rsp = ((Rsp - 8) & -16) | 8;  // align at 16 + 8
    #endif

    {
      const string GUESS = "    --- begin guessed ips ---";
      write (fd, GUESS);
      trace_newline (fd);
    }

    for (; callstack_lines_displayed<1024; callstack_lines_displayed++)
    {
      INT_PTR p*, RetAdr;

      if (Rsp + (int)Rsp'size > mem.BaseAddress + (INT_PTR)mem.RegionSize)
        break;

      if (Rsp >= rbp)
        break;

      p'byte = Rsp'byte;
      RetAdr = p[0] - 2;      // possible return address

      if (in_main_module (RetAdr))
        display_or_record_rip (RetAdr, fd, ref callstack_lines_displayed, ref rip_info);

      #if MEM32
        Rsp += 4;
      #else
        Rsp += 16;
      #endif
    }

    {
      const string GUESS = "    --- end guessed ips ---";
      write (fd, GUESS);
      trace_newline (fd);
    }
  }

  for (; callstack_lines_displayed<1024; callstack_lines_displayed++)
  {
    INT_PTR *p;

    if (rbp < mem.BaseAddress || rbp+(int)(2*rbp'size) > mem.BaseAddress + (INT_PTR)mem.RegionSize)
      break;

    p'byte = rbp'byte;

    rbp = p[0];      // load old rbp
    rip = p[1] - 2;  // load return address,  -2 because we don't want the IP after the call

    if (!dialog_full)
    {
      char[32] location;
      int      line_nr;
      char[64] short_buffer;

      compute_address_details (rip, out location, out line_nr);
      build_short_address_details (rip, location, line_nr, out short_buffer);

      if (strlen(dialog_box_buffer) + 3 + strlen(short_buffer) < dialog_box_buffer'length)
      {
        strcat (ref dialog_box_buffer, ", ");
        strcat (ref dialog_box_buffer, short_buffer);
      }
      else
      {
        dialog_full = true;
      }
    }

    display_or_record_rip (rip, fd, ref callstack_lines_displayed, ref rip_info);
  }

  terminate_repeating_rip (fd, ref rip_info);

  trace_newline (fd);
  write (fd, "Registers:\r\n");

#if MEM32
  trace_register_32 (fd, "EAX", c.Rax);
  trace_register_32 (fd, "RSI", c.Rsi);
  trace_newline (fd);

  trace_register_32 (fd, "EBX", c.Rbx);
  trace_register_32 (fd, "RDI", c.Rdi);
  trace_newline (fd);

  trace_register_32 (fd, "ECX", c.Rcx);
  trace_register_32 (fd, "RBP", c.Rbp);
  trace_newline (fd);

  trace_register_32 (fd, "EDX", c.Rdx);
  trace_register_32 (fd, "RSP", c.Rsp);
  trace_newline (fd);
#else
  trace_register (fd, "RAX", c.Rax);
  trace_register (fd, "R8 ", c.R8);
  trace_newline (fd);

  trace_register (fd, "RBX", c.Rbx);
  trace_register (fd, "R9 ", c.R9);
  trace_newline (fd);

  trace_register (fd, "RCX", c.Rcx);
  trace_register (fd, "R10", c.R10);
  trace_newline (fd);

  trace_register (fd, "RDX", c.Rdx);
  trace_register (fd, "R11", c.R11);
  trace_newline (fd);

  trace_register (fd, "RSI", c.Rsi);
  trace_register (fd, "R12", c.R12);
  trace_newline (fd);

  trace_register (fd, "RDI", c.Rdi);
  trace_register (fd, "R13", c.R13);
  trace_newline (fd);

  trace_register (fd, "RBP", c.Rbp);
  trace_register (fd, "R14", c.R14);
  trace_newline (fd);

  trace_register (fd, "RSP", c.Rsp);
  trace_register (fd, "R15", c.R15);
  trace_newline (fd);
#endif

  if (_gf_create_crash_report_file)
  {
    flush (fd);
    list_all_DLLs (fd);
  }

  {
    clear str;
    i = 100;
    str[0:i] = {all=>'-'};
    write (fd, str[0:100]);
    write (fd, "\r\n");
  }

  close (fd);

  if (_gf_display_fatal_error_message_box)
    FatalAppExitA (0, &dialog_box_buffer);     // !! does not support a string longer than 260 !!
  else
    ExitProcess (0xFFFFFFFF);
}

//----------------------------------------------------------------------------------------

[callback]
LONG pvectored_exception_handler (EXCEPTION_POINTERS e)
{
  ref EXCEPTION_RECORD r = *e.ExceptionRecord;
  int                  i;

  for (i=0; i<exception_type'length; i++)
  {
    if (exception_type[i].ExceptionCode == r.ExceptionCode)
      handle_the_exception (e);
  }

  // let this exception travel up the chain, otherwise APIs like GetOpenFileNameA() fail.
  return 0;  // EXCEPTION_CONTINUE_SEARCH (0x0)
}

//----------------------------------------------------------------------------------------

public void arm_exception_handler (bool create_crash_report_file        = true,
                                   bool display_fatal_error_message_box = true)
{
  _gf_create_crash_report_file        = create_crash_report_file;
  _gf_display_fatal_error_message_box = display_fatal_error_message_box;

  AddVectoredExceptionHandler (0, pvectored_exception_handler);  // 0 = last handler to be called
}

//----------------------------------------------------------------------------------------

public void log_in_crash_report (string format, object[] arg)
{
  char[MAX_EXECUTABLE_FILENAME_LENGTH]      exe_filename;
  char[MAX_EXECUTABLE_FILENAME_LENGTH + 32] crash_report_filename;
  int                                       i;
  FILE                                      file;

  get_executable_filename (out exe_filename);
  strcpy (out crash_report_filename, exe_filename);
  i = strrchr (crash_report_filename, '\\');
  crash_report_filename[i+1] = nul;
  strcat (ref crash_report_filename, CRASH_REPORT);

  fappend (out file, crash_report_filename, encoding => ANSI);
  fprintf (ref file, format, arg);
  fclose (ref file);
}

//---------------------------------------------------------------------
#end unsafe
