
// guitree.h

use ../edline, ../edtext, ../text, ../win/windows;
use ../gui;

enum CONTROL_TYP {TYP_TEXT, TYP_EDIT, TYP_CHECKBOX, TYP_RADIOBUTTON, TYP_BUTTON, TYP_WINDOW, TYP_LISTBOX, 
   TYP_SCROLL, TYP_TREE, TYP_COMBO, TYP_EDITBOX};

const string[] typ_name = {"a text", "an edit", "a checkbox", "a radiobutton", "a button", "a window", "a listbox",
   "a scroll", "a tree", "a combo", "an editbox"};

// control can be clicked and does not move the dialog window, can be reached with TAB
const bool tabable[1+(int)CONTROL_TYP'last] = {false, true, true, true, true, true, true, true, true, true, true};

//--------------------------------------------------------------------

struct EDIT_INFO
{
  EDIT_LINE    edit_line;
  bool         password_mode;     // false = normal, true = display "***"
  int          scroll_x_offset;
  bool         line_marking_in_progress;
  int          line_marking_start_col;
}

struct CHECKBOX_INFO
{
  int          box_width;
  bool         modify_allowed;
  bool         setting;
}

struct RADIOBUTTON_INFO
{
  int          box_width;
  bool         setting;
  bool         modify_allowed;
  RADIOBUTTON_GROUP_ID gid;
}

struct BUTTON_INFO
{
  bool         pressed;
}

struct WINDOW_INFO
{
  HBITMAP       hbitmap;
  HDC           hdcMemory;
  HBITMAP       old_hbitmap;
  bool          refresh;         // false = no refresh (cam), true = refresh
}

struct LISTBOX_INFO
{
  W_TEXT       wtext;                // format: reference:4, selected(0/1):1, line
  int          page;                 // index of first visible listbox line (1 .. ?)
  int          ln;                   // index of current listbox line (0=none .. ?)
  int          nb_screen_lines;      // lines visible on screen
  int          line_length;          // max characters of a line
  int          line_height;          // height of a line, in pixels
  int          scrollbar;            // scrollbar position (1 .. y_size-1)
  int          arrow_box_width;
  int          arrow_box_height;
  bool         allow_user_selection;
  bool         allow_hotkey_search;
  int          hotkey_position;
  bool         multiple_mode;        // false = single mode, true = multiple mode
  int          selected_line;        // for single mode : the selected line, or 0 for none, 
                                     // for multi mode : the start line of a range, or 0 for none.
  int          clicked_scrollbox_y;  // clicked y for dragging scrollbox
  int          clicked_page;         // current page nr at click
  bool         show_focus;           // only if working with keyboard
}

packed struct PREFIX      // line prefix (14 bytes) MUST BE MULTIPLE OF 2 (WCHAR)
{
  int8  reference;        // user reference
  bool  selected;         // line is selected by user (displayed in reverse)
  bool  is_folder;        // true = contains children, false = single item
  uint2 level;            // tree level (0 = root)
  uint2 type;             // type of icon (for future use)
}

const int PREFIX_WLEN = (int)PREFIX'size / 2;   // nb of wchars to store PREFIX

const int MAX_TREE_LEVELS = 32;   // must be power of 2

bool tree_line_is_expanded (ref W_TEXT wtext, int ln, uint2 level);

enum TSCROLL {HSCROLL, VSCROLL};

struct SCROLL_INFO
{
  TSCROLL      tscroll;
  int          page;                 // index of first visible col line (1 .. ?)
  int          range;                // extern range (#total)
  int          shown;                // intern range (#visible)
  int          arrow_box_width_or_height;  // width for HSCROLL, height for VSCROLL
  int          clicked_scrollbox_c;  // clicked x for dragging hscrollbox, y for dragging vscrollbox
  int          clicked_page;         // current page nr at click
  int          small_y_increment;
}

struct LISTBOX_EXTRA
{
  bool         listbox_shown;
  int          y;             // relative position and
  int          y_size;        // size of listbox
}

struct TREE_INFO
{
  LISTBOX_INFO  listbox;
}

struct COMBO_INFO
{
  EDIT_INFO     edit;
  LISTBOX_INFO  listbox;
  LISTBOX_EXTRA lb;
}

struct EDITBOX_INFO
{
  EDIT_TEXT    text;
  int          nb_full_visible_lines;
  int          page;             // line number of first screen line (>= 1)
  int          font_width;
  int          scroll;           // column of first screen column (>= 0)
  bool         text_marking_in_progress;
  TEXT_MARK    text_marking_start;
}

struct CONTROL_INFO (CONTROL_TYP typ)
{
  CONTROL_ID    id;
  int           x, y;             // relative position and
  int           x_size, y_size;   // size of object
  FONT          font;
  wstring^      text;             // allocated on heap (can be null)
  wchar         hotkey;
  CONTROL_INFO^ prev, next;       // circular list (for TAB/UNTAB)
  bool          hide;             // true = invisible, false = visible
  uint[5]       colors;

  switch (typ)
  {
    case TYP_TEXT:
      null;

    case TYP_EDIT:
      EDIT_INFO edit;

    case TYP_CHECKBOX:
      CHECKBOX_INFO checkbox;

    case TYP_RADIOBUTTON:
      RADIOBUTTON_INFO radiobutton;

    case TYP_BUTTON:
      BUTTON_INFO button;

    case TYP_WINDOW:
      WINDOW_INFO window;

    case TYP_LISTBOX:
      LISTBOX_INFO listbox;

    case TYP_SCROLL:
      SCROLL_INFO scroll;

    case TYP_TREE:
      TREE_INFO tree;

    case TYP_COMBO:
      COMBO_INFO combo;

    case TYP_EDITBOX:
      EDITBOX_INFO editbox;
  }
}

//--------------------------------------------------------------------

struct DIALOG_INFO
{
  DIALOG_ID      d;
  HWND           hwnd;
  DIALOG_HANDLER handler;
  uint           ShowWindowArg;
  int            border_drag_size; // size of border than can be dragged with the mouse to resize the window (default = 0)
  int            border_size;      // visual border size, 0, 1 or 2.
  bool           has_close_button;
  FONT           title_font;                  // font & color of title
  int            title_height;
  uint           dialog_colors[5];            // index 0, 1, 3, 4 used for border, index 2 = title bar background
  uint           dialog_background_color;
  int            dialog_transparency_focus;
  int            dialog_transparency_non_focus;
  FONT           default_font;               // current default font for controls
  uint[5]        default_colors;             // current default colors for controls
  bool           default_insert_mode;        // true = insert, false = delete
  bool           redirect_keyboard_to_main_window;

  // work variables
  RECT           rect;           // absolute screen position of dialog
  int            current_transparency;
  bool           has_keyboard_focus;
  bool           mouse_tracking_active;
  int2           previous_mouse_x, previous_mouse_y;

  // fields for controls
  CONTROL_INFO^  list;           // circular list of controls or null
  CONTROL_INFO^  focus;          // control having the focus or null
}

//--------------------------------------------------------------------

HWND hInstance;
APPLICATION_EVENT_HANDLER app_handler;   // user function that handles main window events
RECT main_win_rect;                      // position and size of main window
int  y_title, x_border, y_border;        // size of borders of main window
bool caret_visible;
int  caret_x, caret_y, caret_height, caret_new_x, caret_new_y, caret_new_height;
bool g_shift_pressed, g_control_pressed, g_alt_pressed, g_middle_mouse_button_pressed;

#begin unsafe
WINDOWPOS*   main_window_windowpos;
WINDOWPOS*   sub_window_windowpos;
DIALOG_INFO* g_dialog_ptr;       // current dialog
DIALOG_INFO* g_new_dialog_ptr;   // next dialog
#end unsafe

bool g_mouse_tracking_active;
int2 g_previous_mouse_x, g_previous_mouse_y;

INT_PTR g_basic_mouse_shape = IDC_ARROW;   // IDC_ARROW or IDC_BEAM
INT_PTR g_current_mouse_shape;             // IDC_ARROW, IDC_BEAM, IDC_NO, IDC_WAIT
HCURSOR g_hcursor_shape;
bool g_drag_selected;
bool g_drag_moved;
bool g_user_drop_allowed;
bool g_hourglass;

#begin unsafe
CONTROL_INFO* g_last_control_left_clicked;
#end unsafe

//--------------------------------------------------------------------
