
// guiutils.c

use ../strings, ../win/windows;
use ../gui, ../edtext;
use guitree;

//--------------------------------------------------------------------

FONT  cached_font;
HFONT hcached_font;

HDC cached_hdc;

//--------------------------------------------------------------------

bool same_font (FONT a, FONT b)
{
  return strcmp (a.name, b.name) == 0 && a.height == b.height && a.style == b.style;
}

//--------------------------------------------------------------------

HFONT create_font (FONT font)
{
  char name[MAX_FONT_NAME_LENGTH];

  strcpy (out name, font.name);
  name[MAX_FONT_NAME_LENGTH-1] = nul;   // font name must never exceed 31 chars + nul

#begin unsafe
  return CreateFontA (font.height,
                      0,             // use matching width
                      0, 0,
                      (font.style & BOLD) != 0 ? FW_BOLD : FW_DONTCARE,
                      (DWORD)((font.style & ITALIC) != 0),
                      (DWORD)((font.style & UNDERLINED) != 0),
                      (DWORD)FALSE, 
                      ANSI_CHARSET, OUT_DEFAULT_PRECIS,
                      CLIP_DEFAULT_PRECIS, DRAFT_QUALITY,
                      DEFAULT_PITCH,
                      &name);
#end unsafe
}

//--------------------------------------------------------------------

public HFONT get_cached_font (FONT font)
{
#begin unsafe
  if (!same_font (font, cached_font))
  {
    if (hcached_font != 0)
      assert DeleteObject (hcached_font) != 0;

    cached_font = font;
    hcached_font = create_font (font);
    assert hcached_font != 0;
  }

  return hcached_font;
#end unsafe
}

//--------------------------------------------------------------------

public HDC get_cached_hdc ()
{
#begin unsafe
  if (cached_hdc == 0)
  {
    HDC hdc = GetDC (0);
    cached_hdc = CreateCompatibleDC (hdc);
    assert cached_hdc != 0;
    ReleaseDC (0, hdc);
  }
  return cached_hdc;
#end unsafe
}

//--------------------------------------------------------------------

public void compute_scrollbox_data
      (int height,           // must be (nb_lines*text_height)
       int scrollbar_arrow_box_height,
       int current_page_nr,
       int nb_listbox_lines,
       int nb_screen_lines,
   out int scrollbox_y_offset,
   out int scrollbox_height)
{
  int  scrollbox_position_num;
  int  scrollbox_position_den;
  int  scrollbox_height_num;
  int  scrollbox_height_den;
  int  move_height;
  int  sh;

  sh = height - 2 * scrollbar_arrow_box_height;

  scrollbox_height_num = nb_screen_lines;
  scrollbox_height_den = nb_listbox_lines;

  if (scrollbox_height_den > 0)
    scrollbox_height = sh * scrollbox_height_num / scrollbox_height_den;
  else
    scrollbox_height = sh;

  if (scrollbox_height < 4)
    scrollbox_height = 4;
  if (scrollbox_height > sh)
    scrollbox_height = sh;

  // compute sum of free area above & below box
  move_height = sh - scrollbox_height;


  scrollbox_position_num = current_page_nr - 1;    // [0..max_page[
  scrollbox_position_den = nb_listbox_lines - nb_screen_lines;

  if (scrollbox_position_den > 0)
    scrollbox_y_offset = move_height * scrollbox_position_num / scrollbox_position_den;
  else
    scrollbox_y_offset = 0;

  if (scrollbox_y_offset < 0)
    scrollbox_y_offset = 0;
  if (scrollbox_y_offset > move_height)
    scrollbox_y_offset = move_height;
}

//--------------------------------------------------------------------------

public void recompute_font_width (ref CONTROL_INFO o)
{
  assert o.typ == TYP_EDITBOX;
  o.editbox.font_width = text_width_of2 ("W", o.font);
}

//--------------------------------------------------------------------------

public void compute_editbox_scroll_field (ref CONTROL_INFO o)
{
  ref EDITBOX_INFO editbox = o . editbox;
  int nb_visible_cols      = o.x_size / editbox.font_width;
  int col                  = edit_text_get_col (editbox.text);

  if (editbox.scroll >= col)
  {
    editbox.scroll = col-1;
    if (editbox.scroll < 0)
      editbox.scroll = 0;
  }

  if (editbox.scroll < col - nb_visible_cols + 1)
    editbox.scroll = col - nb_visible_cols + 1;
}

//--------------------------------------------------------------------------

public void compute_editbox_page (ref CONTROL_INFO o)
{
  ref EDITBOX_INFO editbox = o . editbox;
  int ln                   = edit_text_get_ln (editbox.text);

  editbox.nb_full_visible_lines = o.y_size / o.font.height;
  if (editbox.nb_full_visible_lines == 0)
    editbox.nb_full_visible_lines = 1;

  if (editbox.page > edit_text_count_lines(editbox.text) - editbox.nb_full_visible_lines + 1)  // page too large
    editbox.page = edit_text_count_lines(editbox.text) - editbox.nb_full_visible_lines + 1;

  if (editbox.page < ln - (editbox.nb_full_visible_lines - 1))   // page too small
    editbox.page = ln - (editbox.nb_full_visible_lines - 1);

  if (editbox.page < 1)    // page too small
    editbox.page = 1;
  
  if (editbox.page > ln)   // page too large (must be <= ln)
    editbox.page = ln;
}

//--------------------------------------------------------------------------
