
// thread.c

use strings, win/windows;

//---------------------------------------------------------------
#begin unsafe
//---------------------------------------------------------------

struct SHARED_OBJECT    // full type
{
  CRITICAL_SECTION cs;
  uint  init;    // 0 = new, 1 = initialization in progress, 2 = initialized
}

uint pid;   // current process id

uint granularity;

//---------------------------------------------------------------

// returns old value of variable
public uint InterlockedExchange (ref uint variable,      // ebp+12  RBP+24
                                 uint     new_value)     // ebp+8   RBP+16
{
#if MEM32
  _asm { 0x8B 0x45 0x08 };  // mov   eax,ebp+8   ; load new value
  _asm { 0x8B 0x75 0x0C };  // mov   esi,ebp+12  ; load addr of variable
  _asm { 0x87 0x06 };       // xchg  eax,[esi]   ; exchange with variable
  _asm { 0xC9 };            // leave
  _asm { 0xC2 0x08 0x00 };  // ret 8
#else
  _asm { 0x8B 0x45 0x10 };       // mov   eax,rbp+16  ; load new value
  _asm { 0x48 0x8B 0x75 0x18 };  // mov   rsi,rbp+24  ; load addr of variable
  _asm { 0x87 0x06 };            // xchg  eax,[rsi]   ; exchange with variable
  _asm { 0xC9 };                 // leave
  _asm { 0xC2 0x10 0x00 };       // ret 16
#endif
}

//---------------------------------------------------------------

// adds delta to variable in an indivisible operation,
// even if several threads call this function at the same time.

public void InterlockedAdd (ref int variable,    // ebp+12  RBP+24
                                int delta)       // ebp+8   RBP+16
{
#if MEM32
  _asm { 0x8B 0x45 0x08 };       // mov  eax,ebp+8   ; load new value
  _asm { 0x8B 0x75 0x0C };       // mov  esi,ebp+12  ; load addr of variable
  _asm { 0xF0 0x01 0x06 };       // lock  add  dword ptr [esi],eax
  _asm { 0xC9 };                 // leave
  _asm { 0xC2 0x08 0x00 };       // ret 8
#else
  _asm { 0x8B 0x45 0x10 };       // mov   eax,rbp+16  ; load new value
  _asm { 0x48 0x8B 0x75 0x18 };  // mov   rsi,rbp+24  ; load addr of variable
  _asm { 0xF0 0x01 0x06 };       // lock  add  dword ptr [rsi],eax
  _asm { 0xC9 };                 // leave
  _asm { 0xC2 0x10 0x00 };       // ret 16
#endif
}

//---------------------------------------------------------------

public void InterlockedAnd (ref uint variable, uint value)
{
#if MEM32
  _asm { 0x8B 0x45 0x08 };       // mov  eax,ebp+8   ; load new value
  _asm { 0x8B 0x75 0x0C };       // mov  esi,ebp+12  ; load addr of variable
  _asm { 0xF0 0x21 0x06 };       // lock  and  dword ptr [esi],eax
  _asm { 0xC9 };                 // leave
  _asm { 0xC2 0x08 0x00 };       // ret 8
#else
  _asm { 0x8B 0x45 0x10 };       // mov   eax,rbp+16  ; load new value
  _asm { 0x48 0x8B 0x75 0x18 };  // mov   rsi,rbp+24  ; load addr of variable
  _asm { 0xF0 0x21 0x06 };       // lock  and  dword ptr [rsi],eax
  _asm { 0xC9 };                 // leave
  _asm { 0xC2 0x10 0x00 };       // ret 16
#endif
}

//---------------------------------------------------------------

public void InterlockedOr (ref uint variable, uint value)
{
#if MEM32
  _asm { 0x8B 0x45 0x08 };       // mov  eax,ebp+8   ; load new value
  _asm { 0x8B 0x75 0x0C };       // mov  esi,ebp+12  ; load addr of variable
  _asm { 0xF0 0x09 0x06 };       // lock  or  dword ptr [esi],eax
  _asm { 0xC9 };                 // leave
  _asm { 0xC2 0x08 0x00 };       // ret 8
#else
  _asm { 0x8B 0x45 0x10 };       // mov   eax,rbp+16  ; load new value
  _asm { 0x48 0x8B 0x75 0x18 };  // mov   rsi,rbp+24  ; load addr of variable
  _asm { 0xF0 0x09 0x06 };       // lock  or  dword ptr [rsi],eax
  _asm { 0xC9 };                 // leave
  _asm { 0xC2 0x10 0x00 };       // ret 16
#endif
}

//---------------------------------------------------------------

public void set_timer (out TIMER t, uint nsecs)
{
  t = GetTickCount() + nsecs * 1000;
}

//---------------------------------------------------------------

// returns true if timer has elapsed, false otherwise.

public bool timer_elapsed (TIMER t)
{
  // note: GetTickCount() wraps to zero after 49.7 days
  // so we check if the timer elapsed at least 10 days ago.
  return GetTickCount() - t < 864000000;   /* 10 days */
}

//---------------------------------------------------------------

// system timer incremented TICKS_PER_SEC times per second;
// the timer may overflow after some days.

public uint ticks ()
{
  return GetTickCount();
}

//---------------------------------------------------------------

// set granularity of ticks in msecs (default is 16 msec, best possible is 1 msec)
// returns 0 if success, -1 if value not supported.

public int set_granularity (uint msecs)
{
  if (granularity != 0)
    timeEndPeriod (granularity);    // undo old granularity

  if (timeBeginPeriod (msecs) != 0)  // not supported
  {
    granularity = 0;
    return -1;
  }

  granularity = msecs;
  return 0;
}

//---------------------------------------------------------------

public void delay (uint milliseconds)
{
  Sleep (milliseconds);
}

//---------------------------------------------------------------

// returns number of 100-nanosecond intervals since January 1, 1601 (UTC).

public long clock ()
{
  long time;
  GetSystemTimeAsFileTime (&time);
  return time;
}

//---------------------------------------------------------------

// very high precision timer
// returns value of intern cpu clock

public long performance_timer ()
{
  long l;
  QueryPerformanceCounter (&l);
  return l;
}

//---------------------------------------------------------------

// returns nb of cpu clock increments per second

public long performance_timer_frequency ()
{
  long l;
  if (QueryPerformanceFrequency (&l) == 0)
    return 0;
  return l;
}

//---------------------------------------------------------------

// 1) object must be aligned on 32-bit boundary.
// 2) object must be cleared before use.

// to be the only thread executing a piece of code,
// enclose it between enter_shared_object() and leave_shared_object().

public void enter_shared_object (ref SHARED_OBJECT o)
{
  uint old;

  if (o.init == 2)  // initialized
  {
    EnterCriticalSection (&o.cs);
    return;
  }

  old = InterlockedExchange (ref o.init, 1);

  switch (old)
  {
    case 0:   // we're the first caller
      InitializeCriticalSection (&o.cs);
      (void)InterlockedExchange (ref o.init, 2);  // set it to 2 (initialized)
      break;

    case 1:   // someone else is taking care of initialization
      // wait until it switches to 2 ..
      while (o.init != 2)
        sleep 0;
      break;

    case 2:   // it was already initialized but we did set it back to 1 !
      (void)InterlockedExchange (ref o.init, 2);  // set it back to 2 (initialized)
      break;

    default:
      abort;
  }

  EnterCriticalSection (&o.cs);
}

//---------------------------------------------------------------

public void leave_shared_object (ref SHARED_OBJECT o)
{
  assert (o.init != 0);
  LeaveCriticalSection (&o.cs);
}

//---------------------------------------------------------------

public void destroy_shared_object (ref SHARED_OBJECT o)
{
  if (o.init != 0)
  {
    DeleteCriticalSection (&o.cs);
    o.init = 0;
  }
}

//---------------------------------------------------------------

public uint get_current_process_id ()
{
  if (pid == 0)
    pid = GetCurrentProcessId ();
  return pid;
}

//---------------------------------------------------------------

public uint get_current_thread_id ()
{
  return GetCurrentThreadId ();
}

//---------------------------------------------------------------

public int start_process (string executable, string arguments, bool wait = true)
{
  int                 rc;
  STARTUPINFOA        s;
  PROCESS_INFORMATION p;
  char                cmdline[260+260+64];

  clear s;
  s.cb = s'size;
  s.wShowWindow = SW_SHOWDEFAULT;

  if (strchr (executable, ' ') != -1)   /* contains spaces */
  {
    sprintf (out cmdline, "\"%s\"", executable);
  }
  else
  {
    sprintf (out cmdline, "%s", executable);
  }

  if (strlen(arguments) > 0)
  {
    strcat (ref cmdline, " ");
    strcat (ref cmdline, arguments);
  }

  rc = CreateProcessA (null, &cmdline, null, null, FALSE, 0, null, null, &s, &p);
  if (rc != 0)
  {
    if (wait)
      WaitForInputIdle (p.hProcess, 30000);   /* timeout: 30 secs */
    CloseHandle (p.hProcess);
    CloseHandle (p.hThread);
    return 0;
  }

  return negative_last_windows_error();
}

//---------------------------------------------------------------

public int wstart_process (wstring executable, wstring arguments, bool wait = true)
{
  int                 rc;
  STARTUPINFOA        s;
  PROCESS_INFORMATION p;
  wchar               cmdline[260+260+64];

  clear s;
  s.cb = s'size;
  s.wShowWindow = SW_SHOWDEFAULT;

  if (wstrchr (executable, L' ') != -1)   /* contains spaces */
  {
    wsprintf (out cmdline, L"\"%S\"", executable);
  }
  else
  {
    wsprintf (out cmdline, L"%S", executable);
  }

  if (wstrlen(arguments) > 0)
  {
    wstrcat (ref cmdline, L" ");
    wstrcat (ref cmdline, arguments);
  }

  rc = CreateProcessW (null, &cmdline, null, null, FALSE, 0, null, null, &s, &p);
  if (rc != 0)
  {
    if (wait)
      WaitForInputIdle (p.hProcess, 30000);   /* timeout: 30 secs */
    CloseHandle (p.hProcess);
    CloseHandle (p.hThread);
    return 0;
  }

  return negative_last_windows_error();
}

//---------------------------------------------------------------

// terminates the current process and returns a code to the calling process

public void exit (int code = 0)
{
  ExitProcess ((uint)code);
}

//---------------------------------------------------------------

public int create_signal ()
{
  int rc = (int)CreateSemaphoreA (null, lInitialCount => 0, lMaximumCount => 1, null);
  return (rc == 0) ? -1 : rc;
}

//---------------------------------------------------------------

public void raise_signal (int s)
{
  (void)ReleaseSemaphore (s, lReleaseCount => 1, lpPreviousCount => null);
}

//---------------------------------------------------------------

// returns -1 if timeout, 0 if signal was raised.

public int wait_signal (int s, uint timeout_msecs = uint'max)
{
  DWORD rc;
  HANDLE hs = s;
  rc = WaitForMultipleObjectsEx (1, &hs, fWaitAll => 0, timeout_msecs, alertable => 0);
  if (rc == WAIT_TIMEOUT)
    return -1;
  else if (rc == 0)
    return 0;
  else
    abort;
}

//---------------------------------------------------------------

public void close_signal (int s)
{
  assert CloseHandle(s) != 0;
}

//---------------------------------------------------------------

// -2 to +2
public int get_thread_priority ()
{
  return GetThreadPriority (GetCurrentThread());
}

//---------------------------------------------------------------

// -2 to +2
public void set_thread_priority (int priority)
{
  SetThreadPriority (GetCurrentThread(), priority);
}

//---------------------------------------------------------------
#end unsafe
//---------------------------------------------------------------
