
/* fsm1.c : table for finite state machine 1 */

/******************************************************************/

/* start states: */
const int START_WHITE =   0;
const int START_BLACK = 213;

/* operation codes: */
const uint2 PUT_WHITE = 1;
const uint2 PUT_BLACK = 2;
const uint2 OP_ERROR  = 3;

/* error state: */
const uint2 AN_ERROR = 426;

/******************************************************************/

struct ACTION
{
  uint2 operation;
  uint2 length;
  uint2 zero;
  uint2 one;
}

const ACTION action[427] =
{
 /*   0 */  { 0,            0,        1,        16 },
 /*   1 */  { 0,            0,        2,        13 },
 /*   2 */  { 0,            0,        9,         3 },
 /*   3 */  { 0,            0,       36,         4 },
 /*   4 */  { 0,            0,        5,        32 },
 /*   5 */  { 0,            0,       75,         6 },
 /*   6 */  { 0,            0,        7,       141 },
 /*   7 */  { 0,            0,      133,         8 },
 /*   8 */  { PUT_WHITE,    0,      214,       226 },
 /*   9 */  { 0,            0,       39,        10 },
 /*  10 */  { 0,            0,       54,        11 },
 /*  11 */  { 0,            0,       52,        12 },
 /*  12 */  { PUT_WHITE,    1,      214,       226 },
 /*  13 */  { 0,            0,       33,        14 },
 /*  14 */  { 0,            0,      137,        15 },
 /*  15 */  { PUT_WHITE,    2,      214,       226 },
 /*  16 */  { 0,            0,       17,        22 },
 /*  17 */  { 0,            0,       18,        20 },
 /*  18 */  { 0,            0,       19,        28 },
 /*  19 */  { PUT_WHITE,    3,      214,       226 },
 /*  20 */  { 0,            0,       30,        21 },
 /*  21 */  { PUT_WHITE,    4,      214,       226 },
 /*  22 */  { 0,            0,       23,        25 },
 /*  23 */  { 0,            0,       24,        42 },
 /*  24 */  { PUT_WHITE,    5,      214,       226 },
 /*  25 */  { 0,            0,       26,        27 },
 /*  26 */  { PUT_WHITE,    6,      214,       226 },
 /*  27 */  { PUT_WHITE,    7,      214,       226 },
 /*  28 */  { 0,            0,      135,        29 },
 /*  29 */  { PUT_WHITE,    8,      214,       226 },
 /*  30 */  { 0,            0,       31,        46 },
 /*  31 */  { PUT_WHITE,    9,      214,       226 },
 /*  32 */  { PUT_WHITE,   10,      214,       226 },
 /*  33 */  { 0,            0,       34,        65 },
 /*  34 */  { 0,            0,       35,        49 },
 /*  35 */  { PUT_WHITE,   11,      214,       226 },
 /*  36 */  { 0,            0,       37,        57 },
 /*  37 */  { 0,            0,       38,        71 },
 /*  38 */  { PUT_WHITE,   12,      214,       226 },
 /*  39 */  { 0,            0,       60,        40 },
 /*  40 */  { 0,            0,       63,        41 },
 /*  41 */  { PUT_WHITE,   13,      214,       226 },
 /*  42 */  { 0,            0,       43,       134 },
 /*  43 */  { 0,            0,       44,        45 },
 /*  44 */  { PUT_WHITE,   14,      214,       226 },
 /*  45 */  { PUT_WHITE,   15,      214,       226 },
 /*  46 */  { 0,            0,       47,        48 },
 /*  47 */  { PUT_WHITE,   16,      214,       226 },
 /*  48 */  { PUT_WHITE,   17,      214,       226 },
 /*  49 */  { 0,            0,       73,        50 },
 /*  50 */  { 0,            0,      174,        51 },
 /*  51 */  { PUT_WHITE,   18,      214,       226 },
 /*  52 */  { 0,            0,       53,        81 },
 /*  53 */  { PUT_WHITE,   19,      214,       226 },
 /*  54 */  { 0,            0,       55,        87 },
 /*  55 */  { 0,            0,       56,        84 },
 /*  56 */  { PUT_WHITE,   20,      214,       226 },
 /*  57 */  { 0,            0,       94,        58 },
 /*  58 */  { 0,            0,      101,        59 },
 /*  59 */  { PUT_WHITE,   21,      214,       226 },
 /*  60 */  { 0,            0,       77,        61 },
 /*  61 */  { 0,            0,      104,        62 },
 /*  62 */  { PUT_WHITE,   22,      214,       226 },
 /*  63 */  { 0,            0,       64,       107 },
 /*  64 */  { PUT_WHITE,   23,      214,       226 },
 /*  65 */  { 0,            0,       66,       119 },
 /*  66 */  { 0,            0,       67,        69 },
 /*  67 */  { 0,            0,       68,       110 },
 /*  68 */  { PUT_WHITE,   24,      214,       226 },
 /*  69 */  { 0,            0,      113,        70 },
 /*  70 */  { PUT_WHITE,   25,      214,       226 },
 /*  71 */  { 0,            0,      116,        72 },
 /*  72 */  { PUT_WHITE,   26,      214,       226 },
 /*  73 */  { 0,            0,       74,       127 },
 /*  74 */  { PUT_WHITE,   27,      214,       226 },
 /*  75 */  { 0,            0,       76,       130 },
 /*  76 */  { PUT_WHITE,   28,      214,       226 },
 /*  77 */  { 0,            0,      182,        78 },
 /*  78 */  { 0,            0,       79,        80 },
 /*  79 */  { PUT_WHITE,   29,      214,       226 },
 /*  80 */  { PUT_WHITE,   30,      214,       226 },
 /*  81 */  { 0,            0,       82,        83 },
 /*  82 */  { PUT_WHITE,   31,      214,       226 },
 /*  83 */  { PUT_WHITE,   32,      214,       226 },
 /*  84 */  { 0,            0,       85,        86 },
 /*  85 */  { PUT_WHITE,   33,      214,       226 },
 /*  86 */  { PUT_WHITE,   34,      214,       226 },
 /*  87 */  { 0,            0,       88,        91 },
 /*  88 */  { 0,            0,       89,        90 },
 /*  89 */  { PUT_WHITE,   35,      214,       226 },
 /*  90 */  { PUT_WHITE,   36,      214,       226 },
 /*  91 */  { 0,            0,       92,        93 },
 /*  92 */  { PUT_WHITE,   37,      214,       226 },
 /*  93 */  { PUT_WHITE,   38,      214,       226 },
 /*  94 */  { 0,            0,       95,        98 },
 /*  95 */  { 0,            0,       96,        97 },
 /*  96 */  { PUT_WHITE,   39,      214,       226 },
 /*  97 */  { PUT_WHITE,   40,      214,       226 },
 /*  98 */  { 0,            0,       99,       100 },
 /*  99 */  { PUT_WHITE,   41,      214,       226 },
 /* 100 */  { PUT_WHITE,   42,      214,       226 },
 /* 101 */  { 0,            0,      102,       103 },
 /* 102 */  { PUT_WHITE,   43,      214,       226 },
 /* 103 */  { PUT_WHITE,   44,      214,       226 },
 /* 104 */  { 0,            0,      105,       106 },
 /* 105 */  { PUT_WHITE,   45,      214,       226 },
 /* 106 */  { PUT_WHITE,   46,      214,       226 },
 /* 107 */  { 0,            0,      108,       109 },
 /* 108 */  { PUT_WHITE,   47,      214,       226 },
 /* 109 */  { PUT_WHITE,   48,      214,       226 },
 /* 110 */  { 0,            0,      111,       112 },
 /* 111 */  { PUT_WHITE,   49,      214,       226 },
 /* 112 */  { PUT_WHITE,   50,      214,       226 },
 /* 113 */  { 0,            0,      114,       115 },
 /* 114 */  { PUT_WHITE,   51,      214,       226 },
 /* 115 */  { PUT_WHITE,   52,      214,       226 },
 /* 116 */  { 0,            0,      117,       118 },
 /* 117 */  { PUT_WHITE,   53,      214,       226 },
 /* 118 */  { PUT_WHITE,   54,      214,       226 },
 /* 119 */  { 0,            0,      120,       136 },
 /* 120 */  { 0,            0,      121,       124 },
 /* 121 */  { 0,            0,      122,       123 },
 /* 122 */  { PUT_WHITE,   55,      214,       226 },
 /* 123 */  { PUT_WHITE,   56,      214,       226 },
 /* 124 */  { 0,            0,      125,       126 },
 /* 125 */  { PUT_WHITE,   57,      214,       226 },
 /* 126 */  { PUT_WHITE,   58,      214,       226 },
 /* 127 */  { 0,            0,      128,       129 },
 /* 128 */  { PUT_WHITE,   59,      214,       226 },
 /* 129 */  { PUT_WHITE,   60,      214,       226 },
 /* 130 */  { 0,            0,      131,       132 },
 /* 131 */  { PUT_WHITE,   61,      214,       226 },
 /* 132 */  { PUT_WHITE,   62,      214,       226 },
 /* 133 */  { PUT_WHITE,   63,      214,       226 },
 /* 134 */  { PUT_WHITE,   64,        1,        16 },
 /* 135 */  { PUT_WHITE,  128,        1,        16 },
 /* 136 */  { PUT_WHITE,  192,        1,        16 },
 /* 137 */  { 0,            0,      144,       138 },
 /* 138 */  { 0,            0,      149,       139 },
 /* 139 */  { 0,            0,      167,       140 },
 /* 140 */  { PUT_WHITE,  256,        1,        16 },
 /* 141 */  { 0,            0,      142,       143 },
 /* 142 */  { PUT_WHITE,  320,        1,        16 },
 /* 143 */  { PUT_WHITE,  384,        1,        16 },
 /* 144 */  { 0,            0,      180,       145 },
 /* 145 */  { 0,            0,      146,       152 },
 /* 146 */  { 0,            0,      147,       148 },
 /* 147 */  { PUT_WHITE,  448,        1,        16 },
 /* 148 */  { PUT_WHITE,  512,        1,        16 },
 /* 149 */  { 0,            0,      150,       160 },
 /* 150 */  { 0,            0,      151,       157 },
 /* 151 */  { PUT_WHITE,  576,        1,        16 },
 /* 152 */  { 0,            0,      154,       153 },
 /* 153 */  { PUT_WHITE,  640,        1,        16 },
 /* 154 */  { 0,            0,      155,       156 },
 /* 155 */  { PUT_WHITE,  704,        1,        16 },
 /* 156 */  { PUT_WHITE,  768,        1,        16 },
 /* 157 */  { 0,            0,      158,       159 },
 /* 158 */  { PUT_WHITE,  832,        1,        16 },
 /* 159 */  { PUT_WHITE,  896,        1,        16 },
 /* 160 */  { 0,            0,      161,       164 },
 /* 161 */  { 0,            0,      162,       163 },
 /* 162 */  { PUT_WHITE,  960,        1,        16 },
 /* 163 */  { PUT_WHITE, 1024,        1,        16 },
 /* 164 */  { 0,            0,      165,       166 },
 /* 165 */  { PUT_WHITE, 1088,        1,        16 },
 /* 166 */  { PUT_WHITE, 1152,        1,        16 },
 /* 167 */  { 0,            0,      168,       171 },
 /* 168 */  { 0,            0,      169,       170 },
 /* 169 */  { PUT_WHITE, 1216,        1,        16 },
 /* 170 */  { PUT_WHITE, 1280,        1,        16 },
 /* 171 */  { 0,            0,      172,       173 },
 /* 172 */  { PUT_WHITE, 1344,        1,        16 },
 /* 173 */  { PUT_WHITE, 1408,        1,        16 },
 /* 174 */  { 0,            0,      175,       178 },
 /* 175 */  { 0,            0,      176,       177 },
 /* 176 */  { PUT_WHITE, 1472,        1,        16 },
 /* 177 */  { PUT_WHITE, 1536,        1,        16 },
 /* 178 */  { 0,            0,      179,       181 },
 /* 179 */  { PUT_WHITE, 1600,        1,        16 },
 /* 180 */  { PUT_WHITE, 1664,        1,        16 },
 /* 181 */  { PUT_WHITE, 1728,        1,        16 },
 /* 182 */  { 0,            0,      183,       188 },
 /* 183 */  { 0,            0,      184,  AN_ERROR },
 /* 184 */  { 0,            0,      185,  AN_ERROR },
 /* 185 */  { 0,            0,      186,  AN_ERROR },
 /* 186 */  { 0,            0,      186,       187 },
 /* 187 */  { 0,            0,        1,        16 },
 /* 188 */  { 0,            0,      189,       192 },
 /* 189 */  { 0,            0,      190,       199 },
 /* 190 */  { 0,            0,      191,       196 },
 /* 191 */  { PUT_WHITE, 1792,        1,        16 },
 /* 192 */  { 0,            0,      193,       206 },
 /* 193 */  { 0,            0,      194,       195 },
 /* 194 */  { PUT_WHITE, 1856,        1,        16 },
 /* 195 */  { PUT_WHITE, 1920,        1,        16 },
 /* 196 */  { 0,            0,      197,       198 },
 /* 197 */  { PUT_WHITE, 1984,        1,        16 },
 /* 198 */  { PUT_WHITE, 2048,        1,        16 },
 /* 199 */  { 0,            0,      200,       203 },
 /* 200 */  { 0,            0,      201,       202 },
 /* 201 */  { PUT_WHITE, 2112,        1,        16 },
 /* 202 */  { PUT_WHITE, 2176,        1,        16 },
 /* 203 */  { 0,            0,      204,       205 },
 /* 204 */  { PUT_WHITE, 2240,        1,        16 },
 /* 205 */  { PUT_WHITE, 2304,        1,        16 },
 /* 206 */  { 0,            0,      207,       210 },
 /* 207 */  { 0,            0,      208,       209 },
 /* 208 */  { PUT_WHITE, 2368,        1,        16 },
 /* 209 */  { PUT_WHITE, 2432,        1,        16 },
 /* 210 */  { 0,            0,      211,       212 },
 /* 211 */  { PUT_WHITE, 2496,        1,        16 },
 /* 212 */  { PUT_WHITE, 2560,        1,        16 },
 /* 213 */  { 0,            0,      214,       226 },
 /* 214 */  { 0,            0,      215,       224 },
 /* 215 */  { 0,            0,      216,       230 },
 /* 216 */  { 0,            0,      217,       233 },
 /* 217 */  { 0,            0,      242,       218 },
 /* 218 */  { 0,            0,      238,       219 },
 /* 219 */  { 0,            0,      220,       241 },
 /* 220 */  { 0,            0,      248,       221 },
 /* 221 */  { 0,            0,      264,       222 },
 /* 222 */  { 0,            0,      267,       223 },
 /* 223 */  { PUT_BLACK,    0,        1,        16 },
 /* 224 */  { 0,            0,      225,       229 },
 /* 225 */  { PUT_BLACK,    1,        1,        16 },
 /* 226 */  { 0,            0,      228,       227 },
 /* 227 */  { PUT_BLACK,    2,        1,        16 },
 /* 228 */  { PUT_BLACK,    3,        1,        16 },
 /* 229 */  { PUT_BLACK,    4,        1,        16 },
 /* 230 */  { 0,            0,      232,       231 },
 /* 231 */  { PUT_BLACK,    5,        1,        16 },
 /* 232 */  { PUT_BLACK,    6,        1,        16 },
 /* 233 */  { 0,            0,      235,       234 },
 /* 234 */  { PUT_BLACK,    7,        1,        16 },
 /* 235 */  { 0,            0,      237,       236 },
 /* 236 */  { PUT_BLACK,    8,        1,        16 },
 /* 237 */  { PUT_BLACK,    9,        1,        16 },
 /* 238 */  { 0,            0,      239,       240 },
 /* 239 */  { PUT_BLACK,   10,        1,        16 },
 /* 240 */  { PUT_BLACK,   11,        1,        16 },
 /* 241 */  { PUT_BLACK,   12,        1,        16 },
 /* 242 */  { 0,            0,      256,       243 },
 /* 243 */  { 0,            0,      244,       246 },
 /* 244 */  { 0,            0,      245,       250 },
 /* 245 */  { PUT_BLACK,   13,        1,        16 },
 /* 246 */  { 0,            0,      253,       247 },
 /* 247 */  { PUT_BLACK,   14,        1,        16 },
 /* 248 */  { 0,            0,      249,       261 },
 /* 249 */  { PUT_BLACK,   15,        1,        16 },
 /* 250 */  { 0,            0,      272,       251 },
 /* 251 */  { 0,            0,      341,       252 },
 /* 252 */  { PUT_BLACK,   16,        1,        16 },
 /* 253 */  { 0,            0,      254,       269 },
 /* 254 */  { 0,            0,      255,       319 },
 /* 255 */  { PUT_BLACK,   17,        1,        16 },
 /* 256 */  { 0,            0,      395,       257 },
 /* 257 */  { 0,            0,      258,       278 },
 /* 258 */  { 0,            0,      259,       275 },
 /* 259 */  { 0,            0,      260,       326 },
 /* 260 */  { PUT_BLACK,   18,        1,        16 },
 /* 261 */  { 0,            0,      282,       262 },
 /* 262 */  { 0,            0,      286,       263 },
 /* 263 */  { PUT_BLACK,   19,        1,        16 },
 /* 264 */  { 0,            0,      265,       299 },
 /* 265 */  { 0,            0,      266,       296 },
 /* 266 */  { PUT_BLACK,   20,        1,        16 },
 /* 267 */  { 0,            0,      268,       309 },
 /* 268 */  { PUT_BLACK,   21,        1,        16 },
 /* 269 */  { 0,            0,      289,       270 },
 /* 270 */  { 0,            0,      306,       271 },
 /* 271 */  { PUT_BLACK,   22,        1,        16 },
 /* 272 */  { 0,            0,      273,       312 },
 /* 273 */  { 0,            0,      274,       323 },
 /* 274 */  { PUT_BLACK,   23,        1,        16 },
 /* 275 */  { 0,            0,      338,       276 },
 /* 276 */  { 0,            0,      347,       277 },
 /* 277 */  { PUT_BLACK,   24,        1,        16 },
 /* 278 */  { 0,            0,      279,       332 },
 /* 279 */  { 0,            0,      280,       329 },
 /* 280 */  { 0,            0,      281,       359 },
 /* 281 */  { PUT_BLACK,   25,        1,        16 },
 /* 282 */  { 0,            0,      355,       283 },
 /* 283 */  { 0,            0,      284,       285 },
 /* 284 */  { PUT_BLACK,   26,        1,        16 },
 /* 285 */  { PUT_BLACK,   27,        1,        16 },
 /* 286 */  { 0,            0,      287,       288 },
 /* 287 */  { PUT_BLACK,   28,        1,        16 },
 /* 288 */  { PUT_BLACK,   29,        1,        16 },
 /* 289 */  { 0,            0,      290,       293 },
 /* 290 */  { 0,            0,      291,       292 },
 /* 291 */  { PUT_BLACK,   30,        1,        16 },
 /* 292 */  { PUT_BLACK,   31,        1,        16 },
 /* 293 */  { 0,            0,      294,       295 },
 /* 294 */  { PUT_BLACK,   32,        1,        16 },
 /* 295 */  { PUT_BLACK,   33,        1,        16 },
 /* 296 */  { 0,            0,      297,       298 },
 /* 297 */  { PUT_BLACK,   34,        1,        16 },
 /* 298 */  { PUT_BLACK,   35,        1,        16 },
 /* 299 */  { 0,            0,      300,       303 },
 /* 300 */  { 0,            0,      301,       302 },
 /* 301 */  { PUT_BLACK,   36,        1,        16 },
 /* 302 */  { PUT_BLACK,   37,        1,        16 },
 /* 303 */  { 0,            0,      304,       305 },
 /* 304 */  { PUT_BLACK,   38,        1,        16 },
 /* 305 */  { PUT_BLACK,   39,        1,        16 },
 /* 306 */  { 0,            0,      307,       308 },
 /* 307 */  { PUT_BLACK,   40,        1,        16 },
 /* 308 */  { PUT_BLACK,   41,        1,        16 },
 /* 309 */  { 0,            0,      310,       311 },
 /* 310 */  { PUT_BLACK,   42,        1,        16 },
 /* 311 */  { PUT_BLACK,   43,        1,        16 },
 /* 312 */  { 0,            0,      313,       316 },
 /* 313 */  { 0,            0,      314,       315 },
 /* 314 */  { PUT_BLACK,   44,        1,        16 },
 /* 315 */  { PUT_BLACK,   45,        1,        16 },
 /* 316 */  { 0,            0,      317,       318 },
 /* 317 */  { PUT_BLACK,   46,        1,        16 },
 /* 318 */  { PUT_BLACK,   47,        1,        16 },
 /* 319 */  { 0,            0,      320,       351 },
 /* 320 */  { 0,            0,      321,       322 },
 /* 321 */  { PUT_BLACK,   48,        1,        16 },
 /* 322 */  { PUT_BLACK,   49,        1,        16 },
 /* 323 */  { 0,            0,      324,       325 },
 /* 324 */  { PUT_BLACK,   50,        1,        16 },
 /* 325 */  { PUT_BLACK,   51,        1,        16 },
 /* 326 */  { 0,            0,      327,       336 },
 /* 327 */  { 0,            0,      328,       367 },
 /* 328 */  { PUT_BLACK,   52,        1,        16 },
 /* 329 */  { 0,            0,      361,       330 },
 /* 330 */  { 0,            0,      364,       331 },
 /* 331 */  { PUT_BLACK,   53,        1,        16 },
 /* 332 */  { 0,            0,      333,       354 },
 /* 333 */  { 0,            0,      334,       376 },
 /* 334 */  { 0,            0,      335,       373 },
 /* 335 */  { PUT_BLACK,   54,        1,        16 },
 /* 336 */  { 0,            0,      370,       337 },
 /* 337 */  { PUT_BLACK,   55,        1,        16 },
 /* 338 */  { 0,            0,      339,       345 },
 /* 339 */  { 0,            0,      340,       383 },
 /* 340 */  { PUT_BLACK,   56,        1,        16 },
 /* 341 */  { 0,            0,      342,       349 },
 /* 342 */  { 0,            0,      343,       344 },
 /* 343 */  { PUT_BLACK,   57,        1,        16 },
 /* 344 */  { PUT_BLACK,   58,        1,        16 },
 /* 345 */  { 0,            0,      386,       346 },
 /* 346 */  { PUT_BLACK,   59,        1,        16 },
 /* 347 */  { 0,            0,      348,       389 },
 /* 348 */  { PUT_BLACK,   60,        1,        16 },
 /* 349 */  { 0,            0,      350,       358 },
 /* 350 */  { PUT_BLACK,   61,        1,        16 },
 /* 351 */  { 0,            0,      352,       353 },
 /* 352 */  { PUT_BLACK,   62,        1,        16 },
 /* 353 */  { PUT_BLACK,   63,        1,        16 },
 /* 354 */  { PUT_BLACK,   64,      214,       226 },
 /* 355 */  { 0,            0,      356,       357 },
 /* 356 */  { PUT_BLACK,  128,      214,       226 },
 /* 357 */  { PUT_BLACK,  192,      214,       226 },
 /* 358 */  { PUT_BLACK,  256,      214,       226 },
 /* 359 */  { 0,            0,      392,       360 },
 /* 360 */  { PUT_BLACK,  320,      214,       226 },
 /* 361 */  { 0,            0,      362,       363 },
 /* 362 */  { PUT_BLACK,  384,      214,       226 },
 /* 363 */  { PUT_BLACK,  448,      214,       226 },
 /* 364 */  { 0,            0,      365,       366 },
 /* 365 */  { PUT_BLACK,  512,      214,       226 },
 /* 366 */  { PUT_BLACK,  576,      214,       226 },
 /* 367 */  { 0,            0,      368,       369 },
 /* 368 */  { PUT_BLACK,  640,      214,       226 },
 /* 369 */  { PUT_BLACK,  704,      214,       226 },
 /* 370 */  { 0,            0,      371,       372 },
 /* 371 */  { PUT_BLACK,  768,      214,       226 },
 /* 372 */  { PUT_BLACK,  832,      214,       226 },
 /* 373 */  { 0,            0,      374,       375 },
 /* 374 */  { PUT_BLACK,  896,      214,       226 },
 /* 375 */  { PUT_BLACK,  960,      214,       226 },
 /* 376 */  { 0,            0,      377,       380 },
 /* 377 */  { 0,            0,      378,       379 },
 /* 378 */  { PUT_BLACK, 1024,      214,       226 },
 /* 379 */  { PUT_BLACK, 1088,      214,       226 },
 /* 380 */  { 0,            0,      381,       382 },
 /* 381 */  { PUT_BLACK, 1152,      214,       226 },
 /* 382 */  { PUT_BLACK, 1216,      214,       226 },
 /* 383 */  { 0,            0,      384,       385 },
 /* 384 */  { PUT_BLACK, 1280,      214,       226 },
 /* 385 */  { PUT_BLACK, 1344,      214,       226 },
 /* 386 */  { 0,            0,      387,       388 },
 /* 387 */  { PUT_BLACK, 1408,      214,       226 },
 /* 388 */  { PUT_BLACK, 1472,      214,       226 },
 /* 389 */  { 0,            0,      390,       391 },
 /* 390 */  { PUT_BLACK, 1536,      214,       226 },
 /* 391 */  { PUT_BLACK, 1600,      214,       226 },
 /* 392 */  { 0,            0,      393,       394 },
 /* 393 */  { PUT_BLACK, 1664,      214,       226 },
 /* 394 */  { PUT_BLACK, 1728,      214,       226 },
 /* 395 */  { 0,            0,      396,       401 },
 /* 396 */  { 0,            0,      397,  AN_ERROR },
 /* 397 */  { 0,            0,      398,  AN_ERROR },
 /* 398 */  { 0,            0,      399,  AN_ERROR },
 /* 399 */  { 0,            0,      399,       400 },
 /* 400 */  { 0,            0,        1,        16 },
 /* 401 */  { 0,            0,      402,       405 },
 /* 402 */  { 0,            0,      403,       412 },
 /* 403 */  { 0,            0,      404,       409 },
 /* 404 */  { PUT_BLACK, 1792,      214,       226 },
 /* 405 */  { 0,            0,      406,       419 },
 /* 406 */  { 0,            0,      407,       408 },
 /* 407 */  { PUT_BLACK, 1856,      214,       226 },
 /* 408 */  { PUT_BLACK, 1920,      214,       226 },
 /* 409 */  { 0,            0,      410,       411 },
 /* 410 */  { PUT_BLACK, 1984,      214,       226 },
 /* 411 */  { PUT_BLACK, 2048,      214,       226 },
 /* 412 */  { 0,            0,      413,       416 },
 /* 413 */  { 0,            0,      414,       415 },
 /* 414 */  { PUT_BLACK, 2112,      214,       226 },
 /* 415 */  { PUT_BLACK, 2176,      214,       226 },
 /* 416 */  { 0,            0,      417,       418 },
 /* 417 */  { PUT_BLACK, 2240,      214,       226 },
 /* 418 */  { PUT_BLACK, 2304,      214,       226 },
 /* 419 */  { 0,            0,      420,       423 },
 /* 420 */  { 0,            0,      421,       422 },
 /* 421 */  { PUT_BLACK, 2368,      214,       226 },
 /* 422 */  { PUT_BLACK, 2432,      214,       226 },
 /* 423 */  { 0,            0,      424,       425 },
 /* 424 */  { PUT_BLACK, 2496,      214,       226 },
 /* 425 */  { PUT_BLACK, 2560,      214,       226 },
 /* 426 */  { OP_ERROR,     0, AN_ERROR,  AN_ERROR }};

/******************************************************************/
