
/* fsm2.h : table for finite state machine 2 */

/*
____________________________________________________________________________
 Mode  |      to be coded                  |  Notation   |     Code word
____________________________________________________________________________
 Pass  |         b1, b2                    |     P       |          0001
____________________________________________________________________________
 Horiz |       a0a1, a1a2                  |     H       | 001 + M(a0a1) +
       |                                                 |M(a1a2)(see Note)
____________________________________________________________________________
 Vert  |  a1 just under b1   |  a1b1 = 0   |    V(0)     |             1
____________________________________________________________________________
       |  a1 right of b1     |  a1b1 = 1   |   VR(1)     |           011
       |                     |  a1b1 = 2   |   VR(2)     |        000011
                             |  a1b1 = 3   |   VR(3)     |       0000011
____________________________________________________________________________
       |  a1 left of b(1)    |  a1b1 = 1   |   VL(1)     |           010
                             |  a1b1 = 2   |   VL(2)     |        000010
                             |  a1b1 = 3   |   VL(3)     |       0000010
____________________________________________________________________________
Extension                                                |    0000001xxx
____________________________________________________________________________
*/

/******************************************************************/

/* operation codes: */
const uint2 OPER_PASS  = 1;
const uint2 OPER_HORIZ = 2;
const uint2 OPER_VERT  = 3;
const uint2 OPER_EXT   = 4;
const uint2 OPER_ERROR = 5;

/* special state: */
const uint2 ENDING     = 20;

/******************************************************************/

struct ACTION2
{
  uint2  operation;
  int2   distance;
  uint2  zero;
  uint2  one;
}

const ACTION2 action2[21] =
{
 /*   0 */  { 0,            0,        1,         6 },    /*           */
 /*   1 */  { 0,            0,        2,         7 },    /* 0         */
 /*   2 */  { 0,            0,        3,         5 },    /* 00        */
 /*   3 */  { 0,            0,        9,         4 },    /* 000       */
 /*   4 */  { OPER_PASS,    0,   ENDING,    ENDING },    /* 0001      */
 /*   5 */  { OPER_HORIZ,   0,   ENDING,    ENDING },    /* 001       */
 /*   6 */  { OPER_VERT,   +0,   ENDING,    ENDING },    /* 1         */
 /*   7 */  { 0,            0,       15,         8 },    /* 01        */
 /*   8 */  { OPER_VERT,   +1,   ENDING,    ENDING },    /* 011       */
 /*   9 */  { 0,            0,       12,        10 },    /* 0000      */
 /*  10 */  { 0,            0,       16,        11 },    /* 00001     */
 /*  11 */  { OPER_VERT,   +2,   ENDING,    ENDING },    /* 000011    */
 /*  12 */  { 0,            0,       18,        13 },    /* 00000     */
 /*  13 */  { 0,            0,       17,        14 },    /* 000001    */
 /*  14 */  { OPER_VERT,   +3,   ENDING,    ENDING },    /* 0000011   */
 /*  15 */  { OPER_VERT,   -1,   ENDING,    ENDING },    /* 010       */
 /*  16 */  { OPER_VERT,   -2,   ENDING,    ENDING },    /* 000010    */
 /*  17 */  { OPER_VERT,   -3,   ENDING,    ENDING },    /* 0000010   */
 /*  18 */  { 0,            0,   ENDING,        19 },    /* 000000    */
 /*  19 */  { OPER_EXT,     0,   ENDING,    ENDING },    /* 0000001   */
 /*  20 */  { OPER_ERROR,   0,   ENDING,    ENDING }};

/******************************************************************/
