
// shader_defs.h

const string StructPointLight =
  "struct LIGHT " +
  "{ " +
    "float3 color;       " +  // used for diffuse and specular
    "float  range;       " +  // distance on which it has effect
    "float3 position;    " +  // position of light
    "float  attenuation; " +  // attenuation exponent : 0(no attenuation with distance) .. 1(linear attenuation)
    "float3 direction;   " +  // for spotlight
    "float  cone;        " +  // for spotlight (0.0 = no spotlight)
  "}; ";

const string VarConstantBuffer0 =    // stable data, does not change often, both for vertex and pixel shaders.
  "cbuffer ConstantBuffer0  " +
  "{ " +
    "matrix<float,4,4> viewMatrix;         " +
    "matrix<float,4,4> projectionMatrix;   " +
    "matrix<float,4,4> shadowMatrix; " +
    "float3 eyePosition;               " +
    "float  fogInvRange;         " +
    "float3 fogColor;            " +
    "float  fogStart;            " +
    "float3 CameraOffset;        " +   // used for perlin noise water
    "float  FogInvExponent;      " +
    "float2 filler1;             " +
    "int    nb_of_shadow_maps;   " +
    "int    indexes_supported;   " +   // 0 = amd slow vertex shader, 1 = indexes supported so fast vertex shader
    "float3 AmbiantLightColor; " +
    "float  filler3; " +
    "float3 SunLightColor; " +
    "float  filler2; " +
    "float3 SunLightDirection;   " +
    "int    NbLights;            " +
    "LIGHT  Light[128];          " +
  "}; ";

const string VarConstantBuffer1 =   // dynamic data, changes for each draw, both for vertex and pixel shaders.
  "cbuffer ConstantBuffer1  " +
  "{ " +
    "matrix<float,4,4> worldMatrix; " +
    "float4 MaterialColor; " +
    "float3 EmissiveColor; " +
    "float  SpecularCone; " +
    "float  fact_u_min_1; " +
    "float  fact_v_min_1; " +
    "float  fact_u; " +
    "float  fact_v; " +
    "float  ofs_u; " +
    "float  ofs_v; " +
    "float  tclock; " +
    "float  ShinyFactor; " +
    "int    ambiantlight_on; " +
    "int    sunlight_on; " +
    "uint   ambiant_downwards_and_bump; " +     // byte 0 = ambiant downwards
                                                // byte 1 = bump texture
    "float  parallax_factor; " +    // 0 = none, example -0.05
    "float4 object_id; " +
  "}; ";

const string VarConstantBuffer2 =   // bone matrices, only for rigged mesh vertex shader.
  "cbuffer ConstantBuffer2 " +
  "{ " +
    "matrix<float,4,4> g_bone0[64]; " +
  "}; ";

const string VarConstantBuffer3 =   // bone matrices, only for rigged mesh vertex shader.
  "cbuffer ConstantBuffer3 " +
  "{ " +
    "matrix<float,4,4> g_bone1[64]; " +
  "}; ";

const string VarConstantBuffer4 =   // bone matrices, only for rigged mesh vertex shader.
  "cbuffer ConstantBuffer4 " +
  "{ " +
    "matrix<float,4,4> g_bone2[64]; " +
  "}; ";

//-------------------------------------------------------------------------

const string VertexInputType =      // for basic vertex shader
  "struct VertexInputType " +
  "{ " +
    "float4 position : POSITION;  " +  // automatically converted from float3
    "float3 normal   : NORMAL;    " +
    "float3 tangent  : TANGENT;   " +
    "float4 diffuse  : COLOR;     " +  // diffuse color (uint will be converted to float4)
    "float2 tex      : TEXCOORD;  " +
    "uint   instance : SV_InstanceID; " +
  "}; ";

const string ShadowVertexInputType =
  "struct ShadowVertexInputType " +
  "{ " +
    "float4 position : POSITION;  " +  // automatically converted from float3
    "float3 normal   : NORMAL;    " +  // automatically converted from float3
    "float3 tangent  : TANGENT;   " +  // automatically converted from float3
    "float4 diffuse  : COLOR;     " +  // diffuse color (uint will be converted to float4)
    "float2 tex      : TEXCOORD;  " +
    "uint   instance : SV_InstanceID; " +
   "}; ";

//-------------------------------------------------------------------------

const string VertexInputTypeRiggedMesh =   // for rigged mesh shader
  "struct VertexInputTypeRiggedMesh " +
  "{ " +
    "float4 position  : POSITION;   " +  // automatically converted from float3
    "float3 normal    : NORMAL;     " +
    "float3 tangent   : TANGENT;    " +  // automatically converted from float3
    "float2 tex       : TEXCOORD;   " +
    "uint   bone      : BONE;       " +
    "float  weight[4] : WEIGHT;     " +  // 0.0 = none
    "uint   instance  : SV_InstanceID; " +
  "}; ";


const string ShadowVertexInputTypeRiggedMesh =   // for rigged mesh shader
  "struct ShadowVertexInputTypeRiggedMesh " +
  "{ " +
    "float4 position  : POSITION;   " +  // automatically converted from float3
    "float3 normal    : NORMAL;     " +
    "float3 tangent   : TANGENT;    " +
    "float2 tex       : TEXCOORD;   " +
    "uint   bone      : BONE;       " +
    "float  weight[4] : WEIGHT;     " +  // 0.0 = none
    "uint   instance  : SV_InstanceID; " +
  "}; ";

//-------------------------------------------------------------------------

const string ShadowVertexOutputType =
  "struct ShadowVertexOutputType " +
  "{ " +
    "float4 position          : SV_POSITION;  " +
    "uint   RenderTargetIndex : INT; " +
  "}; ";

const string ShadowGeometricOutputType =
  "struct ShadowGeometricOutputType " +
  "{ " +
    "float4 position          : SV_POSITION;  " +
    "uint   RenderTargetIndex : SV_RenderTargetArrayIndex; " +
  "}; ";

//-------------------------------------------------------------------------

const string PixelInputType =
  "struct PixelInputType " +
  "{ " +
    "float4 position  : SV_POSITION;  " +
    "float4 diffuse   : COLOR;        " + // diffuse color
    "float2 tex       : TEXCOORD;     " +
    "float3 posW      : POSITION;     " + // world position
    "float3 normal    : NORMAL;       " +
    "float3 tangent   : TANGENT;      " +
    "float3 ShadowPos : POSITION2;    " + // shadow position
  "}; ";

//-------------------------------------------------------------------------

const string PixelOutputType =
  "struct PixelOutputType " +
  "{ " +
    "float4 color    : SV_Target0; " +
    "float4 objectid : SV_Target1; " +
  " }; ";

//-------------------------------------------------------------------------
