
// shader_ps.h

//---------------------------------------------------------------------

////////////////////////////////////////////////////////////////////////////////
// basic pixel shader
////////////////////////////////////////////////////////////////////////////////

const string PIXEL_SHADER_SOLID_RENDERING =

 "Texture2D        shaderTexture[2]; " +   // color texture, bump texture
 "Texture2DArray   shadowTexture; " +
 "SamplerState     SampleType[2]; " +

 "SamplerComparisonState ShadowSampler " +
 "{ " +

   // sampler state : FILTER_COMPARISON_MIN_MAG_LINEAR_MIP_POINT
   //              or FILTER_COMPARISON_MIN_MAG_MIP_LINEAR
   //              or FILTER_COMPARISON_ANISOTROPIC
   " Filter = FILTER_COMPARISON_ANISOTROPIC; " +
   " AddressU = BORDER; " +
   " AddressV = BORDER; " +
   " AddressW = BORDER; " +
   " MaxAnisotropy = 1; " +
   " BorderColor = float4(0.0, 0.0, 0.0, 0.0); " +
   " ComparisonFunc = LESS; " +
    "MinLOD = 0.0; " +
    "MaxLOD = 12.0; " +
 "};" +


 " PixelOutputType MyPixelShader(PixelInputType input) " +
 "{ " +
   "PixelOutputType output; " +
   "float4 ElementColor, finalColor; " +
   "float3 LightColor, SpecularLightColor; " +
   "float3 toEye; " +
   "int    i;" +
   "float  sun_factor; " +   // normally 1.0, otherwise 0.0 if shadows

   "if ((ambiant_downwards_and_bump & 0xFF00) != 0) " +   // bump
   "{" +
      "float3 binormal, bumpNormal, V; " +
      "float4 bumpMap; " +

        // Gram-Schmidt orthogonalize
      "input.tangent = normalize ((input.tangent - input.normal * dot(input.tangent, input.normal))); " +

      // compute binormal
      "binormal = cross (input.tangent, input.normal); " +

      // get bump value in range (-1, +1)
      "bumpMap = (shaderTexture[1].Sample(SampleType[0], input.tex) * 2.0) - 1.0; " +

      // parallax effect : bumpMap.z contains height
      "if (parallax_factor != 0.0)" +
      "{" +
        "V = normalize (input.posW.xyz - eyePosition); " +
        "V = float3(dot(V,input.tangent), dot(V,binormal), 0.0); " +
        "input.tex += V.xy * bumpMap.z * parallax_factor; " +
        "bumpMap = (shaderTexture[1].Sample(SampleType[0], input.tex) * 2.0) - 1.0; " +
        "bumpMap.z = sqrt(1.0 - saturate(bumpMap.x * bumpMap.x + bumpMap.y * bumpMap.y)); " +
      "} " +

      "input.normal = normalize ((bumpMap.x * input.tangent) + (bumpMap.y * binormal) + (bumpMap.z * input.normal)); " +
   "} " +

   "if (tclock != 0.0) " +   // means perlin is active
   "{" +
     "float2 t1 = float2(input.posW.xz + CameraOffset.xz) " +
                " * (37.535333333 / 65536.178267652612); " +    // frequency 37.5 within area

     // notes: noise is discontinuous at integer coordinate values because the third dimension, tclock, is not integer !
     "float n1 = noise (float3(t1, 0.37657652 * tclock)); " +
     "float n2 = noise (float3(t1, 0.39657652 * tclock + 0.23565612121)); " +

     "input.tex.x += (1.0 / 128.0) * n1 + sin (tclock * 0.001323212); " +
     "input.tex.y += (1.0 / 128.0) * n2 + cos (tclock * 0.001671212); " +

     "input.normal.x += 0.05 * n1; " +
     "input.normal.z += 0.05 * n2; " +
     "input.normal = normalize (input.normal);  " +
   "}" +


   // ambiant light

   "LightColor = AmbiantLightColor * ambiantlight_on; " +

   "if ((ambiant_downwards_and_bump & 0xFF) != 0) " +
   "{" +
     "LightColor *= (0.85 + 0.15 * input.normal.y); " +  // byte 0 : ambiant light less intense when not shining downwards
   "}" +

   // sun light

   "sun_factor = sunlight_on; " +   // 0 or 1

   "if (input.ShadowPos.z >= 0.0) "  +      // .z equals -1.0 when shadows are OFF
   "{ " +
       "static const float F = 256.0; " +
       "static const float R = 256.0 * (510.0/512.0); " +
       "static const float res[5]  = {     256.0/F,   64.0/F,   16.0/F,   4.0/F, 1.0/F}; " +
       "static const float MAPLIMIT[5] = {  R/256.0,   R/64.0,   R/16.0,   R/4.0,     R}; " +

       "{ " +
           "  float2 xy = input.ShadowPos.xy; " +
           "  float  z = input.ShadowPos.z;   " +      // depth z = 0(far up in sky) to 1(below ground)
           "  float  size = max (abs(xy.x), abs(xy.y)); " +
           "  int    i; " +

           "  for (i=0; i<5 && i<nb_of_shadow_maps; i++) " +
           "  { " +
           "    if (size < MAPLIMIT[i]) " +
           "    { " +
           "      xy = xy * float2(0.5, -0.5) * res[i] + float2(0.5, 0.5); " +

                 "{ " +   // central pixel is in shadow

                   // PCF 3x3
                   "float percentLit = 0.0;  float3 xyz = float3(xy,i); " +
                   " for (int k = -1; k <= 1; k++) { for (int l = -1; l <= 1; l++) { " +
                   "    percentLit += shadowTexture.SampleCmpLevelZero(ShadowSampler, xyz, z, int2(k,l)).r; }} " +
                   " sun_factor *= pow(percentLit /= 9.0f, 2.2); " +

                 "} " +

           "      break; " +
           "    } " +
           "  } " +
       "}" +
   "} " +

   "LightColor += saturate(dot(input.normal, SunLightDirection)) * sun_factor * SunLightColor ; " +


   // specular light

   "SpecularLightColor = float3(0.0, 0.0, 0.0); " +

   // simple atmospherical reflexion
   "if (ambiantlight_on * ShinyFactor > 0.0) " +
   "{" +
     "if (reflect (normalize ((float3)input.posW - eyePosition), input.normal).y >= 0.0)" +
      " SpecularLightColor += AmbiantLightColor * ShinyFactor * 0.064; " +
   "}" +

   // specular part for sunlight
   "if (SpecularCone > 0.0) " +
   "{ " +
     "float3 v = normalize(reflect (-SunLightDirection, input.normal));" +
     "toEye = normalize (eyePosition - (float3)input.posW);" +
     "float specularFactor = pow (max(dot (v, toEye), 0.0f), SpecularCone);" + // [0..1]
     "SpecularLightColor += 2.0*specularFactor * sun_factor * SunLightColor;" +
   "}" +


   // point lights
   "for (i=0; i<NbLights; i++) " +
   "{" +
      "float3 LightVector     = Light[i].position - (float3)input.posW;" +
      "float  LightVectorLen2 = dot (LightVector, LightVector);" +
      "float  LightRange      = Light[i].range; " +
      "if (LightVectorLen2 < LightRange * LightRange && LightVectorLen2 > 0.0) " +
      "{" +
        "float LightVectorLen = sqrt (LightVectorLen2);" +
        "float Factor;" +

        "LightVector /= LightVectorLen;" +          // normalize light vector
        "Factor = dot (LightVector, input.normal);" +   // [0..1]

        "if (Light[i].cone > 0.0f)" +
        "{" +
          "Factor *= pow(max(dot(-LightVector, Light[i].direction), 0.0f), Light[i].cone); " +
        "}" +

        "if (Factor > 0.0f)" +   // [0..1]
        "{" +
          "float Intensity = pow((LightRange - LightVectorLen) / LightRange, Light[i].attenuation); " +
          "if (SpecularCone > 0.0) " +
          "{ " +
            "float3 v = normalize(reflect (-LightVector, input.normal));" +
            "float specularFactor = pow (max(dot (v, toEye), 0.0f), SpecularCone);" + // [0..1]
            "SpecularLightColor += 2.0*specularFactor * Intensity * Light[i].color;" +
          "}" +

          "LightColor += Factor * Intensity * Light[i].color;" +
        "}" +
      "}" +
    "}" +

    // combine all lights
    "ElementColor = shaderTexture[0].Sample(SampleType[0], input.tex) * MaterialColor * input.diffuse;" +
    "finalColor = saturate(float4(LightColor + EmissiveColor + SpecularLightColor, 1.0) * ElementColor " +
                           " + float4(SpecularLightColor * ShinyFactor, 0.0));" +

    // fog
    "if (fogInvRange != 0.0) " +  // static condition
    "{" +
      "float fogFactor = saturate((length(eyePosition - (float3)input.posW) - fogStart) * fogInvRange); " +
      "float factor = pow (fogFactor, FogInvExponent);" +
      "finalColor.rgb = factor * fogColor.rgb + (1.0 - factor) * finalColor.rgb;" +
    "}" +

    // produce output for 2 targets : screen and picking
    "output.color = finalColor; " +
    "output.objectid = 0; " +
    "if (finalColor.a > 0.0 && object_id.a > 0.0)  output.objectid = object_id; " +   // not transparent : store object id

    "return output;" +
   "} ";

//---------------------------------------------------------------------
