
// shader_vs.h : vertex shader

//---------------------------------------------------------------------

const string SHADOW_COMPUTATION =

    "float3 compute_shadow_position (float3 posW, uint instance) " +
    "{ " +
      "float3 position; " +

      "if (shadowMatrix._m33 == 0.0) return float3(-1.0, -1.0, -1.0); " +   // shadows are OFF

      // PosW = world position means relative to camera area
      "position = mul(float4(posW,1.0), shadowMatrix).xyz;  " +

      "static const float res[5] = {256.0, 64.0, 16.0, 4.0, 1.0}; " +

      "position.xy *= res[instance];  " +

      "return position; " +
    "} ";


//---------------------------------------------------------------------

////////////////////////////////////////////////////////////////////////////////
// basic vertex shader
////////////////////////////////////////////////////////////////////////////////

const string VERTEX_SHADER_BASIC_RENDERING =

    SHADOW_COMPUTATION +

    "PixelInputType MyVertexShader (VertexInputType input) " +
    "{  " +
      "PixelInputType output; " +

      // Calculate the position of the vertex against the world, view, and projection matrices.
      "output.posW     = mul(input.position, (float4x3)worldMatrix);  " +
      "output.position = mul(mul(float4(output.posW,1.0), viewMatrix), projectionMatrix);  " +

      // Calculate the normal vector against the world matrix only.
      "output.normal = mul((float3)input.normal, (float3x3)worldMatrix);  " +
      "output.tangent = mul((float3)input.tangent, (float3x3)worldMatrix);  " +

      "output.diffuse = input.diffuse; " +

      // compute the texture coordinates for the pixel shader.
      "output.tex[0] = input.tex[0] * (fact_u_min_1 + 1.0)   +   input.tex[1] * fact_u   +   ofs_u; " +
      "output.tex[1] = input.tex[1] * (fact_v_min_1 + 1.0)   +   input.tex[0] * fact_v   +   ofs_v; " +

      // compute shadow position
      "output.ShadowPos = compute_shadow_position (output.posW, input.instance); " +

      "return output; " +
    "} ";

//---------------------------------------------------------------------

const string VERTEX_SHADER_BASIC_BUILD_SHADOW_MAP =

    SHADOW_COMPUTATION +

    "ShadowVertexOutputType MyVertexShader (ShadowVertexInputType input) " +
    "{  " +
      "ShadowVertexOutputType  output; " +

      // Calculate real world position
      "float3 posW = mul(input.position, (float4x3)worldMatrix);  " +

      // compute shadow position
      "output.position.xyz = compute_shadow_position (posW, input.instance); " +
      "output.position.w = 1.0; " +

      "output.RenderTargetIndex = input.instance; " +

      "return output; " +
    "} ";

//---------------------------------------------------------------------

const string UNROLL_RIGGING =

    "  int   bone; " +
    "  float weight; " +
    "  matrix<float,4,4> m44; " +

    " if (i == 0) { weight = input.weight[0];  bone = input.bone & 255; } " +
    " if (i == 1) { weight = input.weight[1];  bone = (input.bone >> 8) & 255; } " +
    " if (i == 2) { weight = input.weight[2];  bone = (input.bone >> 16) & 255; } " +
    " if (i == 3) { weight = input.weight[3];  bone = (input.bone >> 24) & 255; } " +

    " if (weight == 0.0) " +
    "   break; " +

    " if (bone < 64) " +
    " { " +
      " if (bone < 32) " +
      " { " +
        " if (bone == 0) m44 = g_bone0[0]; " + " if (bone == 1) m44 = g_bone0[1]; " +
        " if (bone == 2) m44 = g_bone0[2]; " + " if (bone == 3) m44 = g_bone0[3]; " +
        " if (bone == 4) m44 = g_bone0[4]; " + " if (bone == 5) m44 = g_bone0[5]; " +
        " if (bone == 6) m44 = g_bone0[6]; " + " if (bone == 7) m44 = g_bone0[7]; " +
        " if (bone == 8) m44 = g_bone0[8]; " + " if (bone == 9) m44 = g_bone0[9]; " +
        " if (bone == 10) m44 = g_bone0[10]; " + " if (bone == 11) m44 = g_bone0[11]; " +
        " if (bone == 12) m44 = g_bone0[12]; " + " if (bone == 13) m44 = g_bone0[13]; " +
        " if (bone == 14) m44 = g_bone0[14]; " + " if (bone == 15) m44 = g_bone0[15]; " +
        " if (bone == 16) m44 = g_bone0[16]; " + " if (bone == 17) m44 = g_bone0[17]; " +
        " if (bone == 18) m44 = g_bone0[18]; " + " if (bone == 19) m44 = g_bone0[19]; " +
        " if (bone == 20) m44 = g_bone0[20]; " + " if (bone == 21) m44 = g_bone0[21]; " +
        " if (bone == 22) m44 = g_bone0[22]; " + " if (bone == 23) m44 = g_bone0[23]; " +
        " if (bone == 24) m44 = g_bone0[24]; " + " if (bone == 25) m44 = g_bone0[25]; " +
        " if (bone == 26) m44 = g_bone0[26]; " + " if (bone == 27) m44 = g_bone0[27]; " +
        " if (bone == 28) m44 = g_bone0[28]; " + " if (bone == 29) m44 = g_bone0[29]; " +
        " if (bone == 30) m44 = g_bone0[30]; " + " if (bone == 31) m44 = g_bone0[31]; " +
      " } " +
      " else " +
      " { " +
        " if (bone == 32) m44 = g_bone0[32]; " + " if (bone == 33) m44 = g_bone0[33]; " +
        " if (bone == 34) m44 = g_bone0[34]; " + " if (bone == 35) m44 = g_bone0[35]; " +
        " if (bone == 36) m44 = g_bone0[36]; " + " if (bone == 37) m44 = g_bone0[37]; " +
        " if (bone == 38) m44 = g_bone0[38]; " + " if (bone == 39) m44 = g_bone0[39]; " +
        " if (bone == 40) m44 = g_bone0[40]; " + " if (bone == 41) m44 = g_bone0[41]; " +
        " if (bone == 42) m44 = g_bone0[42]; " + " if (bone == 43) m44 = g_bone0[43]; " +
        " if (bone == 44) m44 = g_bone0[44]; " + " if (bone == 45) m44 = g_bone0[45]; " +
        " if (bone == 46) m44 = g_bone0[46]; " + " if (bone == 47) m44 = g_bone0[47]; " +
        " if (bone == 48) m44 = g_bone0[48]; " + " if (bone == 49) m44 = g_bone0[49]; " +
        " if (bone == 50) m44 = g_bone0[50]; " + " if (bone == 51) m44 = g_bone0[51]; " +
        " if (bone == 52) m44 = g_bone0[52]; " + " if (bone == 53) m44 = g_bone0[53]; " +
        " if (bone == 54) m44 = g_bone0[54]; " + " if (bone == 55) m44 = g_bone0[55]; " +
        " if (bone == 56) m44 = g_bone0[56]; " + " if (bone == 57) m44 = g_bone0[57]; " +
        " if (bone == 58) m44 = g_bone0[58]; " + " if (bone == 59) m44 = g_bone0[59]; " +
        " if (bone == 60) m44 = g_bone0[60]; " + " if (bone == 61) m44 = g_bone0[61]; " +
        " if (bone == 62) m44 = g_bone0[62]; " + " if (bone == 63) m44 = g_bone0[63]; " +
      " } " +
    " } " +
    " else if (bone < 128) " +
    " { " +
      " if (bone < 96) " +
      " { " +
        " if (bone == 64) m44 = g_bone1[0]; " + " if (bone == 65) m44 = g_bone1[1]; " +
        " if (bone == 66) m44 = g_bone1[2]; " + " if (bone == 67) m44 = g_bone1[3]; " +
        " if (bone == 68) m44 = g_bone1[4]; " + " if (bone == 69) m44 = g_bone1[5]; " +
        " if (bone == 70) m44 = g_bone1[6]; " + " if (bone == 71) m44 = g_bone1[7]; " +
        " if (bone == 72) m44 = g_bone1[8]; " + " if (bone == 73) m44 = g_bone1[9]; " +
        " if (bone == 74) m44 = g_bone1[10]; " + " if (bone == 75) m44 = g_bone1[11]; " +
        " if (bone == 76) m44 = g_bone1[12]; " + " if (bone == 77) m44 = g_bone1[13]; " +
        " if (bone == 78) m44 = g_bone1[14]; " + " if (bone == 79) m44 = g_bone1[15]; " +
        " if (bone == 80) m44 = g_bone1[16]; " + " if (bone == 81) m44 = g_bone1[17]; " +
        " if (bone == 82) m44 = g_bone1[18]; " + " if (bone == 83) m44 = g_bone1[19]; " +
        " if (bone == 84) m44 = g_bone1[20]; " + " if (bone == 85) m44 = g_bone1[21]; " +
        " if (bone == 86) m44 = g_bone1[22]; " + " if (bone == 87) m44 = g_bone1[23]; " +
        " if (bone == 88) m44 = g_bone1[24]; " + " if (bone == 89) m44 = g_bone1[25]; " +
        " if (bone == 90) m44 = g_bone1[26]; " + " if (bone == 91) m44 = g_bone1[27]; " +
        " if (bone == 92) m44 = g_bone1[28]; " + " if (bone == 93) m44 = g_bone1[29]; " +
        " if (bone == 94) m44 = g_bone1[30]; " + " if (bone == 95) m44 = g_bone1[31]; " +
      " } " +
      " else " +
      " { " +
        " if (bone == 96) m44 = g_bone1[32]; " + " if (bone == 97) m44 = g_bone1[33]; " +
        " if (bone == 98) m44 = g_bone1[34]; " + " if (bone == 99) m44 = g_bone1[35]; " +
        " if (bone == 100) m44 = g_bone1[36]; " + " if (bone == 101) m44 = g_bone1[37]; " +
        " if (bone == 102) m44 = g_bone1[38]; " + " if (bone == 103) m44 = g_bone1[39]; " +
        " if (bone == 104) m44 = g_bone1[40]; " + " if (bone == 105) m44 = g_bone1[41]; " +
        " if (bone == 106) m44 = g_bone1[42]; " + " if (bone == 107) m44 = g_bone1[43]; " +
        " if (bone == 108) m44 = g_bone1[44]; " + " if (bone == 109) m44 = g_bone1[45]; " +
        " if (bone == 110) m44 = g_bone1[46]; " + " if (bone == 111) m44 = g_bone1[47]; " +
        " if (bone == 112) m44 = g_bone1[48]; " + " if (bone == 113) m44 = g_bone1[49]; " +
        " if (bone == 114) m44 = g_bone1[50]; " + " if (bone == 115) m44 = g_bone1[51]; " +
        " if (bone == 116) m44 = g_bone1[52]; " + " if (bone == 117) m44 = g_bone1[53]; " +
        " if (bone == 118) m44 = g_bone1[54]; " + " if (bone == 119) m44 = g_bone1[55]; " +
        " if (bone == 120) m44 = g_bone1[56]; " + " if (bone == 121) m44 = g_bone1[57]; " +
        " if (bone == 122) m44 = g_bone1[58]; " + " if (bone == 123) m44 = g_bone1[59]; " +
        " if (bone == 124) m44 = g_bone1[60]; " + " if (bone == 125) m44 = g_bone1[61]; " +
        " if (bone == 126) m44 = g_bone1[62]; " + " if (bone == 127) m44 = g_bone1[63]; " +
      " } " +
    " } " +
    " else " +
    " { " +
      " if (bone < 160) " +
      " { " +
        " if (bone == 128) m44 = g_bone2[0]; " + " if (bone == 129) m44 = g_bone2[1]; " +
        " if (bone == 130) m44 = g_bone2[2]; " + " if (bone == 131) m44 = g_bone2[3]; " +
        " if (bone == 132) m44 = g_bone2[4]; " + " if (bone == 133) m44 = g_bone2[5]; " +
        " if (bone == 134) m44 = g_bone2[6]; " + " if (bone == 135) m44 = g_bone2[7]; " +
        " if (bone == 136) m44 = g_bone2[8]; " + " if (bone == 137) m44 = g_bone2[9]; " +
        " if (bone == 138) m44 = g_bone2[10]; " + " if (bone == 139) m44 = g_bone2[11]; " +
        " if (bone == 140) m44 = g_bone2[12]; " + " if (bone == 141) m44 = g_bone2[13]; " +
        " if (bone == 142) m44 = g_bone2[14]; " + " if (bone == 143) m44 = g_bone2[15]; " +
        " if (bone == 144) m44 = g_bone2[16]; " + " if (bone == 145) m44 = g_bone2[17]; " +
        " if (bone == 146) m44 = g_bone2[18]; " + " if (bone == 147) m44 = g_bone2[19]; " +
        " if (bone == 148) m44 = g_bone2[20]; " + " if (bone == 149) m44 = g_bone2[21]; " +
        " if (bone == 150) m44 = g_bone2[22]; " + " if (bone == 151) m44 = g_bone2[23]; " +
        " if (bone == 152) m44 = g_bone2[24]; " + " if (bone == 153) m44 = g_bone2[25]; " +
        " if (bone == 154) m44 = g_bone2[26]; " + " if (bone == 155) m44 = g_bone2[27]; " +
        " if (bone == 156) m44 = g_bone2[28]; " + " if (bone == 157) m44 = g_bone2[29]; " +
        " if (bone == 158) m44 = g_bone2[30]; " + " if (bone == 159) m44 = g_bone2[31]; " +
      " } " +
      " else " +
      " { " +
        " if (bone == 160) m44 = g_bone2[32]; " + " if (bone == 161) m44 = g_bone2[33]; " +
        " if (bone == 162) m44 = g_bone2[34]; " + " if (bone == 163) m44 = g_bone2[35]; " +
        " if (bone == 164) m44 = g_bone2[36]; " + " if (bone == 165) m44 = g_bone2[37]; " +
        " if (bone == 166) m44 = g_bone2[38]; " + " if (bone == 167) m44 = g_bone2[39]; " +
        " if (bone == 168) m44 = g_bone2[40]; " + " if (bone == 169) m44 = g_bone2[41]; " +
        " if (bone == 170) m44 = g_bone2[42]; " + " if (bone == 171) m44 = g_bone2[43]; " +
        " if (bone == 172) m44 = g_bone2[44]; " + " if (bone == 173) m44 = g_bone2[45]; " +
        " if (bone == 174) m44 = g_bone2[46]; " + " if (bone == 175) m44 = g_bone2[47]; " +
        " if (bone == 176) m44 = g_bone2[48]; " + " if (bone == 177) m44 = g_bone2[49]; " +
        " if (bone == 178) m44 = g_bone2[50]; " + " if (bone == 179) m44 = g_bone2[51]; " +
        " if (bone == 180) m44 = g_bone2[52]; " + " if (bone == 181) m44 = g_bone2[53]; " +
        " if (bone == 182) m44 = g_bone2[54]; " + " if (bone == 183) m44 = g_bone2[55]; " +
        " if (bone == 184) m44 = g_bone2[56]; " + " if (bone == 185) m44 = g_bone2[57]; " +
        " if (bone == 186) m44 = g_bone2[58]; " + " if (bone == 187) m44 = g_bone2[59]; " +
        " if (bone == 188) m44 = g_bone2[60]; " + " if (bone == 189) m44 = g_bone2[61]; " +
        " if (bone == 190) m44 = g_bone2[62]; " + " if (bone == 191) m44 = g_bone2[63]; " +
      " } " +
    " } ";

//---------------------------------------------------------------------

////////////////////////////////////////////////////////////////////////////////
// vertex shader for rigged mesh
////////////////////////////////////////////////////////////////////////////////

const string VERTEX_SHADER_RIGGED_RENDERING =

    SHADOW_COMPUTATION +

    "PixelInputType MyVertexShader (VertexInputTypeRiggedMesh input)  " +
    "{  " +
      "PixelInputType output;  " +
      "int            i;       " +
      "float3         position, normal, tangent; " +

      "position = float3(0.0, 0.0, 0.0);  " +
      "normal   = float3(0.0, 0.0, 0.0);  " +
      "tangent  = float3(0.0, 0.0, 0.0);  " +

      "if (indexes_supported) " +
      "{" +

        "for (i=0; i<4 && input.weight[i] > 0.0; i++)  " +
        "{  " +
          "int bone = (input.bone >> (8*i)) & 255; " +
          "matrix<float,4,4> m44;  " +

          "if (bone < 64)       m44 = g_bone0[(bone)];     " +
          "else if (bone < 128) m44 = g_bone1[(bone-64)];  " +
          "else                 m44 = g_bone2[(bone-128)]; " +

          "position += (float3)(input.weight[i] * mul(m44, input.position));  " +
          "normal   += input.weight[i] * mul((float3x3)m44, input.normal);  " +
          "tangent  += input.weight[i] * mul((float3x3)m44, input.tangent);  " +
        "}  " +

      "}" +
      "else " +
      "{" +
         // AMD does not support dynamic array indexes, so we have to do this with static indexes :
         // static indexes means variables of loops.
        "for (i=0; i<4; i++)  " +
        "{  " +
          UNROLL_RIGGING +
          "position += (float3)(weight * mul(m44, input.position));  " +
          "normal   += weight * mul((float3x3)m44, input.normal);  " +
          "tangent  += weight * mul((float3x3)m44, input.tangent);  " +
        "}  " +
      "}" +

      // Calculate the position of the vertex against the world, view, and projection matrices.
      "output.posW     = mul(float4(position,1.0), (float4x3)worldMatrix);  " +
      "output.position = mul(mul(float4(output.posW,1.0), viewMatrix), projectionMatrix);  " +

      // Calculate the normal vector against the world matrix only.
      "output.normal = mul(normalize(normal), (float3x3)worldMatrix);  " +
      "output.tangent = mul(normalize(tangent), (float3x3)worldMatrix);  " +

      "output.diffuse = float4(1.0, 1.0, 1.0, 1.0);  " +

      // compute the texture coordinates for the pixel shader.
      "output.tex[0] = input.tex[0] * (fact_u_min_1 + 1.0)   +   input.tex[1] * fact_u   +   ofs_u; " +
      "output.tex[1] = input.tex[1] * (fact_v_min_1 + 1.0)   +   input.tex[0] * fact_v   +   ofs_v; " +

      // compute shadow position
      "output.ShadowPos = compute_shadow_position (output.posW, input.instance); " +

      "return output;  " +
    "} ";


//---------------------------------------------------------------------

const string VERTEX_SHADER_RIGGED_BUILD_SHADOW_MAP =

    SHADOW_COMPUTATION +

    "ShadowVertexOutputType MyVertexShader (ShadowVertexInputTypeRiggedMesh input)  " +
    "{  " +
      "ShadowVertexOutputType output;  " +
      "int            i;       " +
      "float3         position;  " +

      "position = float3(0.0, 0.0, 0.0);  " +

      "if (indexes_supported) " +
      "{" +

        "for (i=0; i<4 && input.weight[i] > 0.0; i++)  " +
        "{  " +
          "int bone = (input.bone >> (8*i)) & 255; " +
          "matrix<float,4,4> m44;    " +

          "if (bone < 64)       m44 = g_bone0[(bone)];     " +
          "else if (bone < 128) m44 = g_bone1[(bone-64)];  " +
          "else                 m44 = g_bone2[(bone-128)]; " +

          "position += (float3)(input.weight[i] * mul(m44, input.position));  " +
        "}  " +

      "}" +
      "else " +
      "{" +
         // AMD does not support dynamic array indexes, so we have to do this with static indexes :
         // static indexes means variables of loops.
        "for (i=0; i<4; i++)  " +
        "{  " +
          UNROLL_RIGGING +
          "position += (float3)(weight * mul(m44, input.position));  " +
        "}  " +
      "}" +

      // Calculate the position of the vertex against the world, view, and projection matrices.
      "float3 posW = mul(float4(position,1.0), (float4x3)worldMatrix);  " +

      // compute shadow position
      "output.position.xyz = compute_shadow_position (posW, input.instance); " +
      "output.position.w = 1.0; " +

      "output.RenderTargetIndex = input.instance; " +

      "return output;  " +
    "} ";

//---------------------------------------------------------------------
